import numpy as np
import matplotlib.pyplot as plt
import copy

def nextLine(f):
    l = f.readline().strip()
    while not l:
        l = f.readline().strip()
    return l


class Point:
    def __init__(self, x=0.0, y=0.0, z=0.0):
        self.x, self.y, self.z = x, y, z

    @staticmethod
    def readFromStream(f):
        coords =  map(float, nextLine(f).split())
        return Point(*coords)

    
    def __str__(self):
        return "{:10f} {:10f} {:10f}".format(self.x, self.y, self.z)



class FWIObservationSpecification(list):
    def __init__(self, t_start=0, t_end=1.0, t_delta=0.1):
        list.__init__(self)
        self.t_start, self.t_end, self.t_delta = t_start, t_end, t_delta
                
    @staticmethod
    def readFromStream(f):
        t_start = float(nextLine(f))
        t_end = float(nextLine(f))
        t_delta = float(nextLine(f))
        
        spec = FWIObservationSpecification(t_start, t_end, t_delta)
        
        n_receivers = int(nextLine(f))
        for i in range(n_receivers):
            spec.append(Point.readFromStream(f))
            
        return spec

    def __str__(self):
        return "".join(("FWIObservationSpecification:\n",
                       "  times: [{}, {}],  delta_t: {}\n".format(self.t_start, self.t_end,self.t_delta),
                       "  receiver count: {}\n".format(""),
                       "  reveiver positions: \n    ",
                       "\n    ".join(str(R) for R in self)
                       ))


class FWITimeseries(list):
    def __init__(self):
        list.__init__(self)
        self.receiver = None
    
    @staticmethod
    def readFromStream(f):
        ts = FWITimeseries()
        ts.receiver = Point.readFromStream(f)
        
        n_measurements = int(nextLine(f))
        for i in range(n_measurements):
            _ = list(map(float, nextLine(f).split()))
            ts.append((_[0], np.array(_[1:])))
        
        return ts
    
    def min(self):
        return min(min(m) for t,m in self)
    
    def max(self):
        return max(max(m) for t,m in self)
    
    def argmax(self, comp=0):
        tM, M = 0.0, -1e999
        for t,m in self:
            if m[comp] > M:
                tM,M = t, m[comp]
        return tM
    
    def __sub__(self, ts):
        diff = copy.deepcopy(self)
        for (t1,m1),(t2,m2) in zip(diff,ts):
            m1 -= m2
        return diff
    
    def norm2squared(self):
        return sum(m.T@m for t,m in self)
    
    def __str__(self):
        return "".join(("    receiver: {}\n".format(self.receiver),
                        "\n".join("      t: {:6f} -  {}".format(
                            t, " ".join("{:10f}".format(ml) for ml in m)) for t,m in self)
                        ))
    

class FWISeismogram:
    def __init__(self, label=""):
        self.spec = FWIObservationSpecification()
        self.timeseries = []
        self.state_components = 0
        self.label = label
        
    
    @staticmethod
    def readFromStream(f):
        s = FWISeismogram()
        s.spec = FWIObservationSpecification.readFromStream(f)
        s.state_components = int(nextLine(f))
            
        n_timeseries = int(nextLine(f))
        for i in range(n_timeseries):
            s.timeseries.append(FWITimeseries.readFromStream(f))
       
        s.timeseries.sort(key=lambda ts: ts.receiver.x)
        return s
    
    def min(self):
        return min(ts.min() for ts in self.timeseries)
    
    def max(self):
        return max(ts.max() for ts in self.timeseries)
    
    def __str__(self):
        return "".join(("FWISeismogram:\n",
                        str(self.spec), "\n",
                        "  measurements:\n",
                        "\n".join(str(ts) for ts in self.timeseries)
                        ))
    
    def toArrays(self, comp=0):
        shape = (len(self.timeseries[0]), len(self.timeseries))
        data = np.zeros(shape, dtype=np.float64)
        times = np.zeros(shape[0], dtype=np.float64)
        
        #my_timeseries = self.timeseries.copy()
        #my_timeseries.sort(key=lambda ts: ts.receiver.x)
        
        for j,ts in enumerate(self.timeseries):
            for i,(t,m) in enumerate(ts):
                data[i,j] = m[comp]
                times[i] = t
                
        return times,data
    
    def __sub__(self, s):
        diff = FWISeismogram()
        diff.spec = s.spec
        diff.state_components = s.state_components
        for ts1,ts2 in zip(self.timeseries,s.timeseries):
            diff.timeseries.append(ts1-ts2)
        return diff
    
    def norm2squared(self):
        dt = self.spec.t_delta
        return dt*dt * sum(ts.norm2squared() for ts in self.timeseries)
    
    def plot(self, comp=0, color_factor=0.2, color='black', figure=None, data_range=None, data_factor=1.0):        
        if figure is None:
            figure = plt.figure()
            
        ax = figure.gca()
            
        #plt.figure(figsize=figsize)
        
        times,data = self.toArrays(comp)
            
        if data_range is None:
            data_range = np.min(data), np.max(data)
            
        R = data.shape[1]
        #R = 1
        
        amplitude = max(map(abs, data_range))
        y_limits = (0, 2*R*amplitude)
        y_bnd_offsets = (-data_factor*amplitude, data_factor*amplitude)
        ax.set_ylim(y_limits[0] + y_bnd_offsets[0], y_limits[1] + y_bnd_offsets[1])
    
            
        #m,M = np.min(data), np.max(data)
        m,M = data_range
        absM = max(abs(m),abs(M))        
    
        ym,yM = y_limits
        dY = (yM-ym)/R
        
        data = data * data_factor*dY/(2*absM)
        
        for i in range(R):
            tr = data[:,i].copy()
            #ts = list(range(len(tr)))
            offset_y = dY/2 + i*dY
            tr_offsetted = tr + offset_y
            ax.plot(times, tr_offsetted, c=color, linewidth=1.0, label=(self.label if self.label != "obs" else None))
            ax.plot(times, np.ones(len(times))*offset_y, color="black", linewidth=0.5)
            maximum, minimum = np.max(tr), np.min(tr)
            #ax.fill_between(times, tr_offsetted, 
                             #offset_y+color_factor*maximum, 
                             #where=(tr>=color_factor*maximum), 
                             #facecolor=color, interpolate=True)
            #ax.fill_between(times, tr_offsetted, 
                             #offset_y+color_factor*maximum, 
                             #where=(tr<color_factor*minimum), 
                             #facecolor="white", interpolate=True, zorder=-10)
            
            #plt.fill_between(tr, times, offset_y, where=(tr<offset_y), facecolor="white", zorder=-10)

        #plt.xticks([]);
        plt.yticks([])
        #plt.gca().invert_yaxis()
        #plt.xlim(0.9,1.9)
        #plt.ylim(-18,41)
        #title = plt.title('Seismogram', fontsize=16.5)
        #ax.ylabel("x", fontsize=14.5)
        ax.yaxis.set_label_position("right")
        #plt.ylabel("  t", rotation="horizontal", fontsize=14.5)
        figure.tight_layout(rect=(0,0,1,1))
        plt.legend(prop={'size':10})
        #plt.savefig("/tmp/seismogram.pdf")
        #plt.show()

#fig = plt.figure(figsize=(2, 4))
fig = plt.figure(figsize=(1.8, 4))

s_real = FWISeismogram.readFromStream(open("fwi_CGREGINN_seismogram_observed"))
s_real.label = "obs"

colors = ["blue", "grey", "red", "green"]
s_approx, s_residuals = [], []

#for i in [10]:
for i in []:
    s_current = FWISeismogram.readFromStream(open("fwi_CGREGINN_seismogram_{}".format(i)))
    s_current.label = (i if type(i) == str else "iter {}".format(i))
    s_approx.append(s_current)
    s_residuals.append(s_real-s_current)
    s_residuals[-1].label = s_current.label

plot_seismograms = True
if plot_seismograms:
    #data_range = min(s_real.min(), min(s.min() for s in s_approx)), max(s_real.max(), max(s.max() for s in s_approx))
    data_range = s_real.min(), s_real.max()
    kwargs = dict(data_range=data_range, data_factor=10.0)

    s_real.plot(0, color_factor=1.0, figure=fig, color="black", **kwargs)

    for s,c in zip(s_approx, colors):
        s.plot(0, color_factor=1.0, figure=fig, color=c, **kwargs)
        
else: # plot residuals
    data_range = min(s.min() for s in s_residuals), max(s.max() for s in s_residuals)
    print(data_range)
    kwargs = dict(data_range=data_range, data_factor=2.0)

    for s,c in zip(s_residuals, colors):
        print(s.norm2squared())
        s.plot(0, color_factor=1.0, figure=fig, color=c, **kwargs)

plt.show()