import numpy as np
from numpy.linalg import norm

import scipy as sp
import scipy.optimize as opt

import matplotlib.pyplot as plt
import matplotlib.patches as patches

class TravelLine:
    def __init__(self, points, vs):
        self.points = points
        self.vs = vs
        
    def plot(self, ax=None):
        if ax is None:
            ax = plt.gca()
        
        ax.plot(*zip(*self.points))
        
    def duration(self):
        times = [norm(np.array(pe)-np.array(ps))/v for ps,pe,v in zip(self.points[:-1], self.points[1:], self.vs)]
        return float(sum(times))
        

class Setting:
    def __init__(self, source, source_radius, receivers, d, h, v1, v2, bnd_l, bnd_r, bnd_b, buffer_time):
        self.source = source
        self.source_radius = source_radius
        self.receivers = receivers
        self.d, self.h = d, h
        self.v1, self.v2 = v1, v2
        self.bnd_left = bnd_l
        self.bnd_right = bnd_r
        self.bnd_bottom = bnd_b
        self.buffer_time = buffer_time
        
        
    def get_obstacle_top_travelline(self, s, r):
        x = (s[0]+r[0])/2
        return TravelLine([s, [x,self.d], r], [self.v1, self.v1])
        
        
    def get_obstacle_bottom_travelline(self, s, r):
        v1, v2, d = self.v1, self.v2, self.d
        target = [(s[0]+r[0])/2, d+self.h]
        def objective(x):
            d1 = np.sqrt((s[0]-x)**2 + (s[1]-d)**2)
            d2 = np.sqrt((target[0]-x)**2 + (target[1]-d)**2)
            return d1/v1 + d2/v2
        
        x_best = float(opt.fmin(objective, (s[0]+r[0])/2))
        return TravelLine([s, [x_best,d], target, [r[0]-(x_best-s[0]),d], r], [v1,v2,v2,v1])
        
    def round(self, x):
        #return x
        return np.ceil(x/self.h)*self.h
        
        
    def plot(self):
        ax = plt.gca()
        
        ################################
        # plot source
        sources = [self.source]
        plt.plot(*zip(*sources), "ro")
        for source in sources:
            ax.add_patch(patches.Circle(source, self.source_radius, alpha=0.3))
        
        ################################
        # plot receivers
        receivers = self.receivers
        plt.plot(*zip(*receivers), "bX")
                
        ################################
        # plot min-time distance
        closest_source= self.source
        closest_receiver = min(self.receivers, key=lambda r: r[0])
        Candidates_fastest = [self.get_obstacle_top_travelline(closest_source, closest_receiver),
                               self.get_obstacle_bottom_travelline(closest_source, closest_receiver)]
        T_fastest = min(Candidates_fastest, key=lambda T: T.duration())
        T_fastest.plot(ax)
        
        ################################
        # plot max-time distance
        farest_source= self.source
        farest_receiver = max(self.receivers, key=lambda r: r[0])
        Candidates_slowest = [self.get_obstacle_bottom_travelline(farest_source, farest_receiver),
                              self.get_obstacle_top_travelline(farest_source, farest_receiver)]
        T_slowest = max(Candidates_slowest, key=lambda T: T.duration())
        T_slowest.plot(ax)
        
        print("T_fastest:  {:f}".format(T_fastest.duration() - self.source_radius/self.v1))
        print("T_slowest:  {:f}".format(T_slowest.duration() + self.source_radius/self.v1))
        
        ################################
        # get boundary distances
        T = T_slowest.duration() + self.buffer_time
        travel_dist = T*self.v1
        self.bnd_left = self.round(-(travel_dist + self.source_radius)/2 + (farest_source[0] + closest_receiver[0])/2)
        self.bnd_right = self.round(0.5 * (travel_dist + self.source_radius + farest_receiver[0] - closest_source[0]))
        
        print("Bnd_Left: {:f}".format(self.bnd_left))
        print("Bnd_Right: {:f}".format(self.bnd_right))
        
        # bottom boundary
        h, v1, v2 = self.h, self.v1, self.v2
        def t_bottom(B):
            dx = (closest_source[0] + closest_receiver[0])/2
            dy = B - closest_source[1]
            l = np.sqrt(dy**2 + dx**2)
            return 2 * ( (l - (h*l)/dy)/v1 + h*l/v2 )
        
        self.bnd_bottom = self.round(max(float(opt.fsolve(lambda B: t_bottom(B) - T, self.bnd_bottom)), 0.85))
        M = [(closest_source[0] + closest_receiver[0])/2, self.bnd_bottom]
        TravelLine([closest_source, M, closest_receiver], [-1,-1]).plot(ax)
        
        print("Bnd_Bottom: {:f}".format(self.bnd_bottom))
        
        ################################
        # plot bounding box and set limits
        xs = [self.source[0]] + [r[0] for r in self.receivers] + [self.bnd_left, self.bnd_right]
        ys = [self.source[1]] + [r[1] for r in self.receivers] + [0,1] + [self.bnd_bottom]
        
        min_x, max_x = min(xs), max(xs)
        min_y, max_y = min(ys), max(ys)
        
        offset_x = (max_x-min_x)*0.1
        offset_y = (max_y-min_y)*0.1
        
        plt.xlim(min_x-offset_x, max_x+offset_x)
        plt.ylim(max_y+offset_y, min_y-offset_y)
        
        plt.yticks(np.arange(0, self.bnd_bottom, self.h))
        plt.xticks(np.arange(self.bnd_left, self.bnd_right, self.h))
        plt.grid(True)
        
        plt.plot([self.bnd_left, self.bnd_right], [0,0], 'k-')
        plt.plot([self.bnd_left, self.bnd_left],  [0,self.bnd_bottom], 'k-')
        plt.plot([self.bnd_left, self.bnd_right], [self.bnd_bottom,self.bnd_bottom], 'k-')
        plt.plot([self.bnd_right, self.bnd_right],  [0,self.bnd_bottom], 'k-')
        
        # plot obstacle
        ax = plt.gca()
        ax.add_patch(patches.Rectangle((min_x,self.d), 
                                       max_x-min_x, self.h,
                                       alpha=0.3))
        
        plt.show()
        
        
y = 0.0625
setting = Setting([-0.0625, y],    # source
                  0.0625,          # source radius
                  [[x,y] for x in np.linspace(0.15,0.6,10)], # receivers
                  0.625,          # obstacle position
                  0.125,        # obstacle width
                  1, 2,         # wave speeds
                  -1, 1, 1,     # initial boundaries
                  0.1           # buffer time
                  )
setting.plot()
