#include "lib5_solve/solver/preconditioner/Jacobi.hpp"
#include "FWIRKOperator.hpp"
#include "FWIHelpers.hpp"
//#include "../TimeStepping/src/DGViscoAcoustic.hpp"
#include "DGAssemble.hpp"

RKOperator::RKOperator(Matrix *_M, Matrix *_A, const Vector &U, const inteOptions &_opt)
    : M(_M), A(_A), B(U), C(U), opt(_opt) {
  B = 0.0;
  C = 0.0;

}

RKOperator::RKOperator(std::shared_ptr<DGAssemble> _assemble, const Vector &U,
                       const inteOptions &_opt) : B(U), C(U), opt(_opt),
                                                  assemble(_assemble) {
  B = 0.0;
  C = 0.0;
  assembleMatrices(*assemble, U);
}

RKOperator::RKOperator(std::shared_ptr<DGAssemble> _assemble, const Vector &U,
                       const inteOptions &_opt, int rkorder, double dt) : B(U), C(U), opt(_opt),
                                                                          assemble(_assemble) {
  B = 0.0;
  C = 0.0;
  assembleMatrices(*assemble, U);
  initialize(rkorder, dt);
}

void RKOperator::assembleMatrices(DGAssemble &assemble_, const Vector &U) {
  cleanUpMatrizes = true;
  A = std::make_unique<Matrix>(U);
  M = std::make_unique<Matrix>(U);
//    MKappa = new Matrix(U);
  DGAssemble &assembl = static_cast<DGAssemble &>(assemble_);
//    assembl.MassMatrixOnlyKappa(*MKappa);
  assemble_.MassMatrix(*M);
  assemble_.FluxMatrix(*A);
  M->CreateSparse();
  A->CreateSparse();
//    MKappa->CreateSparse();

}

void RKOperator::initialize(int rkorder, double _dt) {
  initialized = true;
  dt = _dt;
  fillQuadrature(c, b, rkorder);
  if (rkorder == -2) {Exit("order -2 not suitable for this operator"); }
  for (size_t stage = 0; stage < c.size(); ++stage) {
    if (rint(c[stage]) == 1) {
      B = *A;
      B *= ((-1.0) * dt * b[stage]);
    }
    if (rint(c[stage]) == 0) {
      C = *A;
      C *= ((-1.0) * dt * b[stage]);
    }
  }
  if (!opt.fwd && opt.debug)
    std::swap(B, C); // macht den RKBwd Test negativ, erzeugt aber dierichtige Adjungierte!
  C.GetData() -= (*M).GetData();
  B.GetData() += (*M).GetData();
  B.CreateSparse();
  C.CreateSparse();
//    delete A;
}

void RKOperator::runLoop(const Vectors &WaveFieldPoint, Vectors &WaveFieldEvaluation) const {
  if (!initialized) {Exit("Run initializer first"); }
  if (opt.fwd) loopFwd(WaveFieldPoint, WaveFieldEvaluation);
  if (!opt.fwd) loopBwd(WaveFieldPoint, WaveFieldEvaluation);
}

void RKOperator::loopFwd(const Vectors &WaveFieldPoint, Vectors &WaveFieldEvaluation) const {
  Date Time;
  WaveFieldEvaluation[0] += (*M) * WaveFieldPoint[0];
  Vector Cu(0.0, WaveFieldEvaluation[0]);
  Vector Bu(0.0, WaveFieldEvaluation[0]);
  Vector tmp(WaveFieldEvaluation[0]);
  double thresholdInt = threshold;
  for (int timeStep = 1; timeStep < WaveFieldEvaluation.size(); ++timeStep) {
    if (static_cast<double>(timeStep) / static_cast<double>(WaveFieldEvaluation.size()) >
        threshold) {
      thresholdInt += thresholdIncrease;
    }
    if (!WaveFieldPoint[timeStep - 1].GetAccumulateFlag()) {
      tmp = WaveFieldPoint[timeStep - 1];
      tmp.Accumulate();
      Cu = C * tmp;
    } else {
      Cu = C * WaveFieldPoint[timeStep - 1];
    }
    if (!WaveFieldPoint[timeStep].GetAccumulateFlag()) {
      tmp = WaveFieldPoint[timeStep];
      tmp.Accumulate();
      Bu = B * tmp;
    } else {
      Bu = B * WaveFieldPoint[timeStep];
    }
    Bu += Cu;
    WaveFieldEvaluation[timeStep] += Bu;
  }
}

void RKOperator::loopFwdOneStep(const Vectors &WaveFieldPoint, Vector &WaveFieldEvaluation,
                                int timeStep) const {
  Date Time;
  if (timeStep == 0) {
    WaveFieldEvaluation += (*M) * WaveFieldPoint[0];
  } else {
    Vector Cu(0.0, WaveFieldEvaluation);
    Vector Bu(0.0, WaveFieldEvaluation);
    Vector tmp(WaveFieldEvaluation);
    double thresholdInt = threshold;
    if (static_cast<double>(timeStep) / static_cast<double>(WaveFieldEvaluation.size()) >
        threshold) {
      thresholdInt += thresholdIncrease;
    }
    if (!WaveFieldPoint[timeStep - 1].GetAccumulateFlag()) {
      tmp = WaveFieldPoint[timeStep - 1];
      tmp.Accumulate();
      Cu = C * tmp;
    } else {
      Cu = C * WaveFieldPoint[timeStep - 1];
    }
    if (!WaveFieldPoint[timeStep].GetAccumulateFlag()) {
      tmp = WaveFieldPoint[timeStep];
      tmp.Accumulate();
      Bu = B * tmp;
    } else {
      Bu = B * WaveFieldPoint[timeStep];
    }
    Bu += Cu;
    WaveFieldEvaluation += Bu;
  }

}

void RKOperator::loopBwd(const Vectors &WaveFieldPoint, Vectors &WaveFieldEvaluation) const {
  if (opt.discrete) loopBwdDiscrete(WaveFieldPoint, WaveFieldEvaluation);
  else {
    Date Time;
    mout << "Starting Evaluation" << endl;
    Vector u(WaveFieldPoint[WaveFieldEvaluation.size() - 1]);
    if (!u.GetAccumulateFlag()) u.Accumulate();
    WaveFieldEvaluation[WaveFieldEvaluation.size() - 1] += (*M) * u;
    Vector Cu(0.0, WaveFieldEvaluation[0]);
    Vector Bu(0.0, WaveFieldEvaluation[0]);
    Vector tmp(WaveFieldEvaluation[0]);

    double thresholdInt = threshold;
    for (int timeStep = 0; timeStep < WaveFieldEvaluation.size() - 1; ++timeStep) {
      double quotient =
          static_cast<double>(timeStep) / static_cast<double>(WaveFieldEvaluation.size());
      if (quotient > thresholdInt) {
        mout << "Evaluation at " << thresholdInt * 100 << "% Time needed: " << Date() - Time
             << endl;
        thresholdInt += thresholdIncrease;
      }
      if (!WaveFieldPoint[timeStep + 1].GetAccumulateFlag()) {
        tmp = WaveFieldPoint[timeStep + 1];
        tmp.Accumulate();
        Cu = C * tmp;
      } else {
        Cu = C * WaveFieldPoint[timeStep + 1];
      }
      if (!WaveFieldPoint[timeStep].GetAccumulateFlag()) {
        tmp = WaveFieldPoint[timeStep];
        tmp.Accumulate();
        Bu = B * tmp;
      } else {
        Bu = B * WaveFieldPoint[timeStep];
      }
      Bu += Cu;
      WaveFieldEvaluation[timeStep] += Bu;
    }
  }
}

void RKOperator::loopBwdOneStep(const Vectors &WaveFieldPoint, Vector &WaveFieldEvaluation,
                                int timeStep) const {
  Date Time;
  if (timeStep == WaveFieldPoint.size() -1) {
    mout << "Starting Evaluation" << endl;
    Vector u(WaveFieldPoint[timeStep]);
    if (!u.GetAccumulateFlag()) u.Accumulate();
    WaveFieldEvaluation += (*M) * u;
  } else {
    Vector Cu(0.0, WaveFieldEvaluation);
    Vector Bu(0.0, WaveFieldEvaluation);
    Vector tmp(WaveFieldEvaluation);
    double thresholdInt = threshold;
      double quotient =
          static_cast<double>(timeStep) / static_cast<double>(WaveFieldEvaluation.size());
//      if (quotient > thresholdInt) {
//        mout << "Evaluation at " << thresholdInt * 100 << "% Time needed: " << Date() - Time
//             << endl;
//        thresholdInt += thresholdIncrease;
//      }
      if (!WaveFieldPoint[timeStep + 1].GetAccumulateFlag()) {
        tmp = WaveFieldPoint[timeStep + 1];
        tmp.Accumulate();
        Cu = C * tmp;
      } else {
        Cu = C * WaveFieldPoint[timeStep + 1];
      }
      if (!WaveFieldPoint[timeStep].GetAccumulateFlag()) {
        tmp = WaveFieldPoint[timeStep];
        tmp.Accumulate();
        Bu = B * tmp;
      } else {
        Bu = B * WaveFieldPoint[timeStep];
      }
      Bu += Cu;
      WaveFieldEvaluation += Bu;
    }
}

void
RKOperator::loopBwdDiscrete(const Vectors &WaveFieldPoint, Vectors &WaveFieldEvaluation) const {
  Date Time;
  Exit("dont use for now");
  mout << "Starting Evaluation" << endl;
  Vector u(WaveFieldPoint[0]);
  Vector Cu(0.0, WaveFieldEvaluation[0]);
  WaveFieldEvaluation[0] = (*M) * u;
  u = WaveFieldPoint[1];
  Cu = C * u;
  Cu.MakeAdditive();
  WaveFieldEvaluation[0] += Cu;
//    WaveFieldEvaluation[0] = 0.0;
  Vector Bu(0.0, WaveFieldEvaluation[0]);
  double thresholdInt = threshold;
  for (int timeStep = 1; timeStep < WaveFieldEvaluation.size() - 1; ++timeStep) {
    double quotient =
        static_cast<double>(timeStep) / static_cast<double>(WaveFieldEvaluation.size());
    if (quotient > thresholdInt) {
//            mout << "Evaluation at " << thresholdInt * 100 << "% Time needed: " << Date()- Time  <<endl;
      thresholdInt += thresholdIncrease;
    }
    Bu = B * WaveFieldPoint[timeStep];
    Cu = C * WaveFieldPoint[timeStep + 1];
    Bu += Cu;
    WaveFieldEvaluation[timeStep] += Bu;
  }
  u = WaveFieldPoint[WaveFieldPoint.size() - 1];
  Bu = B * u;
  WaveFieldEvaluation[WaveFieldEvaluation.size() - 1] += Bu;
}

void RKOperator::calculateRHSRK(Vectors &rhsWF, bool fwd) const {
  Vector tmp(0.0, rhsWF[0]);
  rhsWF[0].Clear();
  for (size_t timeStep = 1; timeStep < rhsWF.size(); ++timeStep) {
    Vector &RHS = rhsWF[timeStep];
    for (size_t stage = 0; stage < c.size(); ++stage) {
      double evaluationTime = (timeStep - (1.0 - c[stage])) * dt;
      assemble->RHS(evaluationTime, tmp);
      RHS += (b[stage]) * tmp;
    }
  }
  rhsWF *= dt;
}

void RKOperator::applyM(Vectors &rhsWF) const {
//    Exit("dont use this feature for the time being");
  Vector tmp(rhsWF[0]);
  for (size_t timeStep = 0; timeStep < rhsWF.size(); ++timeStep) {
    tmp = rhsWF[timeStep];
    rhsWF[timeStep] = (*M) * tmp;
  }
}