//
// Created by cr on 15.10.21.
//

#include "FWIAdjointCalculator.hpp"

void FWIAdjointCalculator::updateTaupPositive(const int Row, const double integral_rho,
                                              const double integral_kappa_0,
                                              const RVector &integral_M, const RVector &integral_D,
                                              Vector &Adjoint_m_mv, const Cell &C) const {
  double alpha = (numL == 0) ? 0 : problem_fwd->alpha(C, C());
  double rho = Material(Row, 0);
  double V_p = Material(Row, 1);
  double grad_taup = 0.0;
  double tau_p = 0.0;
  if (numL > 0) {
    tau_p = Material(Row, 2);
  }
  double integral_kappa_int = integral_kappa_0 * exp(tau_p);
  double rhoexp = exp(rho);
  double rhofac = -(1 + alpha * exp(tau_p)) / exp(rho + 2 * V_p + tau_p);
  double V_pfac = 2.0 * rhofac;
  double grad_rho = rhoexp * integral_rho;
  grad_rho += rhofac * integral_kappa_int;
  double grad_V_P = V_pfac * integral_kappa_int;

  if (numL > 0) {
    double tau_pfac = alpha / exp(2 * V_p + rho);
    double tau_pfac2 = -1.0 / exp(2 * V_p + tau_p + rho);
    grad_taup = tau_pfac * integral_kappa_int;

    for (int l = 1; l <= numL; ++l) {
      double tau_i = Material(Row, 2 + l);
      double integralsum = integral_M[l - 1] + 1.0 / (tau_i) * integral_D[l - 1];
      grad_rho += rhofac * integralsum;
      grad_V_P += V_pfac * integralsum;
      grad_taup += tau_pfac2 * integralsum;
    }
  }
  if (recon.Rho) {
    Adjoint_m_mv(Row, 0) += grad_rho;
  }
  if (recon.Vp == 1) {
    Adjoint_m_mv(Row, 1) += grad_V_P;
  }
  if (recon.Taup == 1 && numL > 0) {
    Adjoint_m_mv(Row, 2) += grad_taup;
  }

}

void FWIAdjointCalculator::updateTaup(int Row, double integral_rho,
                                      double integral_kappa_0, const RVector &integral_M,
                                      const RVector &integral_D, Vector &Adjoint_m_mv,
                                      const Cell &C) const {
  double alpha = (numL == 0) ? 0 : problem_fwd->alpha(C, C());
    double rho = Material(Row, 0);
  double V_p = Material(Row, 1);
  double grad_taup = 0.0;
  double tau_p = (numL > 0) ? Material(Row, 2) : 0.0;


  double rhofac = -(1 + alpha * tau_p) / (squared(rho * V_p));
  double V_pfac = (-2.0) * (1.0 + alpha * tau_p) / (V_p * V_p * V_p * rho);
  double grad_rho = integral_rho;
  grad_rho += rhofac * integral_kappa_0;
  double grad_V_P = V_pfac * integral_kappa_0;

  if (numL > 0) {
    double tau_pfac = alpha / (squared(V_p) * rho);
    double tau_pfac2 = -1.0 / (squared(V_p * tau_p) * rho);
    grad_taup = tau_pfac * integral_kappa_0;
    for (int l = 1; l <= numL; ++l) {
      double tau_i = Material(Row, 2 + l);
      double integralsum = integral_M[l - 1] + 1.0 / (tau_i) * integral_D[l - 1];
      grad_rho += (rhofac / tau_p) * integralsum;
      grad_V_P += (V_pfac / tau_p) * integralsum;
      grad_taup += tau_pfac2 * integralsum;
    }
  }
  if (recon.Rho) {
    Adjoint_m_mv(Row, 0) += grad_rho;
  }
  if (recon.Vp == 1) {
    Adjoint_m_mv(Row, 1) += grad_V_P;
  }
  if (recon.Taup == 1 && numL > 0) {
    Adjoint_m_mv(Row, 2) += grad_taup;
  }
}

void FWIAdjointCalculator::updateNative(const int Row, const double integral_rho,
                                        const double integral_kappa_0, const RVector &integral_M,
                                        const RVector &integral_D, Vector &Adjoint_m_mv) const {
    double gradient_rho = integral_rho;
    double kappa_0 = Material(Row, 1);
    double gradient_kappa_0 = -1.0 / (kappa_0 * kappa_0) * integral_kappa_0;
    RVector gradient_kappa_i(numL + 1);
    RVector gradient_tau_i(numL + 1);
    for (int i = 1; i <= numL; ++i) {
        double kappa_i = Material(Row, 1 + i);
        double tau_i = Material(Row, 1 + numL + i);
        gradient_kappa_i[i - 1] = (-1.0 / (kappa_i * kappa_i) * integral_M[i - 1]
                                   - 1.0 / (tau_i * kappa_i * kappa_i) * integral_D[i - 1]);
    gradient_tau_i[i - 1] = (-1.0 / (tau_i * tau_i * kappa_i) *
                             integral_D[i - 1]); // tau_i gradient not implemented yet !
  }
  if (recon.Rho) {
    Adjoint_m_mv(Row, 0) += gradient_rho;
  }
  if (recon.Kappa0) {
    Adjoint_m_mv(Row, 1) += gradient_kappa_0;
  }

  if (recon.KappaL == 1) {
    for (int k = 1; k <= numL; ++k) {
        Adjoint_m_mv(Row, k + 1) += gradient_kappa_i[k - 1]; // probably NOT right for multiple Kappas - needs Check!
    }
  }

  if (recon.Taus == 1) {
    for (int k = 1; k <= numL; ++k) {
      Adjoint_m_mv(Row, k + 1 + numL) += gradient_tau_i[k - 1];
    }
  }
}

void FWIAdjointCalculator::calculateL2AdjointOneStep(const Vector &U_m_bwd, const Vector &U_m_fd,
                                                     const Vector &LambdaStep,
                                                     Vector &Adjoint_m_mv) const {
  RVector integral_M(0.0, numL);
  RVector integral_D(0.0, numL);
  for (const auto& elemIt : eleCache){
    integral_M = 0.0;
    integral_D = 0.0;
    const DGAcousticElement &elem = *elemIt.second;
    const Cell &C = elem.GetCell();
    int Row = Adjoint_m_mv.find_row(C()).Id();

    double a_inv = 1.0 / elem.Area();
    double cc_rho = 0.0;
    double cc_kappa_0 = 0.0;
    RVector cc_M(0.0, numL + 1); // with partial_t
    RVector cc_D(0.0, numL + 1); // without partial_t
    for (int q = 0; q < elem.nQ(); ++q) {
      double w = elem.QWeight(q);
      const VectorField &V_fd = elem.Velocity(q, U_m_fd);
      const VectorField &V_lamb = elem.Velocity(q, LambdaStep);
      cc_rho += w * V_fd * V_lamb;
      double p_fd_0 = elem.Pressure_i(q, U_m_fd, 0);
      double p_lamb_0 = elem.Pressure_i(q, LambdaStep, 0);
      cc_kappa_0 += w * p_fd_0 * p_lamb_0;
      for (int i = 1; i <= numL; ++i) {
        double p_lamb_i = elem.Pressure_i(q, LambdaStep, i);
        cc_M[i - 1] += w * elem.Pressure_i(q, U_m_fd, i) * p_lamb_i;
        cc_D[i - 1] += w * elem.Pressure_i(q, U_m_bwd, i) * p_lamb_i;
      }
    }
    double integral_rho = cc_rho * a_inv * dt;
    double integral_kappa_0 = cc_kappa_0 * a_inv * dt;
    for (int i = 1; i <= numL; ++i) {
      integral_M[i - 1] += cc_M[i - 1] * a_inv * dt;
      integral_D[i - 1] += cc_D[i - 1] * a_inv * dt;
    }
    if (param == "native") {
      updateNative(Row, integral_rho, integral_kappa_0, integral_M, integral_D, Adjoint_m_mv);
    } else if (param == "taup") {
      updateTaup(Row, integral_rho, integral_kappa_0, integral_M, integral_D, Adjoint_m_mv, C);
    } else if (param == "taup_positive") {
      updateTaupPositive(Row, integral_rho, integral_kappa_0, integral_M, integral_D, Adjoint_m_mv,
                         C);
    } else {
      Exit("Invalid parameterization");
    }
  }
}
void FWIAdjointCalculator::PseudoHessianHelper(const Vector &U_m_bwd, const Vector &U_m_fd,
                                               const Vector &LambdaStep,
                                               Vector &Adjoint_m_mv) const {
  RVector integral_M(0.0, numL);
  RVector integral_D(0.0, numL);
  for (const auto& elemIt : eleCache){
    integral_M = 0.0;
    integral_D = 0.0;
    const DGAcousticElement &elem = *elemIt.second;
    const Cell &C = elem.GetCell();
    int Row = Adjoint_m_mv.find_row(C()).Id();
    double cc_rho = 0.0;
    double cc_kappa_0 = 0.0;
    RVector cc_M(0.0, numL + 1); // with partial_t
    RVector cc_D(0.0, numL + 1); // without partial_t
    for (int q = 0; q < elem.nQ(); ++q) {
      double w = elem.QWeight(q);
      const VectorField &V_fd = elem.Velocity(q, U_m_fd);
      const VectorField &V_lamb = elem.Velocity(q, LambdaStep);
      cc_rho += w * V_fd * V_lamb;
      double p_fd_0 = elem.Pressure_i(q, U_m_fd, 0);
      double p_lamb_0 = elem.Pressure_i(q, LambdaStep, 0);
      cc_kappa_0 += w * p_fd_0 * p_lamb_0;
      for (int i = 1; i <= numL; ++i) {
        double p_lamb_i = elem.Pressure_i(q, LambdaStep, i);
        cc_M[i - 1] += w * elem.Pressure_i(q, U_m_fd, i) * p_lamb_i;
        cc_D[i - 1] += w * elem.Pressure_i(q, U_m_bwd, i) * p_lamb_i;
      }
    }
    double integral_rho = cc_rho;
    double integral_kappa_0 = cc_kappa_0;
    for (int i = 1; i <= numL; ++i) {
      integral_M[i - 1] += cc_M[i - 1];
      integral_D[i - 1] += cc_D[i - 1];
    }
    if (param == "native") {
      updateNative(Row, integral_rho, integral_kappa_0, integral_M, integral_D, Adjoint_m_mv);
    } else if (param == "taup") {
      updateTaup(Row, integral_rho, integral_kappa_0, integral_M, integral_D, Adjoint_m_mv, C);
    } else if (param == "taup_positive") {
      updateTaupPositive(Row, integral_rho, integral_kappa_0, integral_M, integral_D, Adjoint_m_mv,
                         C);
    } else {
      Exit("Invalid parameterization");
    }
  }
}

/*
 * Calculates one step of the integral of \Phi''*
 * The integrand First Component (pressure component):
 * v_fd*v_z + v_linearized_fd*v_lambda
 *  - (1/rho²*v_p²)                                 * (p_fd*p_z + p_linearized_fd*p_lambda)
 *  + (2*rhoDir/(rho^3*vp^2) + 2*vpDir/(rho²*vp^3)) * p_fd*p_lambda
 *
 * Second Component (velocity component):
 *  - (2/rho*v_p^3) *                              (p_fd*p_z + p_linearized_fd*p_lambda)
 *  + (2*rhoDir/(rho²*vp^3) + 6*vpDir/(rho*vp^4)) * p_fd*p_lambda
 */
void
FWIAdjointCalculator::calculateL2AdjointOneStepSO(const Vector &u_m_fd, const Vector &u_prime_fd,
                                                  const Vector &LambdaStep, const Vector &zStep, const Vector &MaterialDirection,
                                                  Vector &Adjoint_m_mv, const bool Psi) const {
    /*
     * Get constants /values necessary for the Integral Calculation in the cell (which happens later / in updateTaup):
     *
     * cc_rho1 = p_fd*p_lambda
     * cc_rho2 = p_fd*p_z + p_linearized_fd*p_lambda
     * cc_v    = v_fd*v_z + v_linearized_fd*v_lambda
     */
//    MOUT(zStep.Size())//24544//This comment is sponsered by terminal featuring nano
//    MOUT(LambdaStep.Size())
    //EleCache genauer unter die Lupe nehmen.
//    MOUT(Adjoint_m_mv.Size())//24544
    bool verboseAdj = false;
    std::vector<int> checkOutput{0,15, 100, 200, 511};
    int iteration = 0;
    for (const auto& elemIt : eleCache){// 512 Elements!
        const DGAcousticElement &elem = *elemIt.second;
        const Cell &C = elem.GetCell();
        int Row = Adjoint_m_mv.find_row(C()).Id();

        double a_inv = 1.0 / elem.Area();
        double cc_rho1 = 0.0;
        double cc_rho2 = 0.0;
        double cc_v = 0.0;
//        double psiccRho = 0.0;// = cc_rho1
        double psiccV = 0.0;

        for (int q = 0; q < elem.nQ(); ++q) {
            double w = elem.QWeight(q);
            const VectorField &v_fd = elem.Velocity(q, u_m_fd);
            const VectorField &v_z = elem.Velocity(q, zStep);
            const VectorField &v_linearized_fd = elem.Velocity(q, u_prime_fd);
            const VectorField &v_lambda = elem.Velocity(q, LambdaStep);
            cc_v += w * (v_fd * v_z  +  v_linearized_fd * v_lambda);

            double p_fd_0 = elem.Pressure_i(q, u_m_fd, 0);
            double p_lambda_0 = elem.Pressure_i(q, LambdaStep, 0);
            double p_z = elem.Pressure_i(q, zStep, 0);
            double p_linearized_fd = elem.Pressure_i(q, u_prime_fd, 0);
            cc_rho1 += w * (p_fd_0 * p_lambda_0);
            cc_rho2 += w * (p_fd_0 * p_z  +  p_linearized_fd * p_lambda_0);
            if(Psi) {
                psiccV += w * v_fd * v_lambda;
//                psiccRho += w * p_fd_0 * p_lambda_0;
            }
            for(int j : checkOutput) {
                if (iteration == j && verboseAdj == true) {
                    mout << "Weight q: " << q << endl;
                    mout << "VectorField v_fd, v_z und v_lambda, v_lin_fd  " << v_fd << ",  " << v_z << ",  " << v_lambda << ", " << v_linearized_fd << endl;
//                    mout << "Pressure p_fd, p_p und p_lambda, p_lin_fd  " << p_fd_0 << ",  " << p_z << ",  " << p_lambda_0 << ", " << p_linearized_fd << endl;
                }
            }
        }

        for(int j : checkOutput) {
            if (iteration == j && verboseAdj == true) {
                mout << "We are in loop iteration " << iteration << " at ''Row'' " << Row << endl;
                mout << "Terme für Berechnung: cc_v, cc_rho1, cc_rho2:  " << cc_v << ",  " << cc_rho1 << ",  " << cc_rho2 << endl;
            }
        }


        /*
         * Integralschritt rechnen
         */
        std::vector<double> fac(4);
        MaterialFactorsSO(Row, MaterialDirection, fac);

//        for(int j : checkOutput) {
//            if (iteration == j && verboseAdj == true) {
//                mout << "Vorfaktoren für Berechnung: " << fac[0] << ",  " << fac[1] << ",  " << fac[2] << ", " << fac[3] << endl;
//            }
//        }
        iteration++;
        /* Sign Testing */
//            cc_rho1 = -1 * cc_rho1;
//            cc_rho2 = -1 * cc_rho2;
//            cc_v = -1 * cc_v;

        double grad_rho = 0.0;
        double grad_v_p = 0.0;
        if (param == "taup") {
            if (recon.Rho) {
                grad_rho = a_inv * dt * (cc_v + (fac[0]*cc_rho2) + (fac[2]*cc_rho1));
                if (Psi) {
                    double grad_rho_psi = a_inv * dt * (psiccV + fac[0] * cc_rho1);
                    Adjoint_m_mv(Row, 0) += grad_rho_psi + 0.5 * grad_rho;
                } else {
                    Adjoint_m_mv(Row, 0) += grad_rho;
                }
            }
            //!!!! Ist hier beim Befüllen alles korrekt? Die Ajdungierte hat im TestAdjSO zur Hälfte nullen!?!?!?
            if (recon.Vp) {
                grad_v_p = a_inv * dt * (fac[1] * cc_rho2 + fac[3] * cc_rho1);
                if (Psi) {
                    double grad_v_psi = a_inv * dt * (fac[1] * cc_rho1);
                    Adjoint_m_mv(Row, 1) += grad_v_psi + 0.5 * grad_v_p;
                } else {
                    Adjoint_m_mv(Row, 1) += grad_v_p;
                }
            }
        } else {
            Exit("Unsupported parameterization for Second Order calculation");
        }
    }//Cell Iterating for loop end
//    vout(3) << "FWIAdjointCalculator endlich fertig" << std::endl;
}

/*
 * Supply Material Factors:
 * factor[0] = - 1/rho²*v_p²
 * factor[1] = - 2/rho*v_p^3
 * factor[2] = 2*rhoDir/(rho^3*vp^2) + 2*vpDir/(rho²*vp^3)
 * factor[3] = 2*rhoDir/(rho²*vp^3) + 6*vpDir/(rho*vp^4)
 */
void FWIAdjointCalculator::MaterialFactorsSO(int Row, const Vector &MaterialDirection, std::vector<double> &factors) const {
    double rho = Material(Row, 0);
    double v_p = Material(Row, 1);
    double rhoDir = MaterialDirection(Row, 0);
    double v_pDir = MaterialDirection(Row, 1);

    factors[0] = (-1.0) / (v_p * v_p * rho * rho);
    factors[1] = (-2.0) / (v_p * v_p * v_p * rho);
    factors[2] = 2.0 * rhoDir / (v_p * v_p * rho * rho * rho) + 2.0 * v_pDir / (v_p * v_p * v_p * rho * rho);
    factors[3] = 2.0 * rhoDir / (v_p * v_p * v_p * rho * rho) + 6.0 * v_pDir / (v_p * v_p * v_p * v_p * rho);
}
