#ifndef _FWIHELPERS_H_
#define _FWIHELPERS_H_

#include "m++.hpp"
#include "Parallel.hpp"
#include "Cell.hpp"
#include "Point.hpp"
#include "Vertex.hpp"
#include "RVector.hpp"
#include "Meshes.hpp"
#include "TimeIntegrator.hpp"
#include "DGDiscretization.hpp"
#include "DGAcousticElement.hpp"

#define FWI_TIMESTAMP  "        at " << Date()

#define my_vector vector
#define TIMEFUNCTION(CODE) mout << "Time before "  << #CODE << " " << Date() << endl; Date(); CODE; mout << "Time after "  << #CODE << " " << Date() << endl;

class reconstructionContainer {
public:
    bool Rho = false;
    bool Kappa0 = false;
    bool KappaL = false;
    bool Taus = false;
    bool Vp = false;
    bool Taup = false;
};

using EleCacheType = std::unordered_map<Point, DGAcousticElement *>;
using FaceEleCacheType = std::unordered_map<Point,std::vector<DGAcousticFaceElement>>;

class FaceEleCacheContainer : public FaceEleCacheType {
  EleCacheType NonOverlap;
public:
  FaceEleCacheContainer(const VectorMatrixBase &Vec, int numL) {
    for (cell c = Vec.cells(); c != Vec.cells_end(); ++c) {
      const Cell &C = *c;
      (*this)[C()] = std::vector<DGAcousticFaceElement>();
      for (int f = 0; f< C.Faces(); ++ f) {
        (*this)[C()].emplace_back(Vec, C, f, numL);
      }
    }
    for (cell c = Vec.overlap(); c != Vec.overlap_end(); ++c) {
      const Cell &C = *c;
      for (int f = 0; f< C.Faces(); ++ f) {
        (*this)[C()].emplace_back(Vec, C, f, numL);
      }
    }
  }
};

class EleCacheContainer  {
  EleCacheType ElemCells;
  EleCacheType ElemCellsOverlap;
public:
    EleCacheContainer(const VectorMatrixBase &Vec, int numL) {
        for (cell c = Vec.cells(); c != Vec.cells_end(); ++c) {
            const Cell &C = *c;
          ElemCells[C()] = new DGAcousticElement(Vec, C, numL);
        }
        for (cell c = Vec.overlap(); c != Vec.overlap_end(); ++c) {
            const Cell &C = *c;
          ElemCellsOverlap[C()] = new DGAcousticElement(Vec, C, numL);
        }
    }
    const EleCacheType& Get() {
      return ElemCells;
    }
  const EleCacheType& GetOverlap() const {
    return ElemCellsOverlap;
  }
  ~EleCacheContainer(){
      for (auto elem : ElemCells){
        delete elem.second;
      }
      for (auto elem: ElemCellsOverlap){
        delete elem.second;
      }
    }
};

class FWIDiscretizationStack {
public:
    LevelPair lev{-1, -1};
    std::shared_ptr<Meshes> meshes = nullptr;
    std::unique_ptr<IDiscretization> mD = nullptr;
    std::unique_ptr<DGDiscretization> wD = nullptr;
    std::shared_ptr<Vector> mV = nullptr;
    std::shared_ptr<Vector> wV = nullptr;

    FWIDiscretizationStack() = default;

    FWIDiscretizationStack(const std::shared_ptr<Meshes> &meshes, int paramComps, int stateComps, int deg,
                           LevelPair Lev)
            : lev(Lev),
              meshes(meshes), mD(std::make_unique<DGDiscretization>(*meshes, 0, paramComps)),
              wD(std::make_unique<DGDiscretization>(*meshes, deg, stateComps)),
              mV(std::make_shared<Vector>(0.0, *mD, lev)),
              wV(std::make_shared<Vector>(0.0, *wD, lev)) {}

    const Mesh &GetMesh() {
        return (*meshes)[lev];
    }

    ~FWIDiscretizationStack() = default;
};
struct RectangularConfig {
  double LeftBnd;
  double RightBnd;
  double BottomBnd;
  double TopBnd;
  int dimX;
  int dimY;
  int bndTop;
  int bndBottom;
  int bndLeft;
  int bndRight;
  std::string name;

  std::vector<int> boundaries() const { return {bndLeft, bndRight, bndBottom, bndTop}; }
};

std::shared_ptr<RectangularGeometry2D> getSharedGeom(const RectangularConfig &conf);
std::shared_ptr<Meshes> getFWIMesh();
class TimeConf {
    void readConfig() {
        Config::Get("t0", t0);
        Config::Get("dt", dt);
        Config::Get("T", T);
        Config::Get("rkorder", rkorder);
        TS = new TimeSeries(t0, T, dt);
        timeSteps = TS->Steps();
    }

    TimeSeries *TS = nullptr;
public:
    double t0 = 0;
    double dt = -1;
    double T = -1;
    int timeSteps = -1;
    int rkorder = 0;

    TimeConf() {
        readConfig();
    }

    TimeConf(double t0_, double dt_, double T_, int rkorder_) : t0(t0_), dt(dt_), T(T_), rkorder(rkorder_) {
        TS = new TimeSeries(t0, T, dt);
    }

    TimeSeries *GetTS() const {
        return new TimeSeries(*TS);
    }
};
class SourceConf {
public:
    std::string SignalName = "RickerDerivative";
    bool FrequencyFiltering = false;
    string FilteringType = "LowPass";
    double SignalFreq = 25.0;
    int FilteringOrder = 1;
//    double LowPassFilterFreq = -1;
    vector<vector<double>> FrequencyLevels;
    int CurrentLevel = 0;
    int MaterialSmoothPasses = 0;
    double SignalDelay = 0;
    bool SaveObservedSeismogram = false;
    bool SaveFilteredSeismogram = false;
    double SignalCutoff = 0;
    bool SrcInversion = false;
    double SrcInversionKappa = 0.0;
    double SrcInversionEps = 0.0;
    SourceConf();
     vector<double> getFreq() const {
      return FrequencyLevels[CurrentLevel];
    }
};

class WPConf {
public:
    int numL = 1;
    int dim = 2;
    std::string s_prob;
    int level = 0;
    int RealDataLevel;
    int InitialIterateLevel;
    TimeConf TConf;
    SourceConf SConf;

    WPConf();
};

auto GetFWITI(int rkorder) -> std::unique_ptr<TimeIntegrator>;

inline double squared(double x) { return x * x; }

inline std::string nr_str(const char *str, int number) {
    if (number >= 0) {
        std::stringstream ss;
        ss << str << "_";
        ss << number;
        return ss.str();
    }
    return str;
}
inline std::string nr_str(const char* str, int number, int number2) {
    if (number >= 0) {
        std::stringstream ss;
        ss << str << "_";
        ss << number;
        if (number2 >=0){
            ss << "_" << number2;
        }
        return ss.str();
    }
    return str;
}


class PointIntHash  {
public:
    size_t operator () (const std::pair<Point,int>& p) const {
        return p.first.Hash() + 324323.43*p.second;
    }
};


// TODO: will not work for 1d and 2d....
inline std::istream& operator >> (std::istream& in, Point& P) {
    in >> P[0];
    in >> P[1];
    //    in >> P[2];
    return in;
}

inline std::istream& operator >> (std::istream& in, VectorField& v) {
    in >> v[0];
    in >> v[1];
    in >> v[2];
    return in;
}




// Copyright 2000 softSurfer, 2012 Dan Sunday
// This code may be freely used and modified for any purpose
// providing that this copyright notice is included with it.
// SoftSurfer makes no warranty for this code, and cannot be held
// liable for any real or imagined damage resulting from its use.
// Users of this code must verify correctness for their application.

// checks whether P2 is left of the line through P0,P1
inline Scalar isLeft( const Point& P0, const Point& P1, const Point& P2 )
{
    return ( (P1[0] - P0[0]) * (P2[1] - P0[1])
            - (P2[0] -  P0[0]) * (P1[1] - P0[1]) );
}
//===================================================================


// cn_PnPoly(): crossing number test for a point in a polygon
//      Return:  0 = outside, 1 = inside
// This code is patterned after [Franklin, 2000]
int cn_PnPoly(const Point &P, const cell &c);
//===================================================================


// wn_PnPoly(): winding number test for a point in a polygon
//      Return:  wn = the winding number (=0 only when P is outside)

int wn_PnPoly(const Point &P, const Cell &C);

int wn_PnPoly(const Point &P, const cell &c);

bool PointInCellDGTS_old(const Point &z, const Cell &C);

bool PointInCellDGTS(const Point &z, const Cell &C);

bool PointInCellDGTS(const Point &z, const cell &c);

inline double rickerDerivativeDelay(double f, double t, double delay, double cutoff) {
    if (t < delay || t > cutoff) return 0;
    double tau = Pi * f * ((t - delay) - 1.5 / f);
    return ((t - delay) - 1.5 / f) * exp(-tau * tau);
}

inline double rickerDelay(double f, double t, double delay, double cutoff) {
    if (t < delay || t > cutoff) return 0;
    double tau = Pi * f * ((t - delay) - 1.5 / f);
    return (1.0 - 2.0 * tau * tau) * exp(-tau * tau);
}

inline double rickerDerivative(double f, double t) {
    if (t == 0) return 0;
    double tau = Pi * f * (t - 1.5 / f);
    return (t - 1.5 / f) * exp(-tau * tau);
}

inline double ricker(double f, double t) {
    if (t == 0) return 0;
    double tau = Pi * f * (t - 1.5 / f);
    return (1.0 - 2.0 * tau * tau) * exp(-tau * tau);
}


void fillQuadrature(RVector &c, RVector &b, int order);

#endif
