#include "FWINonlinearSolvers.hpp"
#include "RegularisationStrategy.hpp"
#include "ForwardModels/FWIForwardModel.hpp"
#include "Spectrum.hpp"
#include "ForwardModels/FWIForwardModelDGTS.hpp"
#include "FWIDataContainerRange.hpp"
#include "FWIDataContainerDomain.hpp"
#include "FWINonlinearSolverSO.hpp"
#include <deque>
#include <memory>
#include <valarray>
#include <iostream>

void FWINonlinearSolver::backtracking(const FWISubproblem &dataObs,
                                      std::vector<double> &residualNorms,
                                      const FWIDataContainerDomain &iterateMUpdate,
                                      FWIDataContainerDomain &iterateNonlinear,
                                      FWISubproblem &nonlinearEvaluation,
                                      FWISubproblem &defect,
                                      std::vector<int> subProb) {
    FWIDataContainerDomain iterateNonlinearTMP(iterateNonlinear);
    iterateNonlinearTMP += iterateMUpdate;
    FWD.NonlinearEvaluation(iterateNonlinearTMP, nonlinearEvaluation, subProb);
    FWD.calculateDefect(dataObs, nonlinearEvaluation, defect, subProb);
    double r_norm = sqrt(FWD.scalarProductL2(defect, defect));
    if (r_norm < residualNorms[iter]) {
        mout << "====================================" << endl;
        mout << "No Backtracking needed! New Residual: " << r_norm << " vs "
             << residualNorms[iter] << endl;
        iterateNonlinear = iterateNonlinearTMP;
        mout << OUT(iter) << endl;
        if (iter == 0) residualNorms[1] = r_norm;
        mout << OUT(residualNorms[iter]) << endl;
        return;
    }
    mout << "iterate Norm before backtracking starts: "
         << FWD.scalarProductL2(iterateNonlinear, iterateNonlinear) << endl;
    mout << "update before backtracking : "
         << FWD.scalarProductL2(iterateMUpdate, iterateMUpdate) << endl;
    double t = 1;
    int back = 1;
    while (r_norm >= residualNorms[iter]) {
        mout << "====================================" << endl;
        mout << "  " + Name() + " d(" << iter << "): Backtracking(" << back - 1
             << "): Starting ( "
             << r_norm << " vs " << residualNorms[iter] << " )";
        iterateNonlinearTMP = iterateNonlinear;
        FWIDataContainerDomain iterateMUpdateTMP(iterateMUpdate);
        t = pow(generalOpt.backtrackingFactor, back);
        iterateMUpdateTMP *= t;
        iterateNonlinearTMP += iterateMUpdateTMP;
        FWD.NonlinearEvaluation(iterateNonlinearTMP, nonlinearEvaluation, subProb);
        FWD.calculateDefect(dataObs, nonlinearEvaluation, defect, subProb);
        r_norm = sqrt(FWD.scalarProductL2(defect, defect));
        mout << "  " + Name() + " d(" << iter << "): Backtracking finished ( "
             << r_norm << " vs " << residualNorms[iter] << " )" << endl;
        ++back;

        mout << " the backtracking factor is: " << t << endl;
    }
    iterateNonlinear = iterateNonlinearTMP;
    if (iter == 0) residualNorms[1] = r_norm;
}

void FWINonlinearSolver::backtrack(const FWISubproblem &dataObs,
                                   std::vector<double> &residualNorms,
                                   const FWIDataContainerDomain &iterateMUpdate,
                                   FWIDataContainerDomain &iterateNonlinear,
                                   FWISubproblem &nonlinearEvaluation,
                                   FWISubproblem &defect,
                                   std::vector<int> subProb) {
    backtracking(dataObs,
                 residualNorms,
                 iterateMUpdate,
                 iterateNonlinear,
                 nonlinearEvaluation,
                 defect,
                 subProb);
}


RMatrix assembleSKPMatrix(FWIForwardModel &FWD,
                          vector<FWIDataContainerRange> &linearizedCalcs) {
    RMatrix matrixEntries(static_cast<int>(linearizedCalcs.size()));
    for (int i = 0; i < static_cast<int>(linearizedCalcs.size()); ++i) {
        for (int j = i; j < static_cast<int>(linearizedCalcs.size()); ++j) {
            matrixEntries[i][j] =
                    FWD.scalarProductL2(linearizedCalcs[i], linearizedCalcs[j]);
            matrixEntries[j][i] = matrixEntries[i][j];
        }
    }
    return matrixEntries;
}

class FWIREGINNBASE : public FWINonlinearSolver {
public:
protected:
    int update_strategy;
    bool setSKP = false;
    std::vector<vector<int>> subproblems;
    ReginnOptions opt;

public:
    FWIREGINNBASE(FWIForwardModel &FWD, std::string name, ReginnOptions _opt) :
        FWINonlinearSolver(name.c_str(), FWD), opt(_opt) {
        verbose = 5;
        iter = 0;
    }

    ~FWIREGINNBASE() {}

    void initializeSubproblems(const std::vector<FWIDataContainerRange> &data_obs,
                               std::vector<FWISubproblem> &nonlinearEvaluations) {
        if (true) {
            bool initialAAO = true;
            Config::Get("AAORecalculateWFEveryOuter", initialAAO);
            subproblems.resize(data_obs.size());
            nonlinearEvaluations.resize(data_obs.size());
            for (int i = 0; i < data_obs.size(); ++i) {
                subproblems[i].resize(1);
                nonlinearEvaluations[i].emplace_back(FWD.getReconstructionType(),
                                                                        FWD.GetWaveVec(),
                                                                        FWD.TConf.timeSteps);
                if (initialAAO) nonlinearEvaluations[i].back().resetWaveData();
                subproblems[i][0] = i;
            }
        } else {
            Exit("leave now and nevercome back!")
            {
                subproblems.resize(1);
                nonlinearEvaluations.resize(1);
                subproblems[0].resize(data_obs.size());
                for (int i = 0; i < data_obs.size(); ++i) {
                    subproblems[0].push_back(i);
                    nonlinearEvaluations[0].push_back(FWIDataContainerRange(FWD.getReconstructionType(),
                                                                            FWD.GetWaveVec(),
                                                                            FWD.TConf.timeSteps));
                }
            }
        }
    }

    void initializeIterateList(const FWIDataContainerDomain &iterateNonlinear,
                               std::vector<FWIDataContainerDomain> &iterateList) {
        if (FWD.getReconstructionType() == "AllAtOnce") {
            bool initialAAO = true;
            Config::Get("AAORecalculateWFEveryOuter", initialAAO);
            if (initialAAO) {
                return;
            }
            FWISeismogram stmp;
            for (int i = 0; i < FWD.numberOfShots(); ++i) {
                iterateList.push_back(iterateNonlinear);
                FWD.Parameter2SeismogramMap(iterateNonlinear.getMat(), stmp, i, &iterateList[i].getVectorsWrite());
            }
        }
    }

    void getIterateFromlist(const std::vector<FWIDataContainerDomain> &iterateList,
                                 const std::vector<int> &prob,
                                 FWIDataContainerDomain &iterateNonlinear) {
        if (FWD.getReconstructionType() == "AllAtOnce") {
            if (prob.size() > 1) {Exit("just for kaczmarcz implemented!")}
            bool initialAAO = true;
            Config::Get("AAORecalculateWFEveryOuter", initialAAO);
            if (initialAAO) {
                FWISeismogram stmp;
                FWD.Parameter2SeismogramMap(iterateNonlinear.getMat(), stmp, prob[0],
                                            &iterateNonlinear.getVectorsWrite());
            }else {
                Exit("Defect is set to zero in AAO, this assumes AAORecalculateWFEveryOuter = 1")
                iterateNonlinear.getVectorsWrite() = iterateList[prob[0]].getVectors();
            }

        }
    }

    void finishIterate(vector<FWIDataContainerDomain> &iterateList, const std::vector<int> &prob,
                       FWIDataContainerDomain &iterateNonlinear, FWISubproblem &nonlinearEvaluations) {
        if (FWD.getReconstructionType() == "AllAtOnce") {
            bool initialAAO = true;
            Config::Get("AAORecalculateWFEveryOuter", initialAAO);
            if (initialAAO) {
                nonlinearEvaluations[0].resetWaveData();
                return;
            } else {
                if (prob.size() > 1) {Exit("just for kaczmarcz implemented!")}
                iterateList[prob[0]].getVectorsWrite() = iterateNonlinear.getVectors();
            }
        }
    }

    virtual double innerLoopOneStep(const FWIDataContainerDomain &iterateNonlinear,
                                    const FWISubproblem &nonlinearEvaluation,
                                    FWISubproblem &defectLinearized,
                                    FWIDataContainerDomain &iterateMUpdate,
                                    int innerIter, size_t shot) = 0;

    void operator()(const IAssemble &, Vector &) override {Exit("not implemented"); }

    NonlinearOutput operator()(const std::vector<FWIDataContainerRange> &dataObs,
                               FWIDataContainerDomain &iterateNonlinear, std::string nameSuff) override {
        mout << "Memory Usage at start" << PPM->TotalMemoryUsage() << endl;
        Start = Date();
        std::vector<FWISubproblem> nonlinearEvaluations;
        std::vector<FWIDataContainerDomain> iterateList;
        FWISubproblem defect;
        int writeSeismoToFileEvery = INT_MAX;
        Config::Get("WriteSeismoToFileEvery", writeSeismoToFileEvery);
//        FWISubproblem defectLinearized;
        initializeIterateList(iterateNonlinear, iterateList);
        initializeSubproblems(dataObs, nonlinearEvaluations);
        mout << "Initialize Subproblems " << PPM->TotalMemoryUsage() << endl;
        size_t num_probs = subproblems.size();
        FWIDataContainerDomain iterateMUpdate(FWD.getReconstructionType(), FWD.GetMatVec(),
                                              FWD.GetWaveVec(), FWD.TConf.timeSteps);
        std::vector<RegularisationStrategyWinkler> regs;
        regs.resize(subproblems.size(),
                    RegularisationStrategyWinkler(opt.gamma, opt.thetaMax, opt.maxInnerLoopSteps));
        d_0 = 0.0;
        std::cout << std::setprecision(15);
        for (; iter < max_iter; ++iter) {
            for (size_t prob = 0; prob < num_probs; ++prob) {
                mout << "before get iterate " << PPM->TotalMemoryUsage() << endl;

                getIterateFromlist(iterateList, subproblems[prob], iterateNonlinear);
                mout << "after get iterate " << PPM->TotalMemoryUsage() << endl;

              FWISubproblem &nonlinearEvaluation = nonlinearEvaluations[prob];
              vout(2) << Name() + " problem " << prob << ": d(" << iter << ")" << endl;

              FWD.NonlinearEvaluation(iterateNonlinear,
                                      nonlinearEvaluation,
                                      subproblems[prob]);
                mout << "after NonlinearEvaluation " << PPM->TotalMemoryUsage() << endl;
                if( iter % writeSeismoToFileEvery  == 0) {
                  FWISeismogram seis(nonlinearEvaluation[0].getSeismogram());
                  seis.collect();
                  seis.writeToFile(Config::GetDataPath()+ "FWI/" + "Seis" + std::to_string(prob) + "_" +
                                                                     std::to_string(iter));
                }
//              FWD.plot(nonlinearEvaluation[0].getVectorsWrite(), nameSuff + "nonlinearEvaluation_iter_"+ to_string(iter));

              FWD.calculateDefect(dataObs,
                                  nonlinearEvaluation,
                                  defect,
                                  subproblems[prob]);
                mout << "after Defect " << PPM->TotalMemoryUsage() << endl;

//              if (iter % 1 == 0 && (&defect[0].getVectorsWrite() != nullptr)) {
//                FWD.plot(defect[0].getVectorsWrite(), nameSuff + "defect_iter_"+ to_string(iter));
//              }
              if (defect.size() != nonlinearEvaluation.size()) {
                Exit(" DataContainer sizes do not match")
              }
              if (iter != 1)
                  regs[prob].addResNorm(sqrt(FWD.scalarProductL2(defect, defect)));
                regs[prob].calcTheta();
                vout(1) << "*************************************" << endl
                        << "  " + Name() + " Subproblem " << prob << ": d(" << iter
                        << "): Starting inner loop "
                        << "( ||r||_2 = " << regs[prob].getResNorm(iter)
                        << ", theta*||r||_2 = " << regs[prob].calcInnerStopp()
                        << ", theta = " << regs[prob].getTheta(iter)
                        << ", m_max = " << regs[prob].calcMaxInner() << " )" << endl;
//                defectLinearized = defect;
//              if (iter % 1 == 0 && (&defect[0].getVectorsWrite() != nullptr)) {
//                FWD.plot(defect[0].getVectorsWrite(), nameSuff + "defect_iter_"+ to_string(iter));
//              }
                int innerIter = 0;
                iterateMUpdate.Clear();
                FWD.GetInitialValueForInner(iterateNonlinear, nonlinearEvaluation, defect,
                                            iterateMUpdate, subproblems[prob]);
                double residualNormTilde = infty;
//                mout << OUT(FWD.scalarProductL2(iterateNonlinear, iterateNonlinear))
//                     << endl;
                while ((innerIter
                        < min(static_cast<unsigned int>(opt.maxInnerLoopSteps), regs[prob].calcMaxInner())
                        && residualNormTilde > regs[prob].calcInnerStopp()) || innerIter < min_iter) {
                    mout << "Before innerstep " << PPM->TotalMemoryUsage() << endl;
                    MOUT(FWD.scalarProductL2(iterateMUpdate, iterateMUpdate));
                    residualNormTilde = innerLoopOneStep(iterateNonlinear,
                                                         nonlinearEvaluation,
                                                         defect,
                                                         iterateMUpdate,
                                                         innerIter,
                                                         prob);
                    MOUT(FWD.scalarProductL2(iterateMUpdate, iterateMUpdate));
                    mout << "after innerstep " << PPM->TotalMemoryUsage() << endl;

                    vout(2) << "  " + Name() + " Subproblem: " << prob << ": d(" << iter
                            << "), inner(" << innerIter << "): "
                            << "||r||_2 = " << residualNormTilde << endl;
                    innerIter++;
                }
                if (iter == 0) regs[prob].addResNorm(residualNormTilde);
                regs[prob].addInnerIter(innerIter);

                vout(1) << "  " + Name() + " Subproblem: " << prob << ": d(" << iter
                        << "): Finished inner loop "
                        << "( " << regs[prob].getInnerIter(iter) << " steps )" << endl
                        << "*************************************" << endl;
                if(FWD.UseGardener()){FWD.Gardener(iterateNonlinear,iterateMUpdate);}
                else {
                  iterateNonlinear += iterateMUpdate;
                }
                if (generalOpt.backtrackingActivated) {
                    if (FWD.getReconstructionType() == "AllAtOnce") {
                        Exit("Defect is set to zero in AAO")
                    }
                    backtrack(dataObs,
                              regs[prob].getResNormVector(iter),
                              iterateMUpdate,
                              iterateNonlinear,
                              nonlinearEvaluation,
                              defect,
                              subproblems[prob]);
                }
                int plotIter = 1;
                Config::Get("PlotFactorIterates", plotIter);
                int plotShot = -1;
                Config::Get("PlotShot", plotShot);

                if (iter % plotIter == 0 && (plotShot < 0 || prob == plotShot)) {
                    FWD.plot(iterateNonlinear,
                             "_fwi_" + Name() + nameSuff + "_iterate_m_allshots_" + to_string(prob) + "_"
                             + to_string(iter));
                }
                mout << "Before finishIterate " << PPM->TotalMemoryUsage() << endl;
                finishIterate(iterateList, subproblems[prob], iterateNonlinear, nonlinearEvaluation);
                mout << "After finishIterate " << PPM->TotalMemoryUsage() << endl;
                if (iter + 1 == max_iter) {
                    mout << "Calculating Final Residual: " << endl;
                    FWD.NonlinearEvaluation(iterateNonlinear,
                                            nonlinearEvaluation,
                                            subproblems[prob]);

                    FWD.calculateDefect(dataObs,
                                        nonlinearEvaluation,
                                        defect,
                                        subproblems[prob]);
                    regs[prob].addResNorm(sqrt(FWD.scalarProductL2(defect, defect)));
                }
            }

        }


        vout(1) << " " + Name() + ": d(" << iter << ")= " << d << "\n";
        NonlinearOutput output;
        for (size_t prob = 0; prob < num_probs; ++prob) {
            output.push_back(regs[prob].getResNormVector(0));
        }
        return output;
//        vout(0) << " " << IterationEnd();
    }
};

class FWICGREGINN_ : public FWIREGINNBASE {
    FWIDataContainerDomain iterateUpdatePN;
    FWIDataContainerDomain iterateUpdatePNMinus1;
    double beta = 0.0;
public:
    virtual auto initializePreconditioner(const FWIDataContainerDomain &nonlinearPoint,
                                          const FWISubproblem &nonlinearEvaluation,
                                          const FWIDataContainerDomain &mNegativeUpdate,
                                          FWISubproblem &defectLinearized,
                                          int innerIter) -> void = 0;

    virtual void applyPreconditioner(const FWIDataContainerDomain &,
                                     FWIDataContainerDomain &,
                                     FWISubproblem &,
                                     int) = 0;

    FWICGREGINN_() = delete;

    FWICGREGINN_(FWIForwardModel &FWD, std::string name, ReginnOptions opt)
            : FWIREGINNBASE(FWD, name, opt),
              iterateUpdatePN(FWD.getReconstructionType(), FWD.GetMatVec(),
                              FWD.GetWaveVec(), FWD.TConf.timeSteps),
              iterateUpdatePNMinus1(FWD.getReconstructionType(), FWD.GetMatVec(),
                                    FWD.GetWaveVec(), FWD.TConf.timeSteps) {}

    double innerLoopOneStep(const FWIDataContainerDomain &iterateNonlinear,
                            const FWISubproblem &nonlinearEvaluation,
                            FWISubproblem &defectLinearized,
                            FWIDataContainerDomain &iterateMUpdate,
                            int innerIter, size_t prob) override {
        FWISubproblem iterateLinearizedTMP(defectLinearized);
        if (innerIter == 0) beta = 0.0;
        vout(2) << "  " + Name() + " Subproblem " << prob << ": d(" << iter << "), inner("
                << innerIter << "): "
                << "Solving backpropagation problem" << endl;
        iterateUpdatePN.Clear();
        MOUT(FWD.scalarProductL2(iterateUpdatePN, iterateUpdatePN));
        FWD.AdjointLinearized(iterateNonlinear, defectLinearized, iterateUpdatePN);
        MOUT(FWD.scalarProductL2(iterateUpdatePN, iterateUpdatePN));
        initializePreconditioner(iterateNonlinear, nonlinearEvaluation, iterateUpdatePN,
                                 defectLinearized, innerIter);
        FWIDataContainerDomain iterateUpdatePNPC(iterateUpdatePN);
      applyPreconditioner(iterateNonlinear,iterateUpdatePNPC,defectLinearized,innerIter);
        MOUT(FWD.scalarProductL2(iterateUpdatePN, iterateUpdatePN));

        double skpPCd = FWD.scalarProductL2(iterateUpdatePNPC, iterateUpdatePN);
        iterateUpdatePN = iterateUpdatePNPC;
        MOUT(FWD.scalarProductL2(iterateUpdatePN, iterateUpdatePN));
        iterateUpdatePNMinus1 *= (beta * skpPCd);
        beta = 1.0 / skpPCd;
        iterateUpdatePN += iterateUpdatePNMinus1;
        vout(2) << "  " + Name() << " Subproblem" << prob << ": d(" << iter << "), inner("
                << innerIter << "):  "
                << "Solving linearized forward problem" << endl;
        FWD.Linearized(iterateNonlinear, iterateUpdatePN, iterateLinearizedTMP);
        double linearizedNorm =
                FWD.scalarProductL2(iterateLinearizedTMP, iterateLinearizedTMP);
        double alpha = skpPCd / linearizedNorm;
        MOUT(alpha);
        iterateUpdatePN *= alpha;
        iterateMUpdate += iterateUpdatePN;
        iterateUpdatePN *= 1.0 / alpha;
        iterateLinearizedTMP *= alpha;
        defectLinearized -= iterateLinearizedTMP;
        iterateUpdatePNMinus1 = iterateUpdatePN;
//        Exit("CIAO");
//MOUT(sqrt(FWD.scalarProductL2(iterateMUpdate, iterateMUpdate)));
        return sqrt(FWD.scalarProductL2(defectLinearized, defectLinearized));
    }
};
class FWICGREGINNCanonic : public FWICGREGINN_ {
public:
    auto initializePreconditioner(const FWIDataContainerDomain &nonlinearPoint,
                                  const FWISubproblem &nonlinearEvaluation,
                                  const FWIDataContainerDomain &mNegativeUpdate,
                                  FWISubproblem &defectLinearized,
                                  int innerIter) -> void override {
        if (FWD.GetPrecon()) {
            FWD.initializePrecon(nonlinearPoint, mNegativeUpdate, nonlinearEvaluation);
        }
    }

    auto applyPreconditioner(const FWIDataContainerDomain &nonlinearPoint,
                             FWIDataContainerDomain &mNegativeUpdate,
                             FWISubproblem &defectLinearized,
                             int innerIter) -> void override {
        if (FWD.GetPrecon()) {
          FWD.applyPrecon(nonlinearPoint, mNegativeUpdate);
        }
    }

    FWICGREGINNCanonic() = delete;

    FWICGREGINNCanonic(FWIForwardModel &FWD, std::string name, ReginnOptions opt)
            : FWICGREGINN_(FWD, name, opt) {}
};

class FWICGREGINNMulti : public FWICGREGINN_ {
    std::valarray<size_t> invMap = std::valarray<size_t>(1, 1);
    bool freezeSKP = true;
    bool normalizeByBackgroundBeforeSKP = false;
    bool normalizeByCurrentIterateBeforeSKP = false;
    bool normalizeByCurrentIterate = false;
    RMatrix SKP;
public:
    FWICGREGINNMulti() = delete;
    FWICGREGINNMulti(FWIForwardModel &FWD, std::string name, ReginnOptions opt)
        : FWICGREGINN_(FWD, name, opt) {
        auto reconlist = FWD.getReconstructionlist();
        invMap.resize(reconlist.size());
        size_t count = 0;
        for (auto it = reconlist.begin(); it != reconlist.end(); ++it) {
            invMap[count] = *it;
            count += 1;
        }
        Config::Get("FWI_FreezeSKP_CGREGINNMulti", freezeSKP);
        Config::Get("FWI_normalizeByBackgroundBeforeSKP_CGREGINNMulti",
                    normalizeByBackgroundBeforeSKP);
        Config::Get("FWI_normalizeByCurrentIterateBeforeSKP_CGREGINNMulti",
                    normalizeByCurrentIterateBeforeSKP);
        Config::Get("FWI_normalizeByCurrentIterate_CGREGINNMulti",
                    normalizeByCurrentIterate);
    }

    auto initializePreconditioner(const FWIDataContainerDomain &nonlinearPoint,
                                  const FWISubproblem &nonlinearEvaluation,
                                  const FWIDataContainerDomain &mNegativeUpdate,
                                  FWISubproblem &defectLinearized,
                                  int innerIter) -> void override {
        if (FWD.getReconstructionType() == "AllAtOnce") {
            if (FWD.GetPrecon()) {
                FWD.initializePrecon(nonlinearPoint.getMat(), mNegativeUpdate.getMat(), nonlinearPoint.getVectors());
            }
        }
        if (innerIter != 0 && freezeSKP) return;
        mout
                << "################ Calculating Small Preconditioning Matrix  ################"
                << endl;
        SKP = RMatrix(RVector(Scalar(1.0), FWD.ParameterComponents()));
        std::vector<FWISubproblem> linearizedCalcs;
        std::unique_ptr<Vector> mNegativeUpdateNormalized = nullptr;
        if (normalizeByCurrentIterateBeforeSKP) {
            mNegativeUpdateNormalized
                    .reset(new Vector(mNegativeUpdate.getMat()));
            Vector mInv(nonlinearPoint.getMat());
            Exit("see below");
//            mInv.inv(Eps);
//        --------------------------------------------------------------
//            Implement this into Vector!
//        --------------------------------------------------------------
//            inv(void *data, double cutoff) const {
//                Vector &v = *_toVector(data);
//                int warnings = 0;
//                for (int i = 0; i < v.size(); ++i) {
//                    if (abs(v[i]) < cutoff) {
//                        v[i] = 0;
//                        warnings++;
//                    }
//                    v[i] = 1. / v[i];
//                }
//                if (warnings > 0) {
//                    mout << "cutoff ratio when inverting: "
//                         << v.size() / static_cast<double>(PPM->Sum(warnings)) << endl;
//                }
//            }
//            (*mNegativeUpdateNormalized) *= mInv;
        }
        for (size_t i = 0; i < invMap.size(); ++i) {
            FWIDataContainerDomain mTemp(mNegativeUpdate);
            if (normalizeByCurrentIterateBeforeSKP) {
                mTemp.getMatWrite() = *mNegativeUpdateNormalized;
            }
            for (size_t j = 0; j < FWD.ParameterComponents(); ++j) {
                if (invMap[i] != j) mTemp.setCompToZero(j);
            }
            linearizedCalcs.push_back(FWISubproblem(defectLinearized));
            FWD.Linearized(nonlinearPoint, mTemp, linearizedCalcs[i]);
        }
        for (size_t i = 0; i < invMap.size(); ++i) {
            size_t pos1 = invMap[i];
            for (size_t j = 0; j < i + 1; ++j) {
                size_t pos2 = invMap[j];
                SKP[pos1][pos2] = FWD.scalarProductL2(linearizedCalcs[i],
                                                      linearizedCalcs[j]);
                if (pos2 != pos1) SKP[pos2][pos1] = SKP[pos1][pos2];
            }
        }
        if (normalizeByBackgroundBeforeSKP) {
            RMatrix SKP2 (RVector(Scalar(1.0),FWD.ParameterComponents()));
            SKP2[0][0] = 2000.0;
            SKP2[1][1] = 3500.0;
            SKP2[2][2] = 0.0767;
            SKP = SKP * SKP2;
            SKP = SKP2 * SKP;
        }
        SKP.Invert();
    }

    auto applyPreconditioner(const FWIDataContainerDomain &nonlinearPoint,
                             FWIDataContainerDomain &mNegativeUpdate,
                             FWISubproblem &defectLinearized,
                             int innerIter) -> void override {

      if (normalizeByCurrentIterate) {
            Vector matInv(nonlinearPoint.getMat());
//            matInv *= matInv;
//            matInv.sqrt();
//            matInv.inv(Eps);
            //            mInv.inv(Eps);
//        --------------------------------------------------------------
//            Implement this into Vector!
//        --------------------------------------------------------------
//            void FWIMppVectorHandlerDGTS::sqrt(void *data) const {
//    Vector &v = *_toVector(data);
//    for (int i = 0; i < v.size(); ++i)
//        v[i] = ::sqrt(v[i]);
//}
//            mNegativeUpdatePC.getMatWrite() *= matInv;
            Exit("CIAO")
        }
      mNegativeUpdate *= SKP;
    }
};

class FWICGREGINNWaveFieldPrecon : public FWICGREGINN_ {
    std::valarray<size_t> invMap = std::valarray<size_t>(1, 1);
    bool freezeSKP = true;
    RMatrix SKP;
public:
    FWICGREGINNWaveFieldPrecon() = delete;

    FWICGREGINNWaveFieldPrecon(FWIForwardModel &FWD, std::string name, ReginnOptions opt)
        : FWICGREGINN_(FWD, name, opt) {
    }

    auto initializePreconditioner(const FWIDataContainerDomain &nonlinearPoint,
                                  const FWISubproblem &nonlinearEvaluation,
                                  const FWIDataContainerDomain &mNegativeUpdate,
                                  FWISubproblem &defectLinearized,
                                  int innerIter) -> void override {
        FWD.initializePrecon(nonlinearPoint, mNegativeUpdate, nonlinearEvaluation);
    }

    auto applyPreconditioner(const FWIDataContainerDomain &nonlinearPoint,
                             FWIDataContainerDomain &mNegativeUpdate,
                             FWISubproblem &defectLinearized,
                             int innerIter) -> void override {
      FWD.applyPrecon(nonlinearPoint, mNegativeUpdate);
    }
};

class FWISDREGINN : public FWIREGINNBASE {
protected:
  FWIDataContainerDomain iterateUpdatePN;
public:
    FWISDREGINN(FWIForwardModel &FWD, std::string name, ReginnOptions opt)
            : FWIREGINNBASE(FWD, name, opt),
              iterateUpdatePN(FWD.getReconstructionType(), FWD.GetMatVec(),
                              FWD.GetWaveVec(), FWD.TConf.timeSteps) {}

    virtual void multiplyWithSteplength(const FWIDataContainerDomain &,
                                        FWIDataContainerDomain &,
                                        FWIDataContainerDomain &,
                                        FWISubproblem &) = 0;

    double innerLoopOneStep(const FWIDataContainerDomain &iterateNonlinear,
                            const FWISubproblem &nonlinearEvaluation,
                            FWISubproblem &defectLinearized,
                            FWIDataContainerDomain &iterateMUpdate,
                            int innerIter, size_t shot) override {
        vout(2) << "  " + Name() + " shot" << shot << ": d(" << iter << "), inner("
                << innerIter << "): "
                << "Solving backpropagation problem" << endl;
        iterateUpdatePN = 0.0;
        std::cout << std::setprecision(15);
        mout << "RAM before adjoint " << PPM->TotalMemoryUsage() << endl;

        FWD.AdjointLinearized(iterateNonlinear, defectLinearized, iterateUpdatePN);
        mout << "RAM after adjoint " << PPM->TotalMemoryUsage() << endl;
        mout << "RAM before multiply " << PPM->TotalMemoryUsage() << endl;

        multiplyWithSteplength(iterateNonlinear,
                               iterateUpdatePN,
                               iterateMUpdate,
                               defectLinearized);
        mout << "RAM after multiply " << PPM->TotalMemoryUsage() << endl;

        return sqrt(FWD.scalarProductL2(defectLinearized, defectLinearized));
    }

    ~FWISDREGINN() {}
};

class FWISDREGINNCanonic : public FWISDREGINN {

public:
    FWISDREGINNCanonic(FWIForwardModel &FWD, std::string name, ReginnOptions opt)
        : FWISDREGINN(FWD, name, opt) {}

    void multiplyWithSteplength(const FWIDataContainerDomain &iterateNonlinear,
                                FWIDataContainerDomain &iterateUpdatePN,
                                FWIDataContainerDomain &iterateMUpdate,
                                FWISubproblem &defectLinearized) override {
        Exit("dont use this atm");
    }
};

class FWISDREGINNMulti : public FWISDREGINN {
    RVector invMap = RVector(1, 1);
public:
    FWISDREGINNMulti(FWIForwardModel &FWD, std::string name, ReginnOptions opt)
            : FWISDREGINN(FWD, name, opt) {
        auto reconlist = FWD.getReconstructionlist();
        invMap.resize(reconlist.size());
        size_t count = 0;
        for (auto it = reconlist.begin(); it != reconlist.end(); ++it) {
            invMap[count] = *it;
            count++;
        }
    }

    ~FWISDREGINNMulti() {}

    auto initializePreconditioner(const FWIDataContainerDomain &nonlinearPoint,
                                  const FWISubproblem &nonlinearEvaluation,
                                  const FWIDataContainerDomain &mNegativeUpdate,
                                  FWISubproblem &defectLinearized,
                                  int innerIter) -> void {
        if (FWD.getReconstructionType() == "AllAtOnce") {
            FWD.initializePrecon(nonlinearPoint.getMat(), mNegativeUpdate.getMat(), nonlinearPoint.getVectors());
        }
    }

    void applyPreconditioner(const FWIDataContainerDomain &nonlinearPoint,
                             FWIDataContainerDomain &mNegativeUpdate) {
        if (FWD.GetPrecon()) {
          FWD.applyPrecon(nonlinearPoint, mNegativeUpdate);
        }

    }

    void multiplyWithSteplength(const FWIDataContainerDomain &iterateNonlinear,
                                FWIDataContainerDomain &iterateUpdatePN,
                                FWIDataContainerDomain &iterateMUpdate,
                                FWISubproblem &defectLinearized) override {
        std::cout << std::setprecision(15);
        bool AAOflag = FWD.getReconstructionType() == "AllAtOnce";
        if (AAOflag) {
            FWD.initializePrecon(iterateNonlinear.getMat(), iterateUpdatePN.getMat(), iterateNonlinear.getVectors());
            FWD.SetPenalty(iterateUpdatePN.getMat(), defectLinearized);


            bool fullFlag = true;
            applyPreconditioner(iterateNonlinear, iterateUpdatePN);
            mout << "RAM before msdProcedure " << PPM->TotalMemoryUsage() << endl;

            if (fullFlag) {
                FWD.mSDProcedure(iterateNonlinear, iterateUpdatePN, iterateMUpdate, defectLinearized);
                mout << "RAM after msdProcedure " << PPM->TotalMemoryUsage() << endl;
                mout << "RAM before restofprocedure " << PPM->TotalMemoryUsage() << endl;

                FWISubproblem linearizedCalcs(defectLinearized);
                FWD.Linearized(iterateNonlinear, iterateUpdatePN, linearizedCalcs);
                defectLinearized -= linearizedCalcs;
                iterateMUpdate += iterateUpdatePN;
                mout << "RAM after restofprocedure " << PPM->TotalMemoryUsage() << endl;

                return;
            }


        }
      FWIDataContainerDomain mTemp(iterateUpdatePN);

      auto reconlist = FWD.getReconstructionlist();
        int vecsize = AAOflag ? invMap.size() + 1 : invMap.size();
        std::vector<FWISubproblem> linearizedCalcs(vecsize, defectLinearized);
        if (AAOflag) {
            mTemp = iterateUpdatePN;
            mTemp.getVectorsWrite() = 0.0;
            for (size_t i = 0; i < invMap.size(); ++i) {
                for (size_t j = 0; j < FWD.ParameterComponents(); ++j) {
                    if (invMap[i] != j) mTemp.setCompToZero(j);
                }
                FWD.Linearized(iterateNonlinear, mTemp, linearizedCalcs[i]);
            }
            mTemp = iterateUpdatePN;
            mTemp.getMatWrite() = 0.0;
            FWD.Linearized(iterateNonlinear, mTemp, linearizedCalcs[invMap.size()]);

        } else {
            for (size_t i = 0; i < invMap.size(); ++i) {
                mTemp = iterateUpdatePN;
                for (size_t j = 0; j < FWD.ParameterComponents(); ++j) {
                    if (invMap[i] != j) mTemp.setCompToZero(j);
                }
                FWD.Linearized(iterateNonlinear, mTemp, linearizedCalcs[i]);
            }
        }
        int matsize = AAOflag ? FWD.ParameterComponents() + 1 : FWD.ParameterComponents();
        SymRMatrix matrixEntries2(matsize);
        matrixEntries2.Identity();
        for (size_t i = 0; i < invMap.size(); ++i) {
            size_t pos1 = invMap[i];
            for (size_t j = 0; j < invMap.size(); ++j) {
                size_t pos2 = invMap[j];
                matrixEntries2(FWD.scalarProductL2(linearizedCalcs[i],
                                                   linearizedCalcs[j]),
                               pos1, pos2);
            }
        }
        if (AAOflag) {
            for (size_t i = 0; i < invMap.size(); ++i) {
                size_t pos1 = invMap[i];
                matrixEntries2(FWD.scalarProductL2(linearizedCalcs[i],
                                                   linearizedCalcs[invMap.size()]),
                               pos1, FWD.ParameterComponents());
            }
            matrixEntries2(FWD.scalarProductL2(linearizedCalcs[invMap.size()],
                                               linearizedCalcs[invMap.size()]),
                           FWD.ParameterComponents(), FWD.ParameterComponents());

        }

        RVector rightSideOne(0.0, matsize);
        RVector calculatedFactors(0.0, matsize);

        if (AAOflag) {
            for (size_t i = 0; i < invMap.size(); ++i) {
                rightSideOne[invMap[i]] =
                        FWD.scalarProductL2(defectLinearized, linearizedCalcs[i]);
            }
            rightSideOne[FWD.ParameterComponents()] = FWD.scalarProductL2(defectLinearized,
                                                                          linearizedCalcs[invMap.size()]);
        } else {
            for (size_t i = 0; i < invMap.size(); ++i) {
                rightSideOne[invMap[i]] =
                        FWD.scalarProductL2(defectLinearized, linearizedCalcs[i]);
                mTemp = iterateUpdatePN;
                for (size_t j = 0; j < FWD.ParameterComponents(); ++j) {
                    if (j != invMap[i]) mTemp.setCompToZero(j);
                }
            }
        }
        SymRMatrix matrixEntries3(matsize);
        matrixEntries3.Identity();
        matrixEntries3 *= 10e-16;
        matrixEntries2 += matrixEntries3;
        MOUT(matrixEntries2);
        matrixEntries2.Invert();
        calculatedFactors = matrixEntries2 * rightSideOne;
        MOUT(rightSideOne);
        MOUT(calculatedFactors);
        vout (4) << "the ts are: ";
        if (AAOflag) {
            for (int i = 0; i < invMap.size(); ++i) {
                std::string a =
                        " t_" + std::to_string(i) + " = " + std::to_string(calculatedFactors[invMap[i]]);
                vout(4) << a;
            }
            std::string b =
                    " t_" + std::to_string(1) + " = " + std::to_string(calculatedFactors[FWD.ParameterComponents()]);
            vout(4) << b;
            MOUT(calculatedFactors[FWD.ParameterComponents()]);
        } else {
            for (int i = 0; i < invMap.size(); ++i) {
                std::string a =
                        " t_" + std::to_string(i) + " = " + std::to_string(calculatedFactors[i]);
                vout(4) << a;
            }
        }
        vout(4) << endl;
        if (AAOflag) {
            if (invMap.size() > 1) {Exit("error!!"); }
            linearizedCalcs[0] *= calculatedFactors[invMap[0]];
            defectLinearized -= linearizedCalcs[0];
            linearizedCalcs[1] *= calculatedFactors[FWD.ParameterComponents()];
            defectLinearized -= linearizedCalcs[1];
        } else {
            for (size_t i = 0; i < invMap.size(); ++i) {
                linearizedCalcs[i] *= calculatedFactors[invMap[i]];
                defectLinearized -= linearizedCalcs[i];
            }
        }
        mTemp = iterateUpdatePN;
        if (AAOflag) {
            mTemp.getMatWrite() *= calculatedFactors[invMap[0]];
            mTemp.getVectorsWrite() *= calculatedFactors[FWD.ParameterComponents()];
            if (invMap.size() > 1) Exit("implementation for vector and smallvector");
        } else {
            mTemp *= calculatedFactors;
        }
        iterateMUpdate += mTemp;
//                iterateMUpdate.getMatWrite().MakeAdditive();
        mout << "norm of the new iterate: "
             << OUT(FWD.scalarProductL2(iterateMUpdate, iterateMUpdate)) << endl;
    };
};

FWINonlinearSolver *GetFWINonlinearSolver(FWIForwardModel &FWD) {
    std::string solver = "FWISteepestDescent";
    Config::Get("FWINonlinearSolver", solver);
//    if (solver == "FWI_CGREGINN_")
//        return new FWICGREGINN(FWD);
    if (solver == "FWI_CGREGINN") {
        std::string name = "CGREGINN";
        std::string type = "Canonic";
        Config::Get("CGREGINNType", type);
        ReginnOptions opt(name);
        name = name + type;
        if (type == "Canonic") {
            return new FWICGREGINNCanonic(FWD, name, opt);
        } else if (type == "Multi") {
            return new FWICGREGINNMulti(FWD, name, opt);
        } else if (type == "WavePrecon") {
            return new FWICGREGINNWaveFieldPrecon(FWD, name, opt);
        }
    }
    if (solver == "FWISteepestDescent_") {
        std::string name = "SteepestDescent";
        ReginnOptions opt(name);
        opt.maxInnerLoopSteps = 1;
        return new FWICGREGINNCanonic(FWD, name, opt);
    } else if (solver == "FWI_mSDREGINN") {
        std::string name = "mSDREGINN";
        ReginnOptions opt(name);
        return new FWISDREGINNMulti(FWD, name, opt);
    } else if (solver == "FWISO"){
      std::string name = "FWISO";

      ReginnOptions opt(name);

      return new FWINonlinearSolverSO(FWD,name,opt);
    }
    Exit("Not implemented");

    return nullptr;
}


