#include "FWIVectorFillFunctions.hpp"
#include "FWIDataContainerDomain.hpp"
#include "FWIDataContainerRange.hpp"
#include <utility>
#include <memory>
#include "Assembler/FWIRhsAssemblerAdjoint.hpp"
#include "DGAssemble.hpp"
#include "TimeIntegrator.hpp"
#include "DGVAcousticAssembleBwd.hpp"
#include "FWIForwardModelDGTSVA.hpp"
#include "FWIForwardModelDGTSVAAllAtOnce.hpp"
#include "Assembler/FWILinearizedAssembler.hpp"
#include "VectorsIntegrator.hpp"
#include "AAO/FWILinearizedIntegralOperator.hpp"
#include "BasicSolver.hpp"
#include "FWIForwardModelDGTS.hpp"
#include "Observation/FWIObservationDGTSNEW.hpp"
#include "FWITimeSolver.hpp"
#include "Reduced/FWILinOpRedFirstOrder.hpp"


void FWIForwardModelDGTS::setProblem(std::shared_ptr<WaveProblemDGTSVA> _prob) {
  if (problem_fwd) { problem_fwd.reset(); }
  problem_fwd = _prob;
  initializeRecon();
  numL = problem_fwd->getNumL();
  stateComps = problem_fwd->getStateComps();
  paramComps = problem_fwd->getParamComps();
  initializeSpaceDiscretization(FWIDiscStack);
  bool PML = false;
  Config::Get("PaddingPML", PML);
  PMLH = std::make_shared<PMLHandler>();
  if (PML) {
    PMLConf PMLConfig;
    vector<double> PMLPaddingLength;
    Config::Get("PMLPaddingLength",PMLPaddingLength);
    PMLConfig.SetPMLPaddingLength(PMLPaddingLength);
    Config::Get("PMLBnds", PMLConfig.pmlBndFlags);
    Config::Get("PMLDistFactor", PMLConfig.distfactor);
    int ts=0;
    Config::Get("PMLTaperStyle", ts);
    double alpha1 = numL>0 ? problem_fwd->alpha1() : -1;
    double alpha2 = numL>0 ? problem_fwd->alpha2() : -1;
    PMLConfig.SetTaperStyle(ts,alpha1,alpha2);
    PMLH->Initialize(FWIDiscStack.wV->GetMesh(), PMLConfig, deg, numL, paramComps);
  }
  plotEnding = static_cast<int>(round(TConf.timeSteps / plotFactor)) + 1;
  eleCacheContainer = std::make_unique<EleCacheContainer>(GetWaveVec(), problem_fwd->getNumL());
  VAPlotter = std::make_shared<DGVAcousticPlotter>(problem_fwd->getNumL(),
                                                   &eleCacheContainer->Get());
  initializeMeasureMask();
  std::shared_ptr<FWIObservationSpecification> spec = nullptr;
  string l;
  Config::Get("Specfile",l);
  if (l.empty()) {
    spec = FWIObsSpecWPCreator().WithT(TConf.T).Withdt(TConf.dt).Witht0(TConf.t0).CreateShared();
  }else{
    int SamplingFactor = 1;
    Config::Get("SpecfileSamplingFactor",SamplingFactor);
    spec = std::make_shared<FWIObservationSpecification>(TConf.t0,TConf.T,TConf.dt,l.c_str());

  }
  Config::Get("UseGardener",useGardener);
  if (PMLH->isPML()) {
    Obs = FWIObservationCreator(PMLH->GetEleCache(), PMLH->GetDiscStack().GetMesh(),
                                problem_fwd->getNumL()).WithSpec(spec).CreateUnique();
  } else {
    Obs = FWIObservationCreator(eleCacheContainer->Get(), FWIDiscStack.GetMesh(),
                                problem_fwd->getNumL()).WithSpec(
        spec).CreateUnique();
  }
  Config::Get("useTaper", taper);
  if (taper) {
    string fn;
    Config::Get("TaperFromFile", fn);
    TaperConf Conf;
    if ( fn.empty()) {
      Taper = std::make_unique<FWITaper>(GetMatVec(), Obs->GetSpaceOp().GetSpecification(), Conf,
                                         *problem_fwd);
    }else {
      Taper = std::make_unique<FWITaper>(GetMatVec(),fn,problem_fwd->getParamComps());
    }

  }

}

void FWIForwardModelDGTS::initializeSpaceDiscretization(FWIDiscretizationStack &discStack) {
  const auto &meshes = *discStack.meshes;
  discStack.mD = std::make_unique<DGDiscretization>(meshes, 0, paramComps);
  meshes.PrintInfo();
  (*discStack.mD)().PrintInfo();
  discStack.mV = std::make_shared<Vector>(0.0, *discStack.mD, discStack.lev);
  discStack.wD = std::make_unique<DGDiscretization>(meshes, deg, GetDim() + 1 + numL);
  (*discStack.mD)().PrintInfo();
  discStack.wV = std::make_shared<Vector>(0.0, *discStack.wD, discStack.lev);
}

void FWIForwardModelDGTS::makeFD(const Vectors &U_m, Vectors &U_m_fd) const {
  U_m_fd = 0.0;
  for (int i = 0; i < U_m.size() - 1; ++i) {
    U_m_fd[i] = U_m[i + 1];
    U_m_fd[i] -= U_m[i];
    U_m_fd[i] *= 1.0 / TConf.dt;
  }
}

double FWIForwardModelDGTS::L2TScalar(const Vectors &u, const Vectors &v) const {
  double l2t = 0.0;
  for (int tsteps = 0; tsteps < u.size(); ++tsteps) {
    if (tsteps == 0 || tsteps == u.size() - 1) {
      l2t += 0.5 * L2tScalar(u[tsteps], v[tsteps]);
    } else {
      l2t += L2tScalar(u[tsteps], v[tsteps]);
    }
  }
  return l2t * TConf.dt;
}


void FWIForwardModelDGTS::Parameter2SeismogramMap(const Vector &m, FWISeismogram &s, int shot,
                                                  Vectors *U_m) {
  problem_fwd->setShot(shot);
  std::shared_ptr<DGAssemble> assemble = getDGAssemble(m, *problem_fwd, *problem_fwd, true);
  FWITimeSolver TSolveR(*FWIDiscStack.wV, PMLH, TConf.rkorder);
  TSolveR.SolveAndMeasure(assemble, *Obs, s, U_m);
}

double FWIForwardModelDGTS::L2tScalar(const Vector &u, const Vector &v) const {
  Scalar e = 0;
  for (const auto &elem: eleCacheContainer->Get()) {
    e += CellWiseStepProduct(u, v, *elem.second, 1.0);
  }
  return PPM->Sum(e);
}

double FWIForwardModelDGTS::L2Scalar(const Vector &u,
                                     const Vector &v) const { // needs to change if not constant materials are used.
  double l2 = 0.0;
  for (const auto &elemIt: eleCacheContainer->Get()) {
    double a = elemIt.second->Area();
    row ru = u.find_row(elemIt.first);
    row rv = v.find_row(elemIt.first);
    for (unsigned int k: reconlist) {
      double para_u = u(ru, k);
      double para_v = v(rv, k);
      l2 += para_u * para_v * a;
    }
  }
  return PPM->Sum(l2);
}

void FWIForwardModelDGTS::projectModel(const Vector &U2, Vector &UM) const {
  for (cell c = U2.cells(); c != U2.cells_end(); c++) {
    for (cell c2 = UM.cells(); c2 != UM.cells_end(); ++c2) {
      if (PointInCellDGTS(c2(), c)) {
        for (int i = 0; i < paramComps; ++i) {
          UM(c2(), i) = U2(c(), i);
        }
      }
    }
  }
}

double
FWIForwardModelDGTS::CellWiseStepProduct(const Vector &U, const Vector &V,
                                         const DGAcousticElement &elem,
                                         double dt_) const {
  double e = 0.0;
  for (int q = 0; q < elem.nQ(); ++q) {
    double w = elem.QWeight(q);
    e += w * dt_ * elem.Velocity(q, U) * elem.Velocity(q, V);
    for (int n = 0; n < elem.p_dimension(); ++n) {
      e += w * dt_ * elem.Pressure_i(q, U, n) * elem.Pressure_i(q, V, n);
    }
  }
  return e;
}

double
FWIForwardModelDGTS::CellWiseTimeProduct(const Vectors &U, const Vectors &V,
                                         const DGAcousticElement &elem,
                                         double dt_) const {
  double skp = 0.0;
  for (int l = 0; l < U.size(); ++l) {
    if (l == 0 || l == U.size() - 1) {
      skp += CellWiseStepProduct(U[l], V[l], elem, 0.5 * dt_);
    } else skp += CellWiseStepProduct(U[l], V[l], elem, dt_);
  }
  return skp;
}


void FWIForwardModelDGTS::initializeRecon() {
  if (problem_fwd->getParametrization() == "native") {
    Config::Get("FWIReconstructRho", recon.Rho);
    Config::Get("FWIReconstructKappa0", recon.Kappa0);
    Config::Get("FWIReconstructKappaL", recon.KappaL);
    Config::Get("FWIReconstructTaus", recon.Taus);
    if (recon.Rho) {
      reconlist.push_back(0);
    }
    if (recon.Kappa0) {
      reconlist.push_back(1);
    }
    if (recon.KappaL) {
      for (int i = 1; i <= numL; ++i) {
        reconlist.push_back(1 + i);
      }
    }
    if (recon.Taus) {
      for (int i = 1; i <= numL; ++i) {
        reconlist.push_back(numL + 1 + i);
      }
    }
    if (reconlist.empty()) {
      Exit("No parameters to be reconstructed!!");
    }
  } else if (problem_fwd->getParametrization() == "taup"
             || problem_fwd->getParametrization() == "taup_positive") {
    Config::Get("FWIReconstructRho", recon.Rho);
    Config::Get("FWIReconstructV_P", recon.Vp);
    Config::Get("FWIReconstructTauP", recon.Taup);
    if (recon.Rho) {
      reconlist.push_back(0);
    }
    if (recon.Vp == 1) {
      reconlist.push_back(1);
    }
    if (numL > 0) {
      if (recon.Taup == 1) {
        reconlist.push_back(2);
      }
    }
    reconVec.resize(reconlist.size());
    int count = 0;
    for (unsigned int &it: reconlist) {
      reconVec[count] = it;
      count++;
    }
  }

}

void FWIForwardModelDGTS::initializeMeasureMask() {
  int MeasureComp = 0;
  Config::Get("compMeasure", MeasureComp);
  Mask = GetMeasureMaskNEW(problem_fwd->getDim(), problem_fwd->getNumL(), MeasureComp);
}

void FWIForwardModelDGTS::assemblePseudoHessianPrecon(const Vector &m, const Vector &m_update,
                                                      const Vectors &u,
                                                      const Vectors &u_fd, Matrix &PHess,
                                                      Vector *Plotti) const {
  FWILinearizedAssembler fwiLinearizedAssembler(problem_fwd, eleCacheContainer->Get(), m);
  int Ninv = reconVec.size();
  Vector DiagonalCorrelation(0.0, m_update);
  Vector OffDiagonalCorrelation(0.0, m_update);
  Vector TmpMat(0.0, m_update);
  FWIAdjointCalculator adj(m, problem_fwd->getParametrization(),
                           eleCacheContainer->Get(),
                           recon,
                           numL, TConf.dt, problem_fwd);

  if (precon.second == 1) {
    Exit("Dont use this now!")
  } else if (precon.second == 2) {
    for (int step = 0; step < u.size(); ++step) {
      double dt_ = (step == 0 || step == u.size() - 1) ? 0.5 * TConf.dt : TConf.dt;
      adj.PseudoHessianHelper(u[step], u_fd[step], u[step], TmpMat);
      for (const auto &elemIt: eleCacheContainer->Get()) {
        const Cell &C = elemIt.second->GetCell();
        for (int param = 0; param < Ninv; param++) {
          DiagonalCorrelation(C(), param) += dt_ * TmpMat(C(), reconVec[param])
                                             * TmpMat(C(), reconVec[param]);
          for (int param2 = param + 1; param2 < Ninv; param2++) {
            OffDiagonalCorrelation(C(), param + param2 - 1)
                += dt_ * TmpMat(C(), reconVec[param]) * TmpMat(C(), reconVec[param2]);
          }
        }
      }
      TmpMat = 0.0;
    }
  }
  double maxi = 0;
  double mini = -infty;
  for (const auto &elemIt: eleCacheContainer->Get()) {
    const Cell &C = elemIt.second->GetCell();
    for (int param = 0; param < Ninv; param++) {
      maxi = max(fabs(maxi), fabs(DiagonalCorrelation(C(), param)));
      mini = min(fabs(mini), fabs(DiagonalCorrelation(C(), param)));
      for (int param2 = param + 1; param2 < Ninv; param2++) {
        maxi = max(fabs(maxi), fabs(OffDiagonalCorrelation(C(), param + param2 - 1)));
        mini = min(fabs(mini), fabs(OffDiagonalCorrelation(C(), param + param2 - 1)));

      }
    }
  }
  double allmax = PPM->Max(maxi);
  double allmin = PPM->Min(mini);
  double normalize = sqrt(allmax / allmin);
//  MOUT(normalize);
//  MOUT(allmax);
//  MOUT(allmin);
  DiagonalCorrelation *= normalize;
  OffDiagonalCorrelation *= normalize;
  double allmaxRegu = allmax * sqrt(allmax / allmin) * PreconRegu;
//  MOUT(allmaxRegu);
//  Vector PlottiOffDiag(*Plotti);
//  PlottiOffDiag = 0.0;
  for (const auto &elemIt: eleCacheContainer->Get()) {
    SymRMatrix BlockMat(Ninv);
    const Cell &C = elemIt.second->GetCell();
    for (int p = 0; p < Ninv; p++) {
      BlockMat(p, p) = DiagonalCorrelation(C(), p);
      for (int p2 = p + 1; p2 < Ninv; p2++) {
        BlockMat(p, p2) = OffDiagonalCorrelation(C(), p + p2 - 1);
      }
    }
    for (int i = 0; i < Ninv; ++i) {
      BlockMat(i, i) += allmaxRegu;
    }
    BlockMat.Invert();
    DGRowEntries PHess_c(PHess, C, C);
    for (int param = 0; param < Ninv; ++param) {
      PHess_c(reconVec[param], reconVec[param]) = BlockMat(param, param);
//      (*Plotti)(C(), param) = BlockMat(param, param);
      for (int param2 = param + 1; param2 < Ninv; ++param2) {
        PHess_c(reconVec[param], reconVec[param2]) = BlockMat(param, param2);
        PHess_c(reconVec[param2], reconVec[param]) = BlockMat(param, param2);
//        PlottiOffDiag(C(), param + param2 - 1) = BlockMat(param, param2);
      }
    }
  }
//    auto &Plotm = mpp::plot("m");
//  Plotm.AddData("Rho", m, 0);
//  Plotm.AddData("Vp", m, 1);
////  Plotm.AddData("Taup", m, 2);
//  Plotm.PlotFile();
//  auto &Plotmupdate = mpp::plot("mupdate");
//  Plotmupdate.AddData("Rho", m_update, 0);
//  Plotmupdate.AddData("Vp", m_update, 1);
////  Plotmupdate.AddData("Taup", m_update, 2);
//  Plotmupdate.PlotFile();
//  auto &Plot = mpp::plot("valuesNEWNEWNEW");
//  Plot.AddData("RhoRho", DiagonalCorrelation, 0);
//  Plot.AddData("VpVp", DiagonalCorrelation, 1);
//  Plot.AddData("TaupTaup", DiagonalCorrelation, 2);
//  Plot.AddData("RhoVp", OffDiagonalCorrelation, 0);
//  Plot.AddData("RhoTaup", OffDiagonalCorrelation, 1);
//  Plot.AddData("TaupVp", OffDiagonalCorrelation, 2);
//  Plot.PlotFile();
//    auto &Plot2 = mpp::plot("valuesNEWNEWNEW222");
//    Plot2.AddData("RhoRho", *Plotti, 0);
//    Plot2.AddData("VpVp", *Plotti, 1);
//    Plot2.AddData("TaupTaup", *Plotti, 2);
//    Plot2.AddData("RhoVp", PlottiOffDiag, 0);
//    Plot2.AddData("RhoTaup", PlottiOffDiag, 1);
//    Plot2.AddData("TaupVp", PlottiOffDiag, 2);
//    Plot2.PlotFile();
//  Exit("lel");
  (*MPC).CreateSparse();
}

//  double reguLambda = 1e-12;

//}



//}

int FWIForwardModelDGTS::GetDim() const { return problem_fwd->getDim(); }

void FWIForwardModelDGTS::fillSourceAndReceiverCellsFromProblem(Vector &M,
                                                                const WaveProblemDGTSVA &prob) const {
    for (cell c = M.cells(); c != M.cells_end(); ++c) {
        for (const auto &re: Obs->GetSpaceOp().GetSpecification()) {
            if (PointInCellDGTS_old(re, *c)) {
              FillCellFromProblem(c, prob, M);
            }
        }
        for (int shot = 0; shot < problem_fwd->getShotCount(); ++shot) {
            problem_fwd->setShot(shot);
            int nrPointSrc = problem_fwd->getNrPointSrc();
            for (int i = 0; i < nrPointSrc; ++i) {
                Point src = problem_fwd->getPointSrcLoc(i);
                if (PointInCellDGTS_old(src, *c)) {
                  FillCellFromProblem(c, prob, M);
                }
            }
        }
    }
}