#include "BasicSolver.hpp"
#include "AAO/FWILinearizedIntegralOperator.hpp"
#include "VectorsIntegrator.hpp"
#include "Assembler/FWILinearizedAssembler.hpp"
#include "FWIForwardModelDGTSVAAllAtOnce.hpp"
#include "FWIForwardModelDGTSVA.hpp"

#include <memory>
#include <utility>
#include "DGVAcousticAssembleBwd.hpp"
#include "FWIForwardModelDGTS.hpp"
#include "TimeIntegrator.hpp"
#include "DGAssemble.hpp"
#include "Assembler/FWIRhsAssemblerAdjoint.hpp"
#include "FWIDataContainerRange.hpp"
#include "FWIDataContainerDomain.hpp"
#include "Reduced/FWILinOpRedFirstOrder.hpp"
#include "FWIVectorFillFunctions.hpp"
#include "FWIProblemWrapperVA.hpp"

FWIForwardModelDGTSVA::FWIForwardModelDGTSVA(std::shared_ptr<Meshes> meshes, LevelPair lev, int deg,
                                             const std::shared_ptr<WaveProblemDGTSVA> &problem_fwd)
        : FWIForwardModelDGTS(std::move(meshes), lev, deg, problem_fwd) {
}


// --------------------------------------------------------------
// From abstract interface
std::shared_ptr<DGAssemble> FWIForwardModelDGTSVA::getDGAssemble(const Vector &m, const FWIBndConditions &bnd,
                                                                 const FWIRhs &rhs, bool fwd) {
    if (fwd) {
        auto assemble = std::make_shared<DGAssemble>(std::move(GetWrappedProblem(m, bnd, rhs, fwd)), true, -1);
        if (PMLH->isPML()) {
            assemble->SetElementCache(&PMLH->GetEleCache());
        } else {
            assemble->SetElementCache(&eleCacheContainer->Get());
        }
        return assemble;
    } else {
        auto assemble = std::make_shared<DGVAcousticAssembleBwd>(std::move(GetWrappedProblem(m, bnd, rhs, fwd)), -1);
        if (PMLH->isPML()) {
            assemble->SetElementCache(&PMLH->GetEleCache());
        } else {
            assemble->SetElementCache(&eleCacheContainer->Get());
        }
        return assemble;
    }
}


// --------------------------------------------------------------
// Parameter-to-solution and Parameter-to-Seismogram


// H

void FWIForwardModelDGTSVA::plot(const Vectors &U, const char *name, int nr, int) {
    if (!allowPlotting_WF) return;
    for (int i = 0; i < plotEnding; ++i) {
        (*VAPlotter).plotWaveVector(U[i * plotFactor], i, name);
    }
}

void FWIForwardModelDGTSVA::plot(const FWIDataContainerDomain &m, const char *name, int nr,
                                 int hierarchylvl) {

    const Vector &M = m.getMat();
    if (!allowPlotting_M) return;
    std::string filename;
    if (hierarchylvl < 0) {
        filename = nr_str(name, nr);
    } else {
        filename = nr_str(name, nr, hierarchylvl);
    }
    plotMatVector(M, filename);
    if (m.isWaveField()) plot(m.getVectors(), (filename + "_WF").c_str());
}

void FWIForwardModelDGTSVA::plot(const Vector &M, const char *name, int nr, int hierarchylvl) {
    if (!allowPlotting_M) return;
    std::string filename;
    if (hierarchylvl < 0) {
        filename = nr_str(name, nr);
    } else {
        filename = nr_str(name, nr, hierarchylvl);
    }
    plotMatVector(M, filename);
};


void plotComponent(const Vector& M, std::string component,int index,const std::string &filename){
    auto& Plot = mpp::plot(component + "_" + filename);
    Plot.AddData(component,M,index);
    Plot.PlotFile();
}

void FWIForwardModelDGTSVA::plotMatVector(const Vector &M, const std::string &filename) {

    LagrangeDiscretization disc(M.GetDisc().GetMeshes(), 0);
    if (recon.Rho) { plotComponent(M,"Rho",0,filename);}
    if (problem_fwd->getParametrization() == "native") {
        if(recon.Kappa0) {
            for (int i = 0; i <= numL; ++i) {
                std::string kappaName = "Kappa" + std::to_string(i);
                plotComponent(M, kappaName, i + 1, filename);
                if (!recon.KappaL) break;
            }
        }
    } else if (problem_fwd->getParametrization() == "taup") {
        if (recon.Vp) {plotComponent(M,"Vp",1,filename);}
        if (recon.Taup && numL > 0) {plotComponent(M,"Taup",2,filename);}
    } else if (problem_fwd->getParametrization() == "taup_positive") {
        Vector M2(M);
        int taupPlotLevel = 0;
        Config::Get("taupPlotLevel", taupPlotLevel);
        for (cell c = M2.cells(); c != M2.cells_end(); ++c) {
            int _comps = 2 + (numL > 0);
//      mout << _comps  << endl;
            for (int i = 0; i < _comps; ++i) {
                M2(c(), i) = exp(M(c(), i));
            }
        }
        if (recon.Vp) {
            if (taupPlotLevel >= 1) {
                plotComponent(M, "Vp", 1, filename);
            }
            plotComponent(M2, "VpExp", 1, filename);
        }
        if (recon.Taup && numL > 0) {
            if (taupPlotLevel >= 1) {
                plotComponent(M, "Taup", 2, filename);
            }
            plotComponent(M2, "TaupExp", 2, filename);
        }
        if (recon.Rho) plotComponent(M2,"RhoExp",0,filename);
    } else {
        Exit("Unknown parameterization");
    }
}


FWIForwardModelDGTSVA::~FWIForwardModelDGTSVA() {
}

void FWIForwardModelDGTSVA::NonlinearEvaluation(const FWIDataContainerDomain &nonlinearPoint, FWISubproblem &evaluation,
                                                std::vector<int> subproblem) {
    if (subproblem.empty()) {Exit("subproblem not properly defined!")}
    if (uM.size() != subproblem.size()) {
        if (uM.size() > subproblem.size()) {
            for (size_t i = subproblem.size(); i < uM.size(); ++i) uM[i].release();
        }
        if (uM.size() < subproblem.size()) {
            for (size_t i = uM.size(); i < subproblem.size(); ++i)
                uM.push_back(std::make_unique<Vectors>(TConf.timeSteps + 1, GetWaveVec()));
        }
    }
    if(subproblem.size() > 1) Exit("Not like this")
    for (size_t i = 0; i < subproblem.size(); ++i) {
        mout << "shot: " << subproblem[i] << ": Solving forward problem..." << endl;
       Nonlinear(nonlinearPoint.getMat(),*uM[i], evaluation[i].getSeismogramWrite(),subproblem[i]);
       mout << "Assembling Linearized Operators" << endl;
       MakeLinearAndAdjointOperator(nonlinearPoint.getMat(), *uM[i]);
        mout << "Done!" << endl;


    }
}

void FWIForwardModelDGTSVA::initializePrecon(const Vector &M, const Vector &M_update,
                                             const Vectors &U_M) {
    Vectors U_M_FD(U_M);
    makeFD(U_M, U_M_FD);
    MPC = std::make_unique<Matrix>(M_update);
    Date PreconStart;
    Vector PlotPC(M_update);
    assemblePseudoHessianPrecon(M, M_update, U_M, U_M_FD, *MPC, &PlotPC);
    mout << "time to assemble precon:" << Date() - PreconStart << endl;
    plot(PlotPC, "PC_Values");
}

void FWIForwardModelDGTSVA::AdjointLinearized(const FWIDataContainerDomain &nonlinearPoint,
                                              const FWISubproblem &adjointLinearizedPoint,
                                              FWIDataContainerDomain &adjointLinearizedEvaluation) {
    FWIDataContainerDomain tmp(getReconstructionType(), GetMatVec(), GetWaveVec(), TConf.timeSteps);
    for (size_t i = 0; i < uM.size(); ++i) {
        AdjOp->CalculateAdjoint(adjointLinearizedPoint[i].getSeismogram(), tmp.getMatWrite(), nullptr);
        adjointLinearizedEvaluation.getMatWrite() += tmp.getMat();
    }
    if (Taper) {
      Taper->ApplyTaperVector(adjointLinearizedEvaluation.getMatWrite());
      Taper->ApplyMinMaxTaper(adjointLinearizedEvaluation.getMatWrite(),*problem_fwd);
    }

    adjointLinearizedEvaluation.getMatWrite().MakeAdditive();
}

void FWIForwardModelDGTSVA::Linearized(const FWIDataContainerDomain &nonlinearPoint,
                                       const FWIDataContainerDomain &linearizedPoint,
                                       FWISubproblem &linearizedEvaluation) {
    for (size_t i = 0; i < uM.size(); ++i) {
        LinOp->SolveAndMeasure(linearizedPoint.getMat(), *Obs, linearizedEvaluation[i]
                .getSeismogramWrite(), nullptr);
    }
}

void FWIForwardModelDGTSVA::calculateDefect(const vector<FWIDataContainerRange> &data_obs,
                                            const FWISubproblem &nonlinearEvaluation, FWISubproblem &defect,
                                            std::vector<int> subproblem) {
    defect.resize(subproblem.size(),
                  FWIDataContainerRange(data_obs[0].getSeismogram()));
    if (subproblem.size() > 1) Exit("Check this for bigger subproblems!")
      for (size_t i = 0; i < subproblem.size(); ++i) {
        defect[i].getSeismogramWrite() = data_obs[subproblem[i]].getSeismogram();
        defect[i].getSeismogramWrite() -= nonlinearEvaluation[i].getSeismogram();
      }
}

std::shared_ptr<ProblemVAcoustic>
FWIForwardModelDGTSVA::GetWrappedProblem(const Vector &m, const FWIBndConditions &bnd, const FWIRhs &rhs, bool fwd) {
    std::unique_ptr<Vector> Mp = nullptr;
    std::unique_ptr<Vector> M_Vec = std::make_unique<Vector>(m);
    M_Vec->Clear();
  ReparametrizeMaterialVector(m,*problem_fwd,*M_Vec);
    if (PMLH->isPML()) {
        Mp = std::make_unique<PMLMatVectorFWI>(*PMLH, *M_Vec,true);
    } else if (PMLHAAO){
        PMLHAAO->SetPMLZero(*M_Vec);
        PMLHAAO->CommunicateBoundaryToPML(*M_Vec);
        PMLHAAO->increaseImpedance(*M_Vec,problem_fwd->getNumL());
        Mp = std::move(M_Vec);
    }else {
      Mp = std::move(M_Vec);
    }

    Mp->Accumulate();
    const auto &rhs_va = dynamic_cast<const FWIRhsVAcoustic &>(rhs);
    const auto &bnd_va = dynamic_cast<const FWIBndConditionsVAcoustic &>(bnd);
    std::shared_ptr<FWIProblemWrapperVA> problem = std::make_shared<FWIProblemWrapperVA>(rhs_va, bnd_va,
                                                                                         std::move(Mp),
                                                                                         problem_fwd->getNumL(),
                                                                                         problem_fwd->getDim(),
                                                                                         problem_fwd);
  auto MatI  =std::make_unique<FWIDGVAcousticMaterialFromVector>(*problem_fwd,m,PMLH.get(),PMLHAAO.get());
  std::shared_ptr<FWIProblemWrapperVA2> problem2 = std::make_shared<FWIProblemWrapperVA2>(rhs_va, bnd_va,
                                                                                       std::move(MatI),
                                                                                       problem_fwd->getNumL(),
                                                                                       problem_fwd->getDim(),
                                                                                       problem_fwd);
    return problem2;
}
