//
// Created by cr on 10.11.22.
//

#include "FWIObservationDGTSNEW.hpp"
#include "FWIMppHelpers.hpp"
#include "Observation/FWIObservationSpecification.hpp"
#include <memory>

void FWIObservationOperatorSpacePoint::initalizeShapeValues(const FWIObservationSpecification &spec,
                                                            const Mesh &b) {
    for (cell c = b.cells(); c != b.cells_end(); ++c) {
        for (const auto &receiver: spec) {
            if (PointInCellDGTS(receiver, c)) {
                SpaceReceiver2SpaceSupport[receiver].push_back(c());
                auto elem = *EleCache.find((*c)())->second;
                int nV = elem.V_dimension();
                int nP = elem.shape_size();
                RVector PV(0.0, nP);
                Point lz = (*c).GlobalToLocal(receiver);
                for (int i = 0; i < nP; ++i) {
                    PV[i] = elem.evaluateShape(lz, i);
                }
                PVals[receiver].emplace(c(), PV);
                if (measureComp == -1) {
                    std::vector<VectorField> VFV(nV, zero);
                    for (int i = 0; i < nV; ++i) {
                        VFV[i] = elem.evaluateVelocityShape(lz, i);
                    }
                    VFVals[receiver].emplace(c(), VFV);
                }
            }
        }
    }
}

void FWIObservationOperatorSpacePoint::SetSpaceValue(const FWIState &Measure, const Point &Receiver,
                                                     const std::list<Point> &MeasureCells,
                                                     Vector &b) const {
    auto PValsAtRec = PVals.at(Receiver);
    auto cell_count = ReceiverToCellCount.at(Receiver);
    for (const auto &MeasureCell: MeasureCells) {
        auto elem = *EleCache.find(MeasureCell)->second;
        int p_dim = elem.p_dimension();
        int nP = elem.shape_size();
        const RVector &PV = PValsAtRec.at(MeasureCell);
        for (int i = 0; i < nP; ++i) {
            for (int n = 0; n < p_dim; ++n) {
                b(elem[0], (b.dim() + n) * nP + i) += Measure[b.dim() + n] * PV[i] * (1.0 / cell_count);
            }
        }
        if (measureComp == -1) {
            const auto &VF = VFVals.at(Receiver).at(MeasureCell);
            for (int i = 0; i < nP; ++i) {
                for (int j = 0; j < b.dim(); ++j) {
                    b(elem[0], j * nP + i) += Measure[j] * VF[i][j] * (1.0 / cell_count);
                }
            }
        }
    }
}

FWIState FWIObservationOperatorSpacePoint::GetSpaceValue(const Vector &U, const Point &Receiver,
                                                         const std::list<Point> &MeasureCells) const {
    FWIState Measure(stateComps);
    auto PValsAtRec = PVals.at(Receiver);
    auto cell_count = ReceiverToCellCount.at(Receiver);
    for (const auto &MeasureCell: MeasureCells) {
        auto elem = *EleCache.find(MeasureCell)->second;
        int p_dim = elem.p_dimension();
        int nP = elem.shape_size();
        const RVector &PV = PValsAtRec.at(MeasureCell);
        for (int i = 0; i < nP; ++i) {
            for (int n = 0; n < p_dim; ++n) {
                Measure[U.dim() + n] += U(elem[0], (U.dim() + n) * nP + i) * PV[i];
            }
        }
        if (measureComp == -1) {
            const auto &VF = VFVals.at(Receiver).at(MeasureCell);
            for (int i = 0; i < nP; ++i) {
                for (int j = 0; j < U.dim(); ++j) {
                    Measure[j] += U(elem[0], j * nP + i) * VF[i][j];
                }
            }
        }
    }
    return Measure * (1.0 / cell_count);
}


void FWIObservationOperatorNEW::MeasureFullWave(const Vectors &U, FWISeismogram &s) const {
    s.ts_clear();
    s.setStateComponents(Mask->GetSeismoComps());
    const auto &spec = spaceOp->GetSpecification();
    s.setSpecification(spec);
    for (const auto &receiver: spaceOp->GetSpecification()) {
        if (!spaceOp->MeasureHere(receiver)) continue;
        for (int step = 0; step < spec.nSteps(); ++step) {
            auto *Measure = new FWIState(0.0, Mask->GetStateComps());
            auto timelist = timeOp->TimeSupport(step);
            for (auto time: timelist) {
                (*Measure) += spaceOp->GetSpaceValue(U[time], receiver, spaceOp->GetSpaceSupport().at(receiver)) *
                              (*timeOp)(spec.step(time));
            }
            Mask->mask(*Measure);
            s.addMeasurement(receiver, spec.step(step), Measure);
        }
    }
}

void FWIObservationOperatorNEW::MeasureSingleWaveField(const Vector &U, FWISeismogram &s, int step) const {
    if (timeOp->TimeSupport(step).size() > 1)
        THROW("Can't use this Measure Single WaveField with a Time Measurement with Support 1")
    for (const auto &receiver: spaceOp->GetSpecification()) {
        if (!spaceOp->MeasureHere(receiver)) continue;
        auto *Measure = new FWIState(0.0, Mask->GetStateComps());
        double t = spaceOp->GetSpecification().step(step);
        (*Measure) = spaceOp->GetSpaceValue(U, receiver, spaceOp->GetSpaceSupport().at(receiver));
        Mask->mask(*Measure);
        s.addMeasurement(receiver, t, *Measure);
        delete Measure;
    }
}

const FWIObservationOperatorSpace &FWIObservationOperatorNEW::GetSpaceOp() const {
    return *spaceOp;
}

const FWIObservationOperatorTime &FWIObservationOperatorNEW::GetTimeOp() const {
    return *timeOp;
}

const MeasureMask &FWIObservationOperatorNEW::GetMask() const {
    return *Mask;
}

void printResult(double t0, double dt, double T, std::list<Point> &receiver_positions) {
    mout << "ReceiverStart: " << t0 << " ";
    mout << "ReceiverEnd: " << T << " ";
    mout << "ReceiverDt: " << dt << endl;
    int i = 0;
    for (auto p = receiver_positions.begin(); p != receiver_positions.end(); ++p, ++i) {
        mout << "  " << i << ": " << *p << " ";
        if (i < 10) mout << " ";
        if (((i + 1) % p->SpaceDim()) == 0) mout << endl;
    }
    mout << endl;
}

FWIObservationSpecification *
GetObservationSpecificationWaypoints(double t0, double dt, double T, vector<Point> &waypoints,
                                     vector<int> &receiverCounts, bool include_last_waypoint) {
    auto waypoint_count = waypoints.size();
    if (waypoint_count - 1 != receiverCounts.size()) Exit("Specify one count fewer than waypoints");
    if (waypoint_count < 2) Exit("At least 2 waypoints needed!");
    mout << "ReceiverStart: " << t0 << " ";
    mout << "ReceiverEnd: " << T << " ";
    mout << "ReceiverDt: " << dt << endl;
    std::list<Point> receiver_positions = PointsOnLine(waypoints, receiverCounts, include_last_waypoint);
    printResult(t0, dt, T, receiver_positions);
    return new FWIObservationSpecification(t0, T, dt, receiver_positions);
}

const FWIObservationSpecification &FWIObservationOperatorSpace::GetSpecification() const {
    return *spec;
}

const std::unordered_map<Point, std::list<Point>> &FWIObservationOperatorSpace::GetSpaceSupport() const {
    return SpaceReceiver2SpaceSupport;
}

bool FWIObservationOperatorSpace::MeasureHere(const Point &receiver) const {
    return SpaceReceiver2SpaceSupport.contains(receiver);
}

int FWIObservationOperatorSpace::GetMeasureComp() const {
    return measureComp;
}

const std::unordered_map<Point, std::unordered_map<Point, RVector>> &FWIObservationOperatorSpace::GetPVals() const {
    return PVals;
}

const std::unordered_map<Point, std::unordered_map<Point, std::vector<VectorField>>> &
FWIObservationOperatorSpace::GetVFVals() const {
    return VFVals;
}
