//
// Created by cr on 10.11.22.
//

#ifndef FWI_FWIOBSERVATIONDGTSNEW_HPP
#define FWI_FWIOBSERVATIONDGTSNEW_HPP

#include <unordered_set>
#include <utility>
#include "FWISeismogram.hpp"
#include "MeasureMask.hpp"
#include "VectorField.hpp"
#include "FWIHelpers.hpp"
#include <memory>
#include "FWIMppHelpers.hpp"
#include "Observation/FWIObservationSpecification.hpp"

FWIObservationSpecification *GetObservationSpecificationWaypoints(double t0, double dt, double T,
                                                                  std::vector<Point> &waypoints,
                                                                  std::vector<int> &receiverCounts,
                                                                  bool include_last_waypoint);

class FWIObsSpecWPCreator {
    double t0 = 0.0;
    double dt = -1;
    double T = -1;
    int WaypointCount = 0;
    std::vector<int> receiverCounts{};
    std::vector<Point> waypoints{};
    bool includeLastWaypoint = true;

    void readConfig() {
        Config::Get("t0", t0);
        Config::Get("dt", dt);
        Config::Get("T", T);
        ReadWaypoints("Receiver", WaypointCount, waypoints, receiverCounts, includeLastWaypoint);
    }

public:
    FWIObsSpecWPCreator() {
        readConfig();
    }

    FWIObsSpecWPCreator WithT(double T_) {
        T = T_;
        return *this;
    }

    FWIObsSpecWPCreator Withdt(double dt_) {
        dt = dt_;
        return *this;
    }

    FWIObsSpecWPCreator Witht0(double t0_) {
        t0 = t0_;
        return *this;
    }

    FWIObsSpecWPCreator WithWaypoints(const std::vector<Point> &waypoints_) {
        waypoints = waypoints_;
        return *this;
    }

    FWIObsSpecWPCreator WithReceiverCounts(const std::vector<int> &receiverCounts_) {
        receiverCounts = receiverCounts_;
        return *this;
    }

    FWIObsSpecWPCreator WithReceiverCounts(bool includeLastWaypoint_) {
        includeLastWaypoint = includeLastWaypoint_;
        return *this;
    }

    FWIObservationSpecification *Create() {
        return GetObservationSpecificationWaypoints(t0, dt, T, waypoints, receiverCounts, includeLastWaypoint);
    }

    std::shared_ptr<FWIObservationSpecification> CreateShared() {
        return std::shared_ptr<FWIObservationSpecification>(Create());
    }

    std::unique_ptr<FWIObservationSpecification> CreateUnique() {
        return std::unique_ptr<FWIObservationSpecification>(Create());
    }
};

std::unique_ptr<MeasureMask> GetMeasureMaskNEW(int dim, int numL, int MeasureComp);


class FWIObservationOperatorTime {
    std::shared_ptr<FWIObservationSpecification> spec;
public:
    FWIObservationOperatorTime() = delete;

    explicit FWIObservationOperatorTime(std::shared_ptr<FWIObservationSpecification> spec_) : spec(std::move(spec_)) {}

    virtual std::list<int> TimeSupport(int) const = 0;

    virtual double operator()(double t) const = 0;

    virtual ~FWIObservationOperatorTime() = default;
};

class FWIObservationOperatorSpace {
protected:
    std::shared_ptr<FWIObservationSpecification> spec;
    const EleCacheType &EleCache;
    int measureComp = 0;
    int stateComps = -1;
    mutable std::unordered_map<Point, std::unordered_map<Point, std::vector<VectorField>>> VFVals;
    mutable std::unordered_map<Point, std::unordered_map<Point, RVector>> PVals;
    mutable std::unordered_map<Point, std::list<Point>> SpaceReceiver2SpaceSupport;
public:
    FWIObservationOperatorSpace() = delete;

    explicit FWIObservationOperatorSpace(const EleCacheType &EleCache_,
                                         std::shared_ptr<FWIObservationSpecification> spec_) : spec(std::move(spec_)),
                                                                                               EleCache(EleCache_),
                                                                                               stateComps(
                                                                                                       EleCache.begin()->second->getDim() +
                                                                                                       EleCache.begin()->second->p_dimension()) {}

    virtual void SetSpaceValue(const FWIState &Measure, const Point &Receiver,
                               const std::list<Point> &MeasureCells,
                               Vector &b) const = 0;

    const std::unordered_map<Point, std::list<Point>> &GetSpaceSupport() const;

    virtual bool MeasureHere(const Point &receiver) const;

    virtual FWIState GetSpaceValue(const Vector &, const Point &, const std::list<Point> &) const = 0;

    virtual ~FWIObservationOperatorSpace() = default;

    const FWIObservationSpecification &GetSpecification() const;

    const std::unordered_map<Point, std::unordered_map<Point, RVector>> &GetPVals() const;

    const std::unordered_map<Point, std::unordered_map<Point, std::vector<VectorField>>> &GetVFVals() const;

    int GetMeasureComp() const;

    virtual std::string Name() const = 0;
};

class FWIObservationOperatorTimePoint : public FWIObservationOperatorTime {
public:
    FWIObservationOperatorTimePoint() = delete;

    explicit FWIObservationOperatorTimePoint(std::shared_ptr<FWIObservationSpecification> spec_)
            : FWIObservationOperatorTime(std::move(spec_)) {}

    std::list<int> TimeSupport(int i) const override { return {i}; }

    double operator()(double t) const override {
        return 1.0;
    }


};

class FWIObservationOperatorSpacePoint : public FWIObservationOperatorSpace {
    std::unordered_map<Point, unsigned> ReceiverToCellCount;

    void initializeReceiverToCellCount(const FWIObservationSpecification &spec) {
        for (const auto &receiver: spec) {
            const unsigned cell_count = PPM->Sum(GetSpaceSupport().contains(receiver) ?
                                                 GetSpaceSupport().at(receiver).size() : 0);
            ReceiverToCellCount[receiver] = cell_count;
        }
    }

    void initalizeShapeValues(const FWIObservationSpecification &spec, const Mesh &b);

public:
    FWIObservationOperatorSpacePoint() = delete;

    FWIObservationOperatorSpacePoint(const EleCacheType &EleCache,
                                     std::shared_ptr<FWIObservationSpecification> spec_,
                                     const Mesh &mesh, int MeasureComp)
            : FWIObservationOperatorSpace(EleCache, std::move(spec_)) {
        measureComp = MeasureComp;
        initalizeShapeValues(*spec, mesh);
        initializeReceiverToCellCount(*spec);
    }

    std::string Name() const override { return "point"; }
    void SetSpaceValue(const FWIState &Measure, const Point &Receiver,
                       const std::list<Point> &MeasureCells,
                       Vector &b) const override;

    FWIState GetSpaceValue(const Vector &U, const Point &Receiver, const std::list<Point> &MeasureCells) const override;
};


class FWIObservationOperatorNEW {
    std::unique_ptr<FWIObservationOperatorTime> timeOp = nullptr;

    std::unique_ptr<FWIObservationOperatorSpace> spaceOp = nullptr;

    std::unique_ptr<MeasureMask> Mask = nullptr;

protected:
public:
    FWIObservationOperatorNEW(std::unique_ptr<FWIObservationOperatorTime> TimeOp_,
                              std::unique_ptr<FWIObservationOperatorSpace> SpaceOp_,
                              std::unique_ptr<MeasureMask> Mask_) :
            timeOp(std::move(TimeOp_)), spaceOp(std::move(SpaceOp_)), Mask(std::move(Mask_)) {}

    ~FWIObservationOperatorNEW() = default;

    void MeasureFullWave(const Vectors &U, FWISeismogram &s) const;

    void MeasureSingleWaveField(const Vector &U, FWISeismogram &s, int step) const;

    const FWIObservationOperatorSpace &GetSpaceOp() const;

    const FWIObservationOperatorTime &GetTimeOp() const;

    const MeasureMask &GetMask() const;

};


#endif //FWI_FWIOBSERVATIONDGTSNEW_HPP
