//
// Created by cr on 03.11.22.
//

#include "PMLHandler.hpp"

vector<double> initializePMInftyVec(int size) {
    vector<double> Box;
    std::generate_n(std::back_inserter(Box), size, [n = 0]() mutable { return (n++ % 2 == 0) ? infty : -infty; });
    return Box;
}

void PMLHandler::findBndFlags(const Mesh &mesh, const vector<double> &MeshBorders, vector<int> &BndFlags,
                              vector<int> &NrBndProcs) {
    int dim = mesh.dim();
    int BoxSize = 2 * dim;
    std::vector<int> bnds(2 * dim, -1);
    for (cell c = mesh.cells(); c != mesh.cells_end(); ++c) {
        for (int f = 0; f < c.Faces(); ++f) {
            const Point &fp = c.Face(f);
            int bnd = mesh.BoundaryFacePart(fp);
            if (bnd != -1) {
                for (int i = 0; i < dim; ++i) {
                    double d = abs(fp[i] - MeshBorders[i * dim]);
                    double d2 = abs(fp[i] - MeshBorders[i * dim + 1]);
                    if (d < Eps) {
                        bnds[i * dim] = bnd;
                        NrBndProcs[i * dim] = 1;
                        break;
                    } else if (d2 < Eps) {
                        bnds[i * dim + 1] = bnd;
                        NrBndProcs[i * dim + 1] = 1;
                        break;
                    }
                }
                std::vector<int> part;
                for (int j = 0; j < dim; ++j) {
                    bool tmp = MeshBorders[dim * j] <= c()[j];
                    bool tmp2 = MeshBorders[dim * j + 1] >= c()[j];
                    int ret = tmp - tmp2;
                    part.push_back(ret);
                }
            }
        }
    }
    PPM->Sum(NrBndProcs);
    std::vector<int> bndsRec(PPM->size() * BoxSize, 0);
    PPM->Allgather(bnds.data(), BoxSize, bndsRec.data());
    for (int i = 0; i < bndsRec.size(); ++i) {
        if (bndsRec[i] != -1) {
            BndFlags[i % BoxSize] = bndsRec[i];
        }
    }
}

void
PMLHandler::findCellNrAndPMLLength(const vector<double> &MeshBorders_, vector<int> &NrCells, vector<int> &nrBoundProcs,
                                   vector<double> &MeshBorderPML, const vector<int> &PMLBnd, int dim,
                                   std::vector<double> &dx) {
    std::vector<int> PaddingCells;
    for (int i = 0; i < dim; ++i) {
        for (int j = 0; j < 2; ++j) {
            double D = ConfigPML.GetPMLPaddingLength(i * dim + j)/ dx[i];
            int DR = round(D);
            if (abs(DR - D) > 1e-8) {
                THROW("Pick a PaddingLength that conforms with your Meshlength");
            }
            PaddingCells.push_back(DR);
        }
    }
    for (int i = 0; i < dim; ++i) {
        int pos = i * dim;
        double d0 = MeshBorders_[pos + 1] - MeshBorders_[pos];
        d0 += dx[i] * (PMLBnd[pos] * PaddingCells[pos] + PMLBnd[pos + 1] * PaddingCells[pos + 1]);
      MeshBorderPML[pos] = MeshBorders_[pos] - ConfigPML.GetPMLPaddingLength(i * dim);
      MeshBorderPML[pos + 1]  = MeshBorders_[pos+1] + ConfigPML.GetPMLPaddingLength(i * dim+1);
        NrCells[i] = (int) round((MeshBorderPML[pos + 1] - MeshBorderPML[pos])/dx[i]);
    }
}

void PMLHandler::findMeshAndProcBorders(const Mesh &mesh, vector<double> &MeshBorders_, vector<double> &ProcBorders_) {
    int dim = mesh.dim();
    int BoxSize = 2 * dim;
    vector<double> Box = initializePMInftyVec(BoxSize);
    for (cell c = mesh.cells(); c != mesh.cells_end(); ++c) {
        for (const auto &Cor: c.AsVector()) {
            for (int i = 0; i < dim; ++i) {
                Box[i * dim] = min(Box[i * dim], Cor[i]);
                Box[i * dim + 1] = max(Box[i * dim + 1], Cor[i]);
            }
        }
    }
    for (int i = 0; i < dim; ++i) {
        MeshBorders_[i * dim] = PPM->Min(Box[i * dim]);
        MeshBorders_[i * dim + 1] = PPM->Max(Box[i * dim + 1]);
        ProcBorders_[PPM->proc() * BoxSize + dim * i] = Box[i * dim];
        ProcBorders_[PPM->proc() * BoxSize + dim * i + 1] = Box[i * dim + 1];
    }
    PPM->Allgather(ProcBorders_.data() + PPM->proc() * BoxSize, BoxSize, ProcBorders_.data());
}

void
PMLHandler::initializePMLMesh(const Mesh &mesh, const vector<double> &MeshBordersI) {
    int dim = mesh.dim();
    std::vector<int> BndFlags(2 * dim, -1);
    std::vector<int> NrBndProcs(2 * dim, 0);
    findBndFlags(mesh, MeshBordersI, BndFlags, NrBndProcs);
    for (int i = 0; i < BndFlags.size(); ++i) {
        BndFlags[i] = (ConfigPML.pmlBndFlags[i] == 0) ? 0 : BndFlags[i];
    }
    vector<int> NrCells(dim, 0);
    vector<double> PMLLength(2 * dim, 0.0);
    auto meshWidth = mesh.MeshWidth();
    if (abs(meshWidth.first - meshWidth.second) > Eps) THROW("Only for quadratic meshes")
    dx = std::vector<double>(dim, meshWidth.first);
    findCellNrAndPMLLength(MeshBordersI, NrCells, NrBndProcs, PMLLength, ConfigPML.pmlBndFlags, dim, dx);
    auto PMLGeom = std::move(
            std::make_shared<RectangularGeometry2D>(std::pair<double, double>({PMLLength[0], PMLLength[1]}),
                                                    std::pair<double, double>({PMLLength[2], PMLLength[3]}),
                                                    NrCells[0] + 1, NrCells[1] + 1,
                                                    std::vector<int>{BndFlags[0], BndFlags[1], BndFlags[2],
                                                                     BndFlags[3]},
                                                    "Rect"));
    FWIDiscStackPML.meshes = std::move(MeshesCreator().WithCoarseGeometry(PMLGeom).WithOverlap(mesh.OverlapName()).
            WithDistribute(mesh.DistributionName()).WithPLevel(0).WithLevel(0).CreateShared());
}

void PMLHandler::initializeEleCache(const Vector &vec_, int numL_) {
  eleCacheContainerPML = std::make_unique<EleCacheContainer>(vec_, numL_);
}

void PMLHandler::fillDiscStack(int deg, int numL_, int paramComps) {
    const auto &meshes = *FWIDiscStackPML.meshes;
    FWIDiscStackPML.mD = std::make_unique<DGDiscretization>(meshes, 0, paramComps);
    meshes.PrintInfo();
    (*FWIDiscStackPML.mD)().PrintInfo();
    FWIDiscStackPML.mV = std::make_shared<Vector>(0.0, *FWIDiscStackPML.mD, FWIDiscStackPML.lev);
    FWIDiscStackPML.wD = std::make_unique<DGDiscretization>(meshes, deg, meshes.dim() + 1 + numL_);
    (*FWIDiscStackPML.mD)().PrintInfo();
    FWIDiscStackPML.wV = std::make_shared<Vector>(0.0, *FWIDiscStackPML.wD, FWIDiscStackPML.lev);
}

void PMLHandler::Initialize(const Mesh &mesh, PMLConf Conf, int deg, int numL_, int paramComps) {
    numL = numL_;
    ConfigPML = Conf;
    ProcBorders = initializePMInftyVec(2 * mesh.dim() * PPM->size());
    MeshBorders.resize(2 * mesh.dim());
    ProcBordersPML = initializePMInftyVec(2 * mesh.dim() * PPM->size());
    MeshBordersPML.resize(2 * mesh.dim());
    findMeshAndProcBorders(mesh, MeshBorders, ProcBorders);
    FWIDiscStackPML.lev = {0, -1};
    initializePMLMesh(mesh, MeshBorders);
    fillDiscStack(deg, numL, paramComps);
    initializeEleCache(GetWaveVector(), numL);
    findMeshAndProcBorders(GetWaveVector().GetMesh(), MeshBordersPML, ProcBordersPML);
    int max = PPM->Max(mesh.CellCount());
    int min = PPM->Min(mesh.CellCount());
    mout << "max: " << max << " " << "min: " << min << endl;
//    if (max != min) Exit("This only works for same sizes on all meshes")
//    InitializeCellToTransferNEW(PMLToNonPML,GetWaveVector().GetMesh(),FWIDiscStackPML.GetMesh(),ProcBorders);
//    InitializeCellToTransferNEW(NonPMLToPML,FWIDiscStackPML.GetMesh(),GetWaveVector().GetMesh(),ProcBordersPML);
    InitializeCellToTransferNEW(NonPMLToPML,mesh,FWIDiscStackPML.GetMesh(),ProcBordersPML);
    InitializeCellToTransferNEW(PMLToNonPML,FWIDiscStackPML.GetMesh(),mesh,ProcBorders);
//    InitializeCellToTransfer(PMLToNonPML, GetWaveVector().GetMesh(), ProcBorders);
//    InitializeCellToTransfer(NonPMLToPML, mesh, ProcBordersPML);
    vector<std::pair<Point, vector<Point>>> Cmids;
    CreatePointsOutsideOfInnerGrid(mesh, Cmids);
  InitializeBoundaryCellToTransferNEW(NonPMLBoundaryToPML, Cmids, GetProcBoundariesPML(),*FWIDiscStackPML.wV );
//  const vector<double> &Boundaries,
//  const Vector &meshfull)
//    InitializeBoundaryCellToTransfer(NonPMLBoundaryToPML, Cmids, GetProcBoundariesPML());
    initialized = true;


}

void PMLHandler::CreatePointsOutsideOfInnerGrid(const Mesh &VecMesh, vector<std::pair<Point, vector<Point>>> &Cmids) {
    int dim = VecMesh.dim();
    auto meshWidth = VecMesh.MeshWidth();
    for (cell c = VecMesh.cells(); c != VecMesh.cells_end(); ++c) {
        vector<Point> Cmid;
        vector<vector<double>> tests(dim);
        int nrbounds = 0;
        for (int f = 0; f < c.Faces(); ++f) {
            const Point &fp = c.Face(f);
            int bnd = VecMesh.BoundaryFacePart(fp);
            if (bnd != -1) {
                nrbounds++;
                int j = 0;
                int ret = 0;
                bool isOnLeftBnd = false;
                for (; j < dim; ++j) {
                    isOnLeftBnd = abs(fp[j] - MeshBorders[j * dim]) < Eps;
                    bool isOnRightBnd = abs(fp[j] - MeshBorders[j * dim + 1]) < Eps;
                    if (isOnLeftBnd || isOnRightBnd) break;
                }
                if (abs(MeshBordersPML[j * dim + (isOnLeftBnd ? 0 : 1)] -
                        MeshBorders[j * dim + (isOnLeftBnd ? 0 : 1)]) <
                    Eps)
                    continue;
                double PMLLastBoundaryCellMid = MeshBordersPML[j * dim + (isOnLeftBnd ? 0 : 1)] -
                                                (isOnLeftBnd ? -1 : 1) * meshWidth.first / 2.0;
                double PMLFirstBoundaryCellMid = fp[j] + 0.5 * (isOnLeftBnd ? -1 : 1) * meshWidth.first;
//                        std::cout << "p: " << PPM->proc() << " " << PMLLastBoundaryCellMid << " " << PMLFirstBoundaryCellMid << endl;
                int test = round(abs(PMLFirstBoundaryCellMid - PMLLastBoundaryCellMid) / meshWidth.first) + 1;
                std::vector<double> test2;
                if (test == 1) {
                    test2.push_back(PMLFirstBoundaryCellMid);
                } else {
                    test2 = ft::linspace(PMLFirstBoundaryCellMid, PMLLastBoundaryCellMid, test);
                }
                for (auto &t: test2) {
//                            MOUT(t);
                    Point P(c());
                    P[j] = t;
                    Cmid.push_back(P);
                }
                tests[j] = test2;
            }
        }
        addCornerPoints(tests, Cmid, dim, nrbounds);
        Cmids.push_back(std::pair<Point, vector<Point>>(Point(c()), Cmid));
    }
}

void PMLHandler::addCornerPoints(const vector<vector<double>> &outerBoundaryPoints, vector<Point> &Cmid, int dim,
                                 int nrbounds) {
    if (dim == 1 && nrbounds == 1) {
        for (auto x: outerBoundaryPoints[0]) {
            Cmid.push_back(Point(x, 0, 0, 0));
        }
    } else if (dim == 2 && nrbounds == 2) {
        for (auto x: outerBoundaryPoints[0]) {
            for (auto y: outerBoundaryPoints[1]) {
                Cmid.push_back(Point(x, y, 0, 0));
            }
        }
    } else if (dim == 3 && nrbounds == 3) {
        for (auto x: outerBoundaryPoints[0]) {
            for (auto &y: outerBoundaryPoints[1]) {
                for (auto &z: outerBoundaryPoints[2]) {
                    Cmid.push_back(Point(x, y, z, 0));
                }
            }
        }
    } else if (dim == 4 && nrbounds == 4) {
        for (auto x: outerBoundaryPoints[0]) {
            for (auto &y: outerBoundaryPoints[1]) {
                for (auto &z: outerBoundaryPoints[2]) {
                    for (auto &t: outerBoundaryPoints[3]) {
                        Cmid.push_back(Point(x, y, z, t));
                    }
                }
            }
        }
    }
}

int PMLHandler::isOnProc(const Point &P, const vector<double> &VectorOfBoxes) {
    int dim = (P.SpaceDim() + P.TimeDim());
    for (int i = 0; i < PPM->size(); ++i) {
        bool cont = true;
        for (int j = 0; j < dim; ++j) {
            bool tmp = VectorOfBoxes[i * 2 * dim + dim * j] <= P[j];
            bool tmp2 = VectorOfBoxes[i * 2 * dim + dim * j + 1] >= P[j];
            if (!tmp || !tmp2) cont = false;
        }
        if (cont) {
            return i;
        }
    }
    return -1;
}

void PMLHandler::InitializeBoundaryCellToTransfer(TransferToPMLBoundary &BondaryCellToTransfer,
                                                  vector<std::pair<Point, std::vector<Point>>> &ExtrapolatedPoints,
                                                  const vector<double> &Boundaries) {
    for (const auto &it: ExtrapolatedPoints) {
        for (const auto &P: it.second) {
            int p = isOnProc(P, Boundaries);
            if (p == -1) continue;
            else {
                BondaryCellToTransfer[it.first].push_back(std::pair<Point, short>(P, p));
            }
        }
    }
}
void
PMLHandler::InitializeBoundaryCellToTransferNEW(TransferToPMLBoundary &BondaryCellToTransfer,
                                                   vector<std::pair<Point, vector<Point>>> &ExtrapolatedPoints,
                                                   const vector<double> &Boundaries,
                                                   const Vector &meshPML) {
  auto* ExBuf = new ExchangeBuffer();
  vector<std::pair<Point, vector<Point>>> NotFound;
  std::unordered_set<Point> LostPoints;
  for (const auto &it: ExtrapolatedPoints) {
    for (const auto &P: it.second) {
      int p = isOnProc(P, Boundaries);
      if (p == -1) continue;
      bool found = meshPML.find_cell(P) != meshPML.cells_end();
      if (found) {
        BondaryCellToTransfer[it.first].emplace_back(P, PPM->Proc());
      } else {
        if(LostPoints.contains(P)){
          THROW("ERRROR!!");
        }
        LostPoints.emplace(P);
        ExBuf->Send(p) << PPM->Proc();
        ExBuf->Send(p) << it.first;
        ExBuf->Send(p) << P;
      }
    }
  }
  ExBuf->Communicate();
  std::unordered_set<Point> FoundPoints;
  for (int l = 0; l < PPM->Size(); ++l) {
    ExchangeBuffer FoundBuff;
    vector<CommData> LostData;
    for (short q1 = 0; q1 < PPM->Size(); ++q1) {
      while (ExBuf->Receive(q1).size() < ExBuf->ReceiveSize(q1)) {
        int BorderProc;
        ExBuf->Receive(q1) >> BorderProc;
        Point BorderPoint;
        ExBuf->Receive(q1) >> BorderPoint;
        Point PMLPoint;
        ExBuf->Receive(q1) >> PMLPoint;
        bool found = meshPML.find_cell(PMLPoint) != meshPML.cells_end();
        if (found) {
          FoundBuff.Send(BorderProc) << BorderPoint;
          FoundBuff.Send(BorderProc) << PMLPoint;
        } else {
          CommData dat{(PPM->Proc() + 1) % PPM->Size(),BorderProc,BorderPoint,PMLPoint};
          LostData.push_back(dat);
        }
      }
    }
    PPM->Barrier();
    ExBuf->ClearBuffers();
    ExBuf->Rewind();
    delete ExBuf;
    ExBuf = new ExchangeBuffer();
    FoundBuff.Communicate();
    for (short q = 0; q < PPM->Size(); ++q) {
      while (FoundBuff.Receive(q).size() < FoundBuff.ReceiveSize(q)) {
        Point BorderPoint;
        FoundBuff.Receive(q) >> BorderPoint;
        Point PMLPoint;
        FoundBuff.Receive(q) >> PMLPoint;
        if(!LostPoints.contains(PMLPoint)){
          THROW("ERROR!!" + to_string(PPM->Proc()));
        }else {
          LostPoints.erase(PMLPoint);
          BondaryCellToTransfer[BorderPoint].emplace_back(PMLPoint, q);
        }
      }
    }
    FoundBuff.ClearBuffers();
    FoundBuff.Rewind();

    for (const auto dat : LostData){
      ExBuf->Send(dat.sendProc) << dat.borderProc;
      ExBuf->Send(dat.sendProc) << dat.BorderPoint;
      ExBuf->Send(dat.sendProc) << dat.PMLPoint;
    }
    PPM->Barrier();

    ExBuf->Communicate();
  }
  PPM->Barrier();
  delete ExBuf;
  ExBuf = new ExchangeBuffer();
  for(const auto& P : BondaryCellToTransfer){
    Point BNDPoint = P.first;
    for(const auto &[PMLPoint,proc] : P.second ){
      ExBuf->Send(proc) << BNDPoint;
      ExBuf->Send(proc) << PMLPoint;
    }
  }
  ExBuf->Communicate();
  for (short q = 0; q < PPM->Size(); ++q) {
    while (ExBuf->Receive(q).size() < ExBuf->ReceiveSize(q)) {
      Point BorderPoint;
      ExBuf->Receive(q) >> BorderPoint;
      Point PMLPoint;
      ExBuf->Receive(q) >> PMLPoint;
      PMLToBoundary[PMLPoint] = BorderPoint;
    }
  }
}
void
PMLHandler::InitializeCellToTransfer(TransferPML &CellToTransfer, const Mesh &mesh, const vector<double> &Boundaries) {
    for (cell c = mesh.cells(); c != mesh.cells_end(); ++c) {
        int p = isOnProc(c(), Boundaries);
        if (p == -1) continue;
        else {
            CellToTransfer[c()] = std::make_pair<short, short>(PPM->proc(), p);
        }
    }
}
void
PMLHandler::InitializeCellToTransferNEW(TransferPML &CellToTransfer, const Mesh &meshFrom,
                                        const Mesh& meshTo,const vector<double> &Boundaries) {
  auto* ExBuf = new ExchangeBuffer();
  vector<std::pair<Point, vector<Point>>> NotFound;
  std::unordered_set<Point> LostPoints;

  for (cell c = meshFrom.cells(); c != meshFrom.cells_end(); ++c) {
    int p = isOnProc(c(), Boundaries);
    if (p == -1) continue;
    bool found = meshTo.find_cell(c()) != meshTo.cells_end();
    if (found)
    {
      CellToTransfer[c()] = std::make_pair<short, short>(PPM->Proc(),PPM->Proc());
    }else {
      if (LostPoints.contains(c())) {
        THROW("ERRROR!!");
      }
      LostPoints.emplace(c());
      ExBuf->Send(p) << PPM->Proc();;
      ExBuf->Send(p) << c();
    }
  }
  ExBuf->Communicate();
  std::unordered_set<Point> FoundPoints;
  for (int l = 0; l < PPM->Size(); ++l) {
    ExchangeBuffer FoundBuff;
    vector<CommDataRed> LostData;
    for (short q1 = 0; q1 < PPM->Size(); ++q1) {
      while (ExBuf->Receive(q1).size() < ExBuf->ReceiveSize(q1)) {
        int ownerCellProc;
        ExBuf->Receive(q1) >> ownerCellProc;
        Point CellPoint;
        ExBuf->Receive(q1) >> CellPoint;
        bool found = meshTo.find_cell(CellPoint) != meshTo.cells_end();
        if (found) {
          FoundBuff.Send(ownerCellProc) << CellPoint;
        } else {
          CommDataRed dat{(PPM->Proc() + 1) % PPM->Size(),ownerCellProc,CellPoint};
          LostData.push_back(dat);
        }
      }
    }
    PPM->Barrier();
    ExBuf->ClearBuffers();
    ExBuf->Rewind();
    delete ExBuf;
    ExBuf = new ExchangeBuffer();
    FoundBuff.Communicate();
    for (short q = 0; q < PPM->Size(); ++q) {
      while (FoundBuff.Receive(q).size() < FoundBuff.ReceiveSize(q)) {
        Point CellPoint;
        FoundBuff.Receive(q) >> CellPoint;
        if(!LostPoints.contains(CellPoint)){
          THROW("ERROR!!" + to_string(PPM->Proc()));
        }else {
          LostPoints.erase(CellPoint);
          CellToTransfer[CellPoint] = std::pair<short, short>(PPM->Proc(),q);
        }
      }
    }
    FoundBuff.ClearBuffers();
    FoundBuff.Rewind();

    for (const auto dat : LostData){
      ExBuf->Send(dat.sendProc) << dat.ownerCellProc;
      ExBuf->Send(dat.sendProc) << dat.CellPoint;
    }
    PPM->Barrier();

    ExBuf->Communicate();
  }
  PPM->Barrier();
  delete ExBuf;
}
const FWIDiscretizationStack &PMLHandler::GetDiscStack() const {
    return FWIDiscStackPML;
}

FWIDiscretizationStack &PMLHandler::GetDiscStack() {
    return FWIDiscStackPML;
}

const EleCacheType &PMLHandler::GetEleCache() const {
    return eleCacheContainerPML->Get();
}


bool PMLHandler::isPML() const {
    return initialized;
}

const Vector &PMLHandler::GetWaveVector() const {
    return *FWIDiscStackPML.wV;
}

const Vector &PMLHandler::GetMatVector() const {
    return *FWIDiscStackPML.mV;
}

const vector<double> &PMLHandler::GetProcBoundariesPML() const {
    return ProcBordersPML;
}

const vector<double> &PMLHandler::GetProcBoundaries() const {
    return ProcBorders;
}

const vector<double> &PMLHandler::GetMeshBordersPML() const {
    return MeshBordersPML;
}

const vector<double> &PMLHandler::GetMeshBorders() const {
    return MeshBorders;
}

const TransferPML &PMLHandler::GetPMLToNonPML() const {
    return PMLToNonPML;
}

const TransferPML &PMLHandler::GetNonPMLToPML() const {
    return NonPMLToPML;
}

const TransferToPMLBoundary &PMLHandler::GetNonPMLBoundaryToPML() const {
    return NonPMLBoundaryToPML;
}

PMLHandler::~PMLHandler() {
}

void PMLVectorFWI::transfer(const TransferPML &Transfer, const Vector &VecSend, Vector &VecRec) {
    ExchangeBuffer exBuffer;
    for (const auto &p: Transfer) {
        const auto &Trans = p.second;
        int i = VecSend.Id(p.first);
        exBuffer.Send(Trans.second) << p.first;
        for (int k = 0; k < VecSend.Dof(i); ++k) {
            exBuffer.Send(Trans.second) << VecSend(p.first, k);
        }
    }
    VecRec.CommunicateVector(exBuffer);
}

void PMLVectorFWI::PaddedToUnpadded(Vector &Vec) {
    transfer(handler.GetPMLToNonPML(), (*this), Vec);
}

void PMLVectorFWI::UnpaddedToPadded(const Vector &Vec) {
    transfer(handler.GetNonPMLToPML(), Vec, (*this));
}

void PMLMatVectorFWI::CommunicateBoundaryToPML(const Vector &VecSend) {
    ExchangeBuffer exBuffer;
    for (const auto &it: handler.GetNonPMLBoundaryToPML()) {
        const auto &SendPoint = it.first;
        for (const auto &P: it.second) {
            short receiveProc = P.second;
            const auto &ReceivePoint = P.first;
            if (ReceivePoint == SendPoint) continue;
            exBuffer.Send(receiveProc) << ReceivePoint;
            int i = VecSend.Id(SendPoint);
            for (int k = 0; k < VecSend.Dof(i); ++k) {
                exBuffer.Send(receiveProc) << VecSend(i, k);
            }
        }
    }
    (*this).CommunicateVector(exBuffer);
}

void PMLMatVectorFWI::increaseImpedance() {
    ExchangeBuffer exBuffer;
    for (const auto &it: handler.GetNonPMLBoundaryToPML()) {
        const auto &SendPoint = it.first;
        for (const auto &P: it.second) {
            short receiveProc = P.second;
            const auto &ReceivePoint = P.first;
            if (ReceivePoint == SendPoint) continue;
            int val1 = 0;
            int val2 = 0;
            double d = -infty;

            exBuffer.Send(receiveProc) << ReceivePoint;
//            exBuffer.Send(receiveProc) << distanceFactor(SendPoint, ReceivePoint);
            exBuffer.Send(receiveProc) << SendPoint;
        }
    }
    (*this).communicateImpedanceFactors(exBuffer);
}

void PMLMatVectorFWI::communicateImpedanceFactors(ExchangeBuffer &exBuffer) {
    exBuffer.Communicate();
    for (short q = 0; q < PPM->Size(CommSplit()); ++q) {
        while (exBuffer.Receive(q).size() < exBuffer.ReceiveSize(q)) {
            Point ReceivePoint;
            exBuffer.Receive(q) >> ReceivePoint;
            int i = Id(ReceivePoint);
            if constexpr (DebugLevel > 0) {
                if (exBuffer.Receive(q).size() + Dof(i) >= exBuffer.ReceiveSize(q)) {
                    Exit(to_string(PPM->Proc(CommSplit())) + ": Out of bounds.");
                }
            }
            Point SendPoint;
            exBuffer.Receive(q) >> SendPoint;

            applydistanceFactors(SendPoint,ReceivePoint,i);

        }
    }
    exBuffer.ClearBuffers();
}

PMLMatVectorFWI::PMLMatVectorFWI(const PMLHandler &handler, const Vector &Unpadded, bool increaseImp) :
        PMLVectorFWI(handler, handler.GetMatVector()) {
    UnpaddedToPadded(Unpadded);
    CommunicateBoundaryToPML(Unpadded);
    if (increaseImp) increaseImpedance();
}

void PMLMatVectorFWI::applydistanceFactors(const Point &SendPoint, const Point &ReceivePoint, int i) {
  if (handler.GetDistFactor() < Eps) return;
  double distX = ReceivePoint[0] - SendPoint[0];
  double distY = ReceivePoint[1] - SendPoint[1];
  if (handler.GetConfig().getTaperStyle() == OLDTAPER){
    double fac = OldTaper(max(abs(distY),abs(distX)));
    (*this)(i,0) *= fac;
    for (int n =0; n <= handler.GetNumL(); ++ n){
      (*this)(i,1+n) /= fac;
    }
    return;
  }
  if(handler.GetConfig().getTaperStyle() == COSTAPER){
    double relDist = calculateRelativeDistance(distX,distY);

    double deltad = handler.GetConfig().GetCosConfig().deltad;
    double vQ = handler.GetConfig().GetCosConfig().vQ;
    double vB = handler.GetConfig().GetCosConfig().vB;
    double C = relDist < deltad ? vQ +  (1-vQ) * cos(relDist/deltad *M_PI_2)
                                : vQ+(vB-vQ)*(relDist-deltad)/(1-deltad);

    if(handler.GetNumL() == 0){
      (*this)(i,0) /= C;
      (*this)(i,1) *= C;
    }else {
      double Kappa0 = (*this)(i, 1);
      double Kappa1 = (*this)(i, 2);
      double taup = Kappa1 / Kappa0;
      double alpha1 = handler.GetConfig().GetCosConfig().alpha1;
      double alpha2 = handler.GetConfig().GetCosConfig().alpha2;

      double QM = (1.0 + taup * alpha1) / (taup*alpha2);
      double Qb = handler.GetConfig().GetCosConfig().Qb;
      double QT = pow(QM,1.0 - relDist) * pow(Qb, relDist); // Exponential Taper from Q Value at
      double TauPML = 1.0/(alpha2*QT-alpha1);
      double quot1 = (1 + alpha1 * TauPML) / (1 + alpha1 * taup);
      double quot2 = (1.0 + handler.GetNumL() * TauPML) / (1.0 + handler.GetNumL() * taup);
      double RhoFac =  sqrt(quot1 / quot2);
      double KappaFac =1.0/ (  sqrt(quot1*quot2)); // these factors are supposed to ensure
      // the correct velocities as well as the correct impedance if the damping changes
      (*this)(i, 0) *= RhoFac* 1.0/sqrt(C);
      (*this)(i, 1) *= KappaFac * sqrt(C);
      for (int n = 1; n <= handler.GetNumL(); ++n) {
        (*this)(i, 1 + n) = (*this)(i, 1)  * TauPML;
      }
    }
  }
  else Exit("Taper Not known");
}

void
PMLHandlerAAO::InitializeBoundaryCellToTransferAAO(TransferToPMLBoundary &BondaryCellToTransfer,
                                                   vector<std::pair<Point, vector<Point>>> &ExtrapolatedPoints,
                                                   const vector<double> &Boundaries,
                                                   const Vector &meshfull) {
  auto* ExBuf = new ExchangeBuffer();
  vector<std::pair<Point, vector<Point>>> NotFound;
  std::unordered_set<Point> LostPoints;
  for (const auto &it: ExtrapolatedPoints) {
    for (const auto &P: it.second) {
      int p = isOnProc(P, Boundaries);
      bool found = meshfull.find_cell(P) != meshfull.cells_end();
      if (found) {
        BondaryCellToTransfer[it.first].emplace_back(P, PPM->Proc());
      } else {
        if(LostPoints.contains(P)){
          THROW("ERRROR!!");
        }
        LostPoints.emplace(P);
        ExBuf->Send(p) << PPM->Proc();
        ExBuf->Send(p) << it.first;
        ExBuf->Send(p) << P;
      }
    }
  }
  ExBuf->Communicate();
  std::unordered_set<Point> FoundPoints;
  for (int l = 0; l < PPM->Size(); ++l) {
    ExchangeBuffer FoundBuff;
    vector<CommData> LostData;
    for (short q1 = 0; q1 < PPM->Size(); ++q1) {
      while (ExBuf->Receive(q1).size() < ExBuf->ReceiveSize(q1)) {
        int BorderProc;
        ExBuf->Receive(q1) >> BorderProc;
        Point BorderPoint;
        ExBuf->Receive(q1) >> BorderPoint;
        Point PMLPoint;
        ExBuf->Receive(q1) >> PMLPoint;
          bool found = meshfull.find_cell(PMLPoint) != meshfull.cells_end();

          if (found) {
          FoundBuff.Send(BorderProc) << BorderPoint;
          FoundBuff.Send(BorderProc) << PMLPoint;
        } else {
          CommData dat{(PPM->Proc() + 1) % PPM->Size(),BorderProc,BorderPoint,PMLPoint};
          LostData.push_back(dat);
        }
      }
    }
    PPM->Barrier();
    ExBuf->ClearBuffers();
    ExBuf->Rewind();
    delete ExBuf;
    ExBuf = new ExchangeBuffer();
    FoundBuff.Communicate();
    for (short q = 0; q < PPM->Size(); ++q) {
      while (FoundBuff.Receive(q).size() < FoundBuff.ReceiveSize(q)) {
        Point BorderPoint;
        FoundBuff.Receive(q) >> BorderPoint;
        Point PMLPoint;
        FoundBuff.Receive(q) >> PMLPoint;
        if(!LostPoints.contains(PMLPoint)){
          THROW("ERROR!!" + to_string(PPM->Proc()));
        }else {
          LostPoints.erase(PMLPoint);
          BondaryCellToTransfer[BorderPoint].emplace_back(PMLPoint, q);
        }
      }
    }
    FoundBuff.ClearBuffers();
    FoundBuff.Rewind();

    for (const auto dat : LostData){
      ExBuf->Send(dat.sendProc) << dat.borderProc;
      ExBuf->Send(dat.sendProc) << dat.BorderPoint;
      ExBuf->Send(dat.sendProc) << dat.PMLPoint;
    }
    PPM->Barrier();

    ExBuf->Communicate();
  }
  PPM->Barrier();
  delete ExBuf;

}

PMLHandlerAAO::PMLHandlerAAO(const AAOPMLConfig &Conf, const Vector &vecfull) {
  lev = vecfull.Level();
  DistFactor = Conf.distfactor;
  ProcBordersPML = initializePMInftyVec(2 * vecfull.GetMesh().dim() * PPM->size());
  MeshBordersPML.resize(2 * vecfull.GetMesh().dim());
  MeshBorders.resize(2 * vecfull.GetMesh().dim());
  MeshBorders[0] = Conf.RectConf.LeftBnd;
  MeshBorders[1] = Conf.RectConf.RightBnd;
  MeshBorders[2] = Conf.RectConf.BottomBnd;
  MeshBorders[3] = Conf.RectConf.TopBnd;
  auto meshWPML = vecfull.GetMesh().MeshWidth();
  dx = std::vector<double>( vecfull.GetMesh().dim(), meshWPML.first);
   std::unordered_set<Point> PMLPoints;
  PMLHandler::findMeshAndProcBorders(vecfull.GetMesh(),MeshBordersPML,ProcBordersPML);
  vector<std::pair<Point, vector<Point>>> Cmids;
  struct position{
    bool left=false;
    bool right=false;
    bool bottom=false;
    bool top=false;
    bool isBnd() const {
      return left || right || bottom || top;
    }
    bool Corner() const {
      return (left && bottom) || (right && bottom) || (right && top) || (left && top);
    }
  };
  for (cell c = vecfull.cells(); c!=vecfull.cells_end();++c){
    position pos;
    if ((abs(c()[0] - dx[0]/2.0 - Conf.RectConf.LeftBnd)<Eps)){
      if(c()[1] >Conf.RectConf.BottomBnd &&c()[1] < Conf.RectConf.TopBnd){
        pos.left = true;
      }
    }
    if ((abs(c()[0] + dx[0]/2.0 - Conf.RectConf.RightBnd)<Eps)){
      if(c()[1] >Conf.RectConf.BottomBnd &&c()[1] < Conf.RectConf.TopBnd){
        pos.right = true;
      }
    }
    if ((abs(c()[1] - dx[0]/2.0 - Conf.RectConf.BottomBnd)<Eps)){
      if(c()[0] >Conf.RectConf.LeftBnd &&c()[0] < Conf.RectConf.RightBnd){
        pos.bottom = true;
      }
    }
    if ((abs(c()[1] + dx[0]/2.0 - Conf.RectConf.TopBnd)<Eps)){
      if(c()[0] >Conf.RectConf.LeftBnd &&c()[0] < Conf.RectConf.RightBnd){
        pos.top = true;
      }
    }
    if (pos.isBnd()){
      int left =0;
      int right =0;
      int bottom = 0;
      int top = 0;
      std::pair<Point, vector<Point>> tmp;
      tmp.first = c();
      if(pos.left){
        int i = 1;
        while (c()[0]-i*dx[0] > MeshBordersPML[0]) {
          tmp.second.emplace_back(c()[0] - i*dx[0], c()[1]);
          ++i;
        }
        left = i-1;
      }
      if(pos.right){
        int i = 1;
        while (c()[0]+i*dx[0] < MeshBordersPML[1]) {
          tmp.second.emplace_back(c()[0] + i*dx[0], c()[1]);
          ++i;
        }
        right = i-1;
      }
      if(pos.bottom){
        int i = 1;
        while (c()[1]-i*dx[0] > MeshBordersPML[2]) {
          tmp.second.emplace_back(c()[0], c()[1] - i*dx[0]);
          ++i;
        }
        bottom = i-1;
      }
      if(pos.top){
        int i = 1;
        while (c()[1]+i*dx[0] < MeshBordersPML[3]) {
          tmp.second.emplace_back(c()[0], c()[1] + i*dx[0]);
          ++i;
        }
        top = i-1;
      }
      if(pos.Corner()){
        int i = pos.left ? left : right;
        int j = pos.bottom ? bottom : top;
        for (int k = 0; k<i; ++k){
          for (int l=0; l<j;++l) {
            tmp.second.emplace_back(tmp.second[k][0],tmp.second[i+l][1]);
          }
        }
      }
      Cmids.push_back(tmp);
    }
  }
  InitializeBoundaryCellToTransferAAO(NonPMLBoundaryToPML, Cmids, ProcBordersPML,vecfull);
}

