//
// Created by cr on 07.10.22.
//

#include <fftw3.h>
#include "FWISourceWavelet.hpp"
#include "RungeKutta.hpp"
#include "CVector.hpp"
#include "FFT.hpp"

std::vector<double> getEvaluationPoints(int rkorder) {
  if (rkorder == -1002 || rkorder == 1002 || rkorder == -2) {
    return {0.5};
  } else if (rkorder == -302) {
    return {0, 1};
  } else {
    auto type = static_cast<TIMEINTEGRATOR>(rkorder);
    auto TI = TimeIntegratorCreator(type).WithLinearSolver(nullptr).WithLinearSolver(nullptr).
        WithLinearSolver(nullptr).WithLinearSolver(nullptr).Create();
    auto rk = dynamic_cast<RungeKutta *>(TI);
    auto c = rk->GetButcherTableau().c.asVector();
    delete TI;
    return c;
  }
}

CVector SourceWaveletBase::GetFFT() const {
  CVector SeismoDataOutput(0.0, evalsOnGrid.size());
  FFT::RealToComplexVector(evalsOnGrid, SeismoDataOutput);
  return SeismoDataOutput;
}
void SourceWaveletBase::Bandpass(double centerFreq, double width, int order){
  int numSamples = evalsOnGrid.size();
  auto f = std::make_unique<Dsp::SimpleFilter<Dsp::Butterworth::BandPass<4>, 1>>();
  f->setup(order, 1.0 / dt, centerFreq,width);
  double *seismoData[1];
  seismoData[0] = evalsOnGrid.asVector().data();
  f->process(numSamples, seismoData);
  seismoData[0] = new double[evalsOnGrid.size()];
  for (int i = 0; i < evalsOnGrid.size(); ++i) {
    seismoData[0][i] = evalsOnGrid[evalsOnGrid.size() - 1 - i];
  }
  f->process(numSamples, seismoData);
  for (int i = 0; i < evalsOnGrid.size(); ++i) {
    evalsOnGrid[evalsOnGrid.size() - 1 - i] = seismoData[0][i];
  }
  SplineRepresent = tk::spline();
  SplineRepresent.set_boundary(tk::spline::second_deriv, 0.0, tk::spline::second_deriv, 0.0);
  SplineRepresent.set_points(timeGrid.asVector(), evalsOnGrid.asVector());
  delete[] seismoData[0];
}
void SourceWaveletBase::BandpassIFOS(double lowerFreq, double upperFreq){
  double fc = lowerFreq;
  double F_HP = upperFreq;
  int NT = evalsOnGrid.size();
  int Npad = (int)(pow(2.0, ceil(log((double)(NT))/log(2.0))+1.0) );
  double df = (double) 1.0/(Npad)/dt;
  vector<double> ff(Npad+1),BFabs(Npad+1);
  int j =0;
  for (j=0;j<Npad/2;j++){
    ff[j+1] = (double) j * df;
  }
  for (j=Npad/2;j<Npad;j++){
    ff[j+1] = -(double) (Npad-j) * df;
  }
  int it1=floor(1/df);
  int it2=round(F_HP/df);
  int it3=floor(fc/df);
  int it4 = 2*it3;
  int flank1 = it2 - it1;
  int flank2 = it4-it3;
  if(PPM->Proc()==0) {
    printf("    Band-pass corner indexes    : it1 = %d, it2 = %d, it3 = %d, it4 = %d \n",
           it1,it2,it3,it4);
    printf("    Band-pass corner frequencies: Fc1 = %.2f, Fc2 = %.2f, Fc3 = %.2f, Fc4 = %.2f \n",
           ff[it1],ff[it2],ff[it3],ff[it4]);
  }
  BFabs[0] = 0.0;
  for (int n=0; n<=Npad; n++) {
    if ((n>=it1) && (n<it3))	BFabs[n] = 1.0;
    if ((n>=it3) && (n<it4))	BFabs[n] = (double) (1.0-(1-cos((double)(n-it3)/flank2*M_PI))/2.0);
    if(n>=it4) BFabs[n] = 0.0; // BFabs[n];
    if(F_HP) {
      if (n<=it1) BFabs[n] = 0.0;
      if ((n>it1) && (n<it2))	BFabs[n] = (double) ((1.0-cos((double)(n-it1)/flank1*M_PI))/2.0);
      if ((n>=it2) && (n<it3))	BFabs[n] = 1.0;
    }
  }
  std::vector<std::complex<double>> in_data(Npad,0.0);
  std::vector<std::complex<double>> out_data(Npad,0.0);
  for(j=0;j<Npad;j++){
    if(j<NT-1){
      in_data[j].real(evalsOnGrid[j+1]);
      in_data[j].imag(0.0);
      out_data[j].real(0.0);
      out_data[j].imag(0.0);
    }else{
      in_data[j].real(0.0);
      in_data[j].imag(0.0);
      out_data[j].real(0.0);
      out_data[j].imag(0.0);
    }
  }
  fftw_plan p_data;
  fftw_plan p_data2;

  p_data = fftw_plan_dft_1d(Npad, reinterpret_cast<fftw_complex *>(in_data.data()),
                            reinterpret_cast<fftw_complex *>(out_data.data()), 1, FFTW_ESTIMATE);
  fftw_execute(p_data);
  for(j=1;j<=Npad/2;j++) {
    out_data[j] = out_data[j] * BFabs[j];
    if(j!= 1) {
      out_data[Npad + 1 - j] = conj(out_data[j]) * BFabs[j];
    }
    in_data[j].imag(0.0);
    in_data[j].real(0.0);
  }
  out_data[0].real(0.0);
  out_data[0].imag(0.0);
  out_data[1].real(0.0);
  out_data[1].imag(0.0);
  fftw_destroy_plan(p_data);
  p_data2  = fftw_plan_dft_1d(Npad, reinterpret_cast<fftw_complex *>(out_data.data()),
                             reinterpret_cast<fftw_complex *>(in_data.data()), -1, FFTW_ESTIMATE);
  fftw_execute(p_data2);
  for(j=1;j<evalsOnGrid.size();j++){
    evalsOnGrid[j]=in_data[j-1].real()/Npad;
  }
  fftw_destroy_plan(p_data2);

}
void SourceWaveletBase::Lowpass(double freq, int order) {
  int numSamples = evalsOnGrid.size();
  auto f = std::make_unique<Dsp::SimpleFilter<Dsp::Butterworth::LowPass<4>, 1>>();
  f->setup(order, 1.0 / dt, freq);
  double *seismoData[1];
  seismoData[0] = evalsOnGrid.asVector().data();
  f->process(numSamples, seismoData);
  seismoData[0] = new double[evalsOnGrid.size()];
  for (int i = 0; i < evalsOnGrid.size(); ++i) {
    seismoData[0][i] = evalsOnGrid[evalsOnGrid.size() - 1 - i];
  }
  f->process(numSamples, seismoData);
  for (int i = 0; i < evalsOnGrid.size(); ++i) {
    evalsOnGrid[evalsOnGrid.size() - 1 - i] = seismoData[0][i];
  }
  SplineRepresent = tk::spline();
  SplineRepresent.set_boundary(tk::spline::second_deriv, 0.0, tk::spline::second_deriv, 0.0);
  SplineRepresent.set_points(timeGrid.asVector(), evalsOnGrid.asVector());
  delete[] seismoData[0];
}

double SourceWaveletBase::operator()(double t) const {
  return SplineRepresent(t);
  THROW("TimeEvaluations do not match")
}

SourceWaveletFromF::SourceWaveletFromF(const std::function<double(double)> &f, double dt_, double T,
                                       double signalFreq_,
                                       int rkorder) : SourceWaveletBase() {
  signalFreq = signalFreq_;
  dt = dt_;
  double t = 0;
  int i = 1;
  evalPoints = getEvaluationPoints(rkorder);
  for (int j = 1; j < evalPoints.size(); ++j) {
    if (mpp_ba::isNear(evalPoints[j], 1))
      evalPoints.erase(evalPoints.begin() + j);
  }
  if (evalPoints.size() == 1) evalsOnGrid.push_back(0.0);
  if (evalPoints.size() == 1) timeGrid.push_back(0.0);
  std::vector<double> x(1,0.0);
  std::vector<double> y(1,0.0);
  while (t < T) {
    t = i * dt;
    if (evalPoints.size() > 1) timeGrid.push_back(t);
    if (evalPoints.size() > 1) evalsOnGrid.push_back(f(t));
    for (double eval: evalPoints) {
      if (evalPoints.size() == 1) timeGrid.push_back(t + eval * dt);
      if (evalPoints.size() == 1) evalsOnGrid.push_back(f(t + eval * dt));
      x.push_back(t + eval * dt);
      y.push_back(f(t + eval * dt));
    }
    i = i + 1;
  }
  SplineRepresent = tk::spline();
  SplineRepresent.set_boundary(tk::spline::second_deriv, 0.0, tk::spline::second_deriv, 0.0);
  SplineRepresent.set_points(x, y);
}


void SourceWaveletBase::WriteToFile(const string &fn) const {
  if (PPM->Proc() == 0) {
    std::ofstream file;
    file.open(fn.c_str());
    file << std::fixed << std::setprecision(15);
    mout << "Writing " + GetName() + " Signal to: " << fn << endl;
//  string head = "\"Time\";\"Value\"; \n";
    string head;
    file << head;
    for (int i = 0; i < timeGrid.size(); ++i) {
      file << i * dt;
      file << ",";
      file << evalsOnGrid[i];
      file << "\n";

    }
    file.close();
  }
}

const RVector &SourceWaveletBase::GetEvalsOnGrid() const {
  return evalsOnGrid;
}

void SourceWaveletBase::SetEvalsOnGrid(const RVector &evalsOnGrid_) {
  if (SourceWaveletBase::evalsOnGrid.size() != evalsOnGrid_.size()){
    Exit("Sizes do not fit");
  }
  evalsOnGrid = evalsOnGrid_;
  SplineRepresent = tk::spline();
  SplineRepresent.set_boundary(tk::spline::second_deriv,0.0,tk::spline::second_deriv,0.0);
  SplineRepresent.set_points(timeGrid.asVector(),evalsOnGrid.asVector(),tk::spline::cspline);
}

std::unique_ptr<SourceWaveletFromF>
GetSourceWavelet(const TimeConf &TConf, const SourceConf &SConf) {
  std::unique_ptr<SourceWaveletFromF> wavelet;
  if (SConf.SignalName == "RickerDerivative") {
    wavelet = std::make_unique<RickerDerivativeWavelet>(SConf.SignalFreq, SConf.SignalDelay,
                                                        SConf.SignalCutoff,
                                                        TConf.dt, TConf.T, TConf.rkorder);
  } else if (SConf.SignalName == "Ricker") {
    wavelet = std::make_unique<RickerWavelet>(SConf.SignalFreq, SConf.SignalDelay,
                                              SConf.SignalCutoff,
                                              TConf.dt, TConf.T, TConf.rkorder);
  }else if (SConf.SignalName == "Sin") {
    double internFac = 1.0;
    Config::Get("SininternFac", internFac);
    wavelet = std::make_unique<SinWavelet>(SConf.SignalFreq, SConf.SignalDelay,internFac,
                                                TConf.dt, TConf.T, TConf.rkorder);
  }else if (SConf.SignalName == "SinIntegral") {
    double internFac = 1.0;
    Config::Get("SinIntegralinternFac", internFac);
    wavelet = std::make_unique<SinWaveletIntegral>(SConf.SignalFreq, SConf.SignalDelay,internFac,
                                           TConf.dt, TConf.T, TConf.rkorder);
  }else if (SConf.SignalName == "Cos") {
    double internFac = 1.0;
    Config::Get("CosinternFac", internFac);
    wavelet = std::make_unique<CosWavelet>(SConf.SignalFreq, SConf.SignalDelay,internFac,
                                           TConf.dt, TConf.T, TConf.rkorder);
  } else {
    Exit("Unknown Source Signal");
  }
  if (SConf.FrequencyFiltering) {
    (*wavelet).WriteToFile(Config::GetDataPath() + SConf.SignalName);
    if (SConf.getFreq()[0] > 1000) {
      MOUT("Stage with Frequency > 1000 detected. No Frequency filtering applied")
    } else if (SConf.SrcInversion) {
      MOUT("Not Filtering Source for SourceInversion")
    }else{
      if(SConf.FilteringType == "LowPass") {
        (*wavelet).Lowpass(SConf.getFreq()[0], SConf.FilteringOrder);
      }else if (SConf.FilteringType == "BandPass"){
//        (*wavelet).Bandpass(SConf.getFreq()[0],SConf.getFreq()[1],SConf.FilteringOrder);
        (*wavelet).BandpassIFOS(SConf.getFreq()[0],SConf.getFreq()[1]);
      }
    }
    string suffix;
      for (double d : SConf.getFreq()) {
        suffix += std::to_string((int) (d * 100));
      }
    (*wavelet).WriteToFile(
        Config::GetDataPath() + SConf.SignalName + std::to_string(SConf.SignalFreq) + SConf.FilteringType + "Filtered"
        + suffix + "Order" +
        std::to_string(SConf.FilteringOrder));

  }
  return std::move(wavelet);
}
double sinDelayIntegral(double signalFreq, double t, double delay,double internFac) {
  if (t < delay ) return 0.0;
  if (t > delay + 1.0/signalFreq ) return 0.0;
  return internFac * (cos(3.0*Pi*signalFreq*(t-delay))-9.0*cos(Pi*signalFreq*(t-delay)))
  /(12*Pi*signalFreq);
}
double sinDelay(double signalFreq, double t, double delay,double internFac) {
  if (t < delay ) return 0.0;
  if (t > delay + 1.0/signalFreq ) return 0.0;
  return internFac * pow(sin(signalFreq * Pi * (t-delay)),3);
}
double CosDelay(double signalFreq,double t, double delay,double internFac){
  if (t < delay ) return 0.0;
  if (t > delay + 1.0/signalFreq ) return 0.0;
  return internFac* 3*Pi*signalFreq * pow(sin(signalFreq * Pi * (t-delay)),2)*cos(signalFreq * Pi * (t-delay));
}
