#ifndef FWI_FWISOURCES_HPP
#define FWI_FWISOURCES_HPP

#include "../../../../DSPFilters/Dsp.h"
#include "FWIHelpers.hpp"
#include "CVector.hpp"
#include <fstream>
#include <functional>
#include "spline.h"
class SourceWaveletBase {
protected:
  double dt = 0.0;
  RVector evalsOnGrid;
public:
  void SetEvalsOnGrid(const RVector &evalsOnGrid_) ;

public:
  const RVector &GetEvalsOnGrid() const;

protected:
  RVector timeGrid;
  tk::spline SplineRepresent;
  double signalFreq=-1.0;
public:
  SourceWaveletBase() = default;

  SourceWaveletBase(const SourceWaveletBase& wav){
    dt = wav.dt;
    evalsOnGrid = wav.evalsOnGrid;
    timeGrid = wav.timeGrid;
    SplineRepresent =wav.SplineRepresent;
    signalFreq=wav.signalFreq;

  }
  void Bandpass(double centerFreq, double width, int order);

  void BandpassIFOS(double,double);

  void Lowpass(double freq, int order);

  CVector GetFFT() const;

  double operator()  (double t) const;

  virtual std::string GetName() const{ return "";};

  void WriteToFile(const string& fn) const;
  double GetSignalFreq() const {
    if(signalFreq == -1){
      Exit("No Dominant Frequency set.")
    }
    return signalFreq;
  }
  virtual ~SourceWaveletBase()  = default;
};

class SourceFromFile : public SourceWaveletBase {
public:
  explicit SourceFromFile(const std::string &fn,tk::spline::spline_type type = tk::spline::cspline ) {
    int cnt =0;
    std::ifstream file(fn);
    if (!file.good()) {
      Exit("No text-file found at: " + fn)
    }
    std::string line;
    while (getline(file, line)) {
      trim(line);
      if (line.length() == 0) {
        continue;
      }
      size_t equalPos = line.find(',');
      if (equalPos == string::npos) {
        continue;
      }
      string x = line.substr(0, equalPos);
      string y = line.substr(equalPos + 1);
      timeGrid.push_back(stod(x));
      evalsOnGrid.push_back(stod(y));
      if (cnt==1) {
        dt = timeGrid[1] - timeGrid[0];
      } else {
        if (!mpp_ba::isNear(dt, timeGrid.back() - timeGrid[timeGrid.size() - 2])) {
          Exit("not equidistant grid");
        }
    }
      cnt++;
  }
  SplineRepresent = tk::spline();
  SplineRepresent.set_boundary(tk::spline::second_deriv,0.0,tk::spline::second_deriv,0.0);
  SplineRepresent.set_points(timeGrid.asVector(),evalsOnGrid.asVector(),type);
  }
  std::string GetName()const override { return "SourceFromFile"; }
};

class SourceWaveletFromF : public SourceWaveletBase {
  std::vector<double> evalPoints;
public:
  SourceWaveletFromF(const std::function<double(double)> &, double, double, double, int);

  SourceWaveletFromF(const SourceWaveletFromF& wav) : SourceWaveletBase(wav) {
    dt = wav.dt;
    evalsOnGrid = wav.evalsOnGrid;
    timeGrid = wav.timeGrid;
    SplineRepresent =wav.SplineRepresent;
    signalFreq=wav.signalFreq;
    evalPoints = wav.evalPoints;
  }

};

class RickerDerivativeWavelet : public SourceWaveletFromF {
public:

  RickerDerivativeWavelet(double signalFreq, double delay, double cutoff, double dt, double T,
                          int rkorder) :
      SourceWaveletFromF(
          std::bind(std::function<double(double, double, double, double)>(rickerDerivativeDelay),
                    signalFreq,
                    std::placeholders::_1, delay, cutoff), dt, T, signalFreq, rkorder) {
  }
  std::string GetName() const override { return "RickerDerivative"; }
};

class RickerWavelet : public SourceWaveletFromF {
public:
  RickerWavelet(double signalFreq, double delay, double cutoff, double dt, double T, int rkorder) :
      SourceWaveletFromF(
          std::bind(std::function<double(double, double, double, double)>(rickerDelay), signalFreq,
                    std::placeholders::_1, delay, cutoff), dt, T, signalFreq, rkorder) {
  }
  std::string GetName() const override { return "Ricker"; }
};
double sinDelayIntegral(double signalFreq, double t, double delay,double internFac);
class SinWaveletIntegral : public SourceWaveletFromF {
public:
  SinWaveletIntegral(double signalFreq, double delay, double internFac,double dt, double T, int rkorder) :
      SourceWaveletFromF(
          std::bind(std::function<double(double, double, double,double)>(sinDelayIntegral), signalFreq,
                    std::placeholders::_1, delay,internFac), dt, T, signalFreq, rkorder) {
  }
  std::string GetName() const override { return "SinIntegral"; }
};
double sinDelay(double signalFreq,double t, double delay,double internFac);

class SinWavelet : public SourceWaveletFromF {
public:
  SinWavelet(double signalFreq, double delay, double internFac,double dt, double T, int rkorder) :
      SourceWaveletFromF(
          std::bind(std::function<double(double, double, double,double)>(sinDelay), signalFreq,
                    std::placeholders::_1, delay,internFac), dt, T, signalFreq, rkorder) {
  }
  std::string GetName() const override { return "Sin"; }
};
double CosDelay(double signalFreq,double t, double delay,double internFac);
class CosWavelet : public SourceWaveletFromF {
public:
  CosWavelet(double signalFreq, double delay, double internFac,double dt, double T, int rkorder) :
      SourceWaveletFromF(
          std::bind(std::function<double(double, double, double,double)>(CosDelay), signalFreq,
                    std::placeholders::_1, delay,internFac), dt, T, signalFreq, rkorder) {
  }
  std::string GetName() const override { return "Cos"; }
};
std::unique_ptr<SourceWaveletFromF> GetSourceWavelet(const TimeConf &TConf, const SourceConf &SConf);

#endif //FWI_FWISOURCES_HPP
