//
// Created by cr on 19.11.22.
//

#include "FWITestLinearized.hpp"
#include "TestEnvironment.hpp"
#include "FWIHelpers.hpp"
#include "Observation/FWIObservationDGTSNEW.hpp"
#include "Observation/FWISeismogram.hpp"
#include "Problems/FWIProblemsDGTSVA.hpp"
#include "FWIVectorFillFunctions.hpp"
#include <experimental/random>
#include "Problems/FWIInversionProblems.hpp"
#include "Observation/FWIObservationSpecification.hpp"
#include "Reduced/FWILinOpRedFirstOrderNormalized.hpp"
#include "FWIProblemWrapperVA.hpp"

struct TestParams {
  std::string model;
  std::string parametrization;
  int param;
  bool Measure;
  std::vector<double> values = std::vector<double>(6, 0.0);

};

class FWITestLinearized : public TestWithParam<TestParams> {
protected:
    std::unique_ptr<WPConf> Conf = nullptr;
    std::shared_ptr<Meshes> meshes;
    std::shared_ptr<WaveProblemDGTSVA> prob;
    std::unique_ptr<FWIDiscretizationStack> discStack;
    std::unique_ptr<EleCacheContainer> eleCache;
    std::shared_ptr<PMLHandler> PMLH;
    std::unique_ptr<DGVAcousticPlotter> plotter;
    std::unique_ptr<FWIObservationOperatorNEW> Obs;
    std::unique_ptr<FWISeismogram> s_m;

    std::unique_ptr<Vector> Mat;
    std::unique_ptr<Vectors> U_m;
    std::unique_ptr<FWILinOpRedFirstOrder> LinOp;
    const std::list<unsigned int> reconlist{0, 1, 2};
  std::unordered_map<Point,double> SeisMNorms;

public:
  FWITestLinearized() {
    Config::Initialize(TestFWILinearized);
    Conf = std::make_unique<WPConf>();
    meshes = getFWIMesh();
    prob = GetInversionProblem(*Conf);
    prob->setParametrization(GetParam().parametrization);
    int deg = 0;
    Config::Get("deg", deg);
    discStack = std::make_unique<FWIDiscretizationStack>(meshes, prob->getParamComps(),
                                                         prob->GetStateComponents(),
                                                         deg, LevelPair{Conf->level, -1});
    eleCache = std::make_unique<EleCacheContainer>(*discStack->wV, prob->getNumL());
    plotter = std::make_unique<DGVAcousticPlotter>(prob->getNumL(), &eleCache->Get());
    std::shared_ptr<FWIObservationSpecification> spec = FWIObsSpecWPCreator().Witht0(Conf->TConf.t0)
        .Withdt(Conf->TConf.dt).WithT(Conf->TConf.T).CreateShared();
    auto mask = GetMeasureMaskNEW(meshes->dim(), prob->getNumL(), 0);
    Obs = FWIObservationCreator(eleCache->Get(), discStack->GetMesh(),
                                prob->getNumL()).CreateUnique();
    PMLH = std::make_shared<PMLHandler>();
    Mat = std::make_unique<Vector>(*discStack->mV);
    Mat->Clear();
    FillVectorFromProblem(*prob, *Mat);
    U_m = std::make_unique<Vectors>(Conf->TConf.timeSteps + 1, *discStack->wV);
    s_m = std::make_unique<FWISeismogram>();
    prob->setShot(0);
    NonlinearEvaluation(*Mat,*U_m,*s_m);
    MakeLinearOp();

  }
  void MakeLinearOp(){
    if(GetParam().model == "L2") {
      LinOp = std::make_unique<FWILinOpRedFirstOrder>(*U_m, *Mat, prob, PMLH, Conf->TConf, eleCache->Get(),
                                            true);
    }else if (GetParam().model == "Normalized"){
      FWITimeSolver TSolveR(*discStack->wV, PMLH, Conf->TConf.rkorder);
      TSolveR.SolveAndMeasure(GetAssemble(*Mat), *Obs, *s_m, U_m.get());
      for (auto  rec : s_m->getSpecification()) {
        auto &ts = s_m->getTimeSeriesWrite(rec);
        double norm = sqrt(ts * ts);
        SeisMNorms[rec] = norm;
      }
      LinOp = std::make_unique<FWILinOpRedFirstOrderNormalized>(*U_m, *s_m, SeisMNorms,  *Mat,prob,
                                                      PMLH, Conf->TConf, eleCache->Get(), true);
    } else{
      Exit("not implemented")
    }
  }
  void NonlinearEvaluation(const Vector& Mat_,Vectors &uM, FWISeismogram& sM) {
    FWITimeSolver TSolveR(*discStack->wV, PMLH, Conf->TConf.rkorder);
    TSolveR.SolveAndMeasure(GetAssemble(Mat_), *Obs, sM, &uM);
    if(GetParam().model == "L2"){
    }else if(GetParam().model == "Normalized"){
      NormalizeTracewise(sM);
    } else{
      Exit("not implemented")
    }
  }
  std::shared_ptr<DGAssemble> GetAssemble(const Vector &Material) {
    auto Mp = std::make_unique<Vector>(*discStack->mV);
    Mp->Clear();
    ReparametrizeMaterialVector(Material, *prob, *Mp);
    Mp->Accumulate();
    auto WrappedProblem = std::make_shared<FWIProblemWrapperVA>(*prob, *prob, std::move(Mp),
                                                                prob->getNumL(), prob->getDim(),
                                                                prob);
    auto assemble = std::make_shared<DGAssemble>(WrappedProblem, true, -1, Conf->TConf.t0,
                                                 Conf->TConf.T,Conf->TConf.dt);
    assemble->SetElementCache(&eleCache->Get());
    return std::move(assemble);
  }

  void TearDown() override {
    Config::Close();
  }
};

TEST_P(FWITestLinearized, FWITestLinearized) {
  Vector Rand(*Mat);
  Rand.Clear();
  auto srcLoc = prob->getPointSrcLoc(0);
  std::experimental::reseed(1);
  for (cell c = Rand.cells(); c != Rand.cells_end(); ++c) {
    if (PointInCellDGTS(srcLoc, *c)) continue;
    auto r = double(std::experimental::randint(0, 10000) / 10000.0);
    Rand(c(), GetParam().param) = r;
  }
  Rand *= 1.0 / nc::L2MaterialNorm(Rand, eleCache->Get(), reconlist);
  Vector MatPlusRand(*Mat + Rand);
  Vectors U_Linearized(*U_m);
  FWISeismogram s_linearized;
  LinOp->SolveAndMeasure(Rand, *Obs, s_linearized, &U_Linearized);
  Vectors U_plusMat(*U_m);
  FWISeismogram s;
  std::vector<double> quotients(6, 0.0);
  for (int i = 0; i < 6; ++i) {
    double fac = pow(10, -i);
    MatPlusRand = *Mat;
    MatPlusRand += fac * Rand;
    U_plusMat.Clear();
    NonlinearEvaluation(MatPlusRand,U_plusMat,s);
    if (GetParam().Measure) {
      s -= *s_m;
      FWISeismogram tmp(s_linearized);
      tmp *= fac;
      s -= tmp;
      quotients[i] = sqrt(s * s) / fac;
    } else {
      U_plusMat -= *U_m;
      U_plusMat -= fac * U_Linearized;
      double DiffNorm = nc::L2TNorm(U_plusMat, eleCache->Get(), Conf->TConf.dt);
      quotients[i] = DiffNorm / fac;
    }
  }
  MOUT(quotients)
  EXPECT_VECTOR_NEAR(quotients, GetParam().values, 1e-6);
}

INSTANTIATE_TEST_SUITE_P(FWITestLinearized, FWITestLinearized, Values(
    TestParams{"L2","native", 0, false,
               {0.81292433, 0.14137084, 0.014445133, 0.0014468184, 0.00014470408, 1.4477827e-05}},
    TestParams{"L2","native", 1, false,
               {0.0024069847, 0.00024170005, 2.4177868e-05, 2.4178628e-06, 2.4204357e-07,
                1.1230511e-05}},
    TestParams{"L2","native", 2, false,
               {0.0013282641, 0.00013694223, 1.3736087e-05, 1.3740282e-06, 1.3798517e-07,
                1.1231312e-05}},
    TestParams{"L2","native", 0, true,
               {2.4134178, 0.437327, 0.044691432, 0.0044762024, 0.00044768816, 4.4825437e-05}},
    TestParams{"L2","native", 1, true,
               {0.0072688637, 0.00073003497, 7.3027954e-05, 7.3030315e-06, 7.3221897e-07,
                9.7866022e-06}},
    TestParams{"L2","native", 2, true,
               {0.0040252297, 0.00041493557, 4.1619436e-05, 4.1632044e-06, 4.1649661e-07,
                9.7930229e-06}},
    TestParams{"L2","taup", 0, false,
               {0.055221703, 0.0076169567, 0.00079051923, 7.9350764e-05, 7.938073e-06,
                1.0671103e-06}},
    TestParams{"L2","taup", 1, false,
               {0.64735807, 0.096746766, 0.0098028566, 0.00098119763, 9.8128494e-05,
                9.8150108e-06}},
    TestParams{"L2","taup", 2, false,
               {1.2780094, 0.40834277, 0.051416425, 0.0052760852, 0.00052898905, 5.2915044e-05}},
    TestParams{"L2","taup", 0, true,
               {0.11980079, 0.016743659, 0.0017392898, 0.00017460158, 1.7466902e-05,
                2.7439665e-06}},
    TestParams{"L2","taup", 1, true,
               {1.9308277, 0.29594415, 0.030001379, 0.0030029827, 0.00030032543, 3.0054824e-05}},
    TestParams{"L2","taup", 2, true,
               {3.1746183, 1.0703428, 0.13591181, 0.013957818, 0.0013995459, 0.00014000118}},
    TestParams{"L2","taup_positive", 0, false,
               {0.24567982, 0.026801928, 0.0027015788, 0.00027037717, 2.7039903e-05,
                2.7039059e-06}},
    TestParams{"L2","taup_positive", 1, false,
               {3.6357826, 2.1837059, 0.24447491, 0.024497232, 0.0024499759, 0.00024499983}},
    TestParams{"L2","taup_positive", 2, false,
               {0.027257068, 0.0025500014, 0.00025264943, 2.5241054e-05, 2.5240245e-06,
                9.0264204e-07}},
    TestParams{"L2","taup_positive", 0, true,
               {0.57603087, 0.063789159, 0.0064006444, 0.00064027369, 6.4029422e-05,
                6.4026776e-06}},
    TestParams{"L2","taup_positive", 1, true,
               {9.4890743, 6.6308463, 0.74867601, 0.07503201, 0.0075040352, 0.00075041093}},
    TestParams{"L2","taup_positive", 2, true,
               {0.065861634, 0.0059742311, 0.00058983981, 5.8907332e-05, 5.8929427e-06,
                3.4200817e-06}},
    TestParams{"Normalized","taup_positive", 0, false,
               {0.24567982,
                   0.026801928,
                   0.0027015788,
                   0.00027037717,
                   2.7039904e-05,
                   2.7039078e-06}},
    TestParams{"Normalized","taup_positive", 1, false,
               {3.6357826,
                   2.1837059,
                   0.24447491,
                   0.024497232,
                   0.0024499759,
                   0.00024499984}}, // only test for pressure wave velocity, since normalizing is a method that is just suited for that
    TestParams{"Normalized","taup", 1, false,
               {0.64735807,
                   0.096746766,
                   0.0098028566,
                   0.00098119763,
                   9.8128493e-05,
                   9.8149938e-06}} // only test for pressure wave velocity, since normalizing is a method that is just suited for that
));

int main(int argc, char **argv) {
  return MppTest(
      MppTestBuilder(argc, argv).
          WithoutDefaultConfig().
          WithScreenLogging().
          WithConfPath("../").
          WithGeoPath("../conf/geo/").
          WithPlotPath("data/vtu/").
          WithPPM()
  ).RUN_ALL_MPP_TESTS();
}