//
// Created by lionbart on 13.02.23.
//

#include "FWITestSecondDerivative.h"
#include "FWITestReduced1D.hpp"
#include "TestEnvironment.hpp"
#include "FWIHelpers.hpp"
#include "Observation/FWIObservationDGTSNEW.hpp"
#include "Observation/FWISeismogram.hpp"
#include "Problems/FWIProblemsDGTSVA.hpp"
#include "FWIVectorFillFunctions.hpp"
#include <experimental/random>
#include "Problems/FWIInversionProblems.hpp"
#include "Reduced/FWILinOpRedSecondOrder.h"
#include "Reduced/FWILinOpRedFirstOrder.hpp"
#include "FWIProblemWrapperVA.hpp"


struct TestParams {
    std::string parametrization;
    int param;
    bool Measure;
    std::vector<double> values = std::vector<double>(6, 0.0);
};


/**
 * @brief Checks Differential Quotient for second derivative.
 *
 * Calculate the Differential Quotient for the second derivative of the
 * Parameter to solution map.
 *  || \Phi'(p + p2) [p1] - \Phi'(p) [p1] - \Phi(p)[p1, p2] || * 1/||p||
 *  
 * @param param
 * @tparam template param
 * @ingroup ???
 */

class FWITestSecondDerivative : public TestWithParam<TestParams> {
protected:
    std::unique_ptr<WPConf> Conf = nullptr;
    std::shared_ptr<Meshes> meshes;
    std::shared_ptr<WaveProblemDGTSVA> prob;
    std::unique_ptr<FWIDiscretizationStack> discStack;
    std::unique_ptr<EleCacheContainer> eleCache;
    std::shared_ptr<PMLHandler> PMLH;
    std::unique_ptr<DGVAcousticPlotter> plotter;
    std::unique_ptr<FWIObservationOperatorNEW> Obs;
    std::unique_ptr<FWISeismogram> s_m;

    std::unique_ptr<Vector> Mat;
    std::unique_ptr<Vectors> U_m;
    std::unique_ptr<FWILinOpRedFirstOrder> LinOp;
    std::unique_ptr<FWILinOpRedSecondOrder> LinOpSO;
    const std::list<unsigned int> reconlist{0, 1};//!!! Vorher  0,1,2
public:
    FWITestSecondDerivative() {
        Config::Initialize(TestFWISecondDerivative); //Details hard coded in .hpp file
        Conf = std::make_unique<WPConf>();
        meshes = getFWIMesh();
        prob = GetInversionProblem(*Conf);
        prob->setParametrization(GetParam().parametrization);
        int deg = 0;
        Config::Get("deg", deg);
        discStack = std::make_unique<FWIDiscretizationStack>(meshes, prob->getParamComps(), prob->GetStateComponents(),
                                                             deg, LevelPair{Conf->level, -1});
        eleCache = std::make_unique<EleCacheContainer>(*discStack->wV, prob->getNumL());
        plotter = std::make_unique<DGVAcousticPlotter>(prob->getNumL(), &eleCache->Get());
        std::shared_ptr<FWIObservationSpecification> spec = FWIObsSpecWPCreator().Witht0(Conf->TConf.t0)
                .Withdt(Conf->TConf.dt).WithT(Conf->TConf.T).CreateShared();
        auto mask = GetMeasureMaskNEW(meshes->dim(), prob->getNumL(), 0);
        Obs = FWIObservationCreator(eleCache->Get(), discStack->GetMesh(), prob->getNumL()).CreateUnique();
        PMLH = std::make_shared<PMLHandler>();
        Mat = std::make_unique<Vector>(*discStack->mV);
        Mat->Clear();
        FillVectorFromProblem(*prob, *Mat);

//      Get Solutions of Forward Problem
        FWITimeSolver TSolveR(*discStack->wV, PMLH, Conf->TConf.rkorder);
        U_m = std::make_unique<Vectors>(Conf->TConf.timeSteps + 1, *discStack->wV);
        s_m = std::make_unique<FWISeismogram>();
        prob->setShot(0);
        TSolveR.SolveAndMeasure(GetAssemble(*Mat), *Obs, *s_m, U_m.get());

        LinOp = std::make_unique<FWILinOpRedFirstOrder>(*U_m, *Mat, prob,PMLH,
                                                        Conf->TConf, eleCache->Get(), false);
    }
    std::shared_ptr<DGAssemble> GetAssemble(const Vector &Material) {
        auto Mp = std::make_unique<Vector>(*discStack->mV);
        Mp->Clear();
        ReparametrizeMaterialVector(Material, *prob, *Mp);
        Mp->Accumulate();
        auto WrappedProblem = std::make_shared<FWIProblemWrapperVA>(*prob, *prob, std::move(Mp),
                                                                    prob->getNumL(), prob->getDim(), prob);
        auto assemble = std::make_shared<DGAssemble>(WrappedProblem, true, -1, Conf->TConf.t0,
                                                     Conf->TConf.T,
                                                     Conf->TConf.dt);
        assemble->SetElementCache(&eleCache->Get());
        return std::move(assemble);
    }

    void TearDown() override {
        Config::Close();
    }
};

TEST_P(FWITestSecondDerivative, FWITestSecondDerivativeDiffQuot) { //name FWITestSecondDerivativeDiffQuot??
    Vector Rand1(*Mat);
    Vector Rand2(*Mat);
    Rand1.Clear();
    Rand2.Clear();
    auto srcLoc = prob->getPointSrcLoc(0); //What is this??
    std::experimental::reseed(1);
    for (cell c = Rand1.cells(); c != Rand1.cells_end(); ++c) {
        if (PointInCellDGTS(srcLoc, *c)) continue;
        auto r1 = double(std::experimental::randint(0, 10000) / 10000.0);//!!! Values around 0!!!
        auto r2 = double(std::experimental::randint(0, 10000) / 10000.0);//!!! Values around 0!!!
        Rand1(c(), GetParam().param) = r1;
        Rand2(c(), GetParam().param) = r2;
    }
    Rand1 *= 1.0 / nc::L2MaterialNorm(Rand1, eleCache->Get(), reconlist);
    Rand2 *= 1.0 / nc::L2MaterialNorm(Rand2, eleCache->Get(), reconlist);

    //Second Order Operator declared here since the random direction is not known in Constructor.
    FWILinOpRedSecondOrder LinOpSO = FWILinOpRedSecondOrder(*U_m, *Mat, Rand1, prob,
                                                            PMLH, Conf->TConf,
                                                            eleCache->Get(), false);

    Vector MatPlusRand2(*Mat);
    Vectors U_SecondOrder(*U_m);
    Vectors U_FirstOrder(*U_m);
    Vectors U_FirstOrderShift(*U_m);

    FWISeismogram s_SO;
    FWISeismogram s_FirstOrder;
    FWISeismogram s_FirstOrderShift;

//  Second derivative \Phi''(x)[Rand1, Rand2]
    LinOpSO.SolveAndMeasure(Rand2, *Obs, s_SO, &U_SecondOrder);
//  Calculate \Phi'(Material) [Rand1]
    LinOp->SolveAndMeasure(Rand1, *Obs, s_FirstOrder, &U_FirstOrder);

    Vectors U_plusMat(*U_m);

    std::vector<double> factors {1.0, 0.1, 0.01, 0.001, 0.0001, 0.00001};
    std::vector<double> quotients(6, 0.0);

    for (int i = 0; i < 6; ++i) {
        Scalar fac = factors[i];
        MatPlusRand2 = *Mat;
        MatPlusRand2 += fac * Rand2;
        U_plusMat.Clear();
//      >>>> (Re)Calculate \Phi'(Material + RandPlusMat2) [Rand1] (not linear in Rand2)
      Vectors lel(*U_m);
      FWITimeSolver TSolveR(*discStack->wV, PMLH, Conf->TConf.rkorder);
      TSolveR.Solve(GetAssemble(MatPlusRand2),lel);
        FWILinOpRedFirstOrder LinOpMaterialShift = FWILinOpRedFirstOrder(lel, MatPlusRand2,
                                                                         prob, PMLH, Conf->TConf,
                                                                         eleCache->Get(), false);
        LinOpMaterialShift.SolveAndMeasure(Rand1, *Obs, s_FirstOrderShift,
                                           &U_FirstOrderShift);

//      Calculate the Differential Quotient
        if (GetParam().Measure) {
            s_FirstOrderShift -= *s_m;
            FWISeismogram tmp(s_SO);
            tmp *= fac;
            s_FirstOrderShift -= tmp;
            quotients[i] = sqrt(s_FirstOrderShift * s_FirstOrderShift) / fac;
        } else {
          U_FirstOrderShift -= U_FirstOrder;
          U_FirstOrderShift -= fac * U_SecondOrder;
          double DiffNorm = nc::L2TNorm(U_FirstOrderShift, eleCache->Get(), Conf->TConf.dt);
            quotients[i] = DiffNorm / fac;
        }
    }
    MOUT(quotients)
    EXPECT_VECTOR_NEAR(quotients, GetParam().values, 1e-6);
}

INSTANTIATE_TEST_SUITE_P(FWITestSecondDerivative, FWITestSecondDerivative, Values(
//                  std::string parametrization; int param; bool Measure;
        TestParams{"taup",0,false, {0.12841808,
                0.019430268,
                0.0020383175,
                0.00023159374,
                0.00011878639,
                0.00011791746}},
                //reconList chagned to {0,1}. old val  0.046147738, 0.0062779927, 0.00064977143, 8.2093646e-05, 5.3168616e-05,5.3055963e-05
        TestParams{"taup",1,false, {18.462818,
            4.4781015,
            0.45454741,
            0.04548007,
            0.0045642512,
            0.00058090853 }})
//        TestParams{"taup",2,false, {1.2780094,0.40834277,0.051416425,0.0052760852,0.00052898905,5.2915044e-05}}
//        TestParams{"taup",0,true, {0.11980079,0.016743659,0.0017392898,0.00017460158,1.7466902e-05,2.7439665e-06}},
//        TestParams{"taup",1,true, {1.9308277,0.29594415,0.030001379,0.0030029827,0.00030032543,3.0054824e-05}},
//        TestParams{"taup",2,true, {3.1746183,1.0703428,0.13591181,0.013957818,0.0013995459,0.00014000118}}
);

int main(int argc, char **argv) {
    return MppTest(
            MppTestBuilder(argc, argv).
                    WithoutDefaultConfig().
                    WithScreenLogging().
                    WithConfPath("../").
                    WithGeoPath("../conf/geo/").
                    WithPlotPath("data/vtu/").
                    WithPPM()
    ).RUN_ALL_MPP_TESTS();
}
#include "FWITestSecondDerivative.h"
