//
// Created by lionbart on 19.03.23.
//

#include "FWITestSecondDerivativeAdjoint.h"
#include "FWITestReduced1D.hpp"
#include "TestEnvironment.hpp"
#include "FWIHelpers.hpp"
#include "Observation/FWIObservationDGTSNEW.hpp"
#include "Observation/FWISeismogram.hpp"
#include "Problems/FWIProblemsDGTSVA.hpp"
#include "FWIVectorFillFunctions.hpp"
#include <experimental/random>
#include "Problems/FWIInversionProblems.hpp"

#include "Reduced/FWILinOpRedSecondOrder.h"
#include "Reduced/FWILinOpRedFirstOrder.hpp"
#include "Reduced/FWILinOpAdjRedSecondOrder.h"
#include "FWIProblemWrapperVA.hpp"


struct TestParams {
    std::string parametrization;
    int param;
    bool Measure;
    std::vector<double> values = std::vector<double>(6, 0.0);
};

/**
 * @brief Checks whether Adjoint fulfills inner product relation
 *
 * <\Phi''(Mat)[Dir1,.] MatlDirection2, g >_L²([0,T],X)   =   <MatDirection2, \Phi''(Material)[Dir1,.]^*g>_(L²(D))²
 *
 * MatDirection2/Dir1 corresponds to Rand2/Rand1
 * Take a wave field / Seismogram g = U_m to start with...
 *
 */

class FWITestSecondDerivativeAdjoint : public TestWithParam<TestParams> {
protected:
    std::unique_ptr<WPConf> Conf = nullptr;
    std::shared_ptr<Meshes> meshes;
    std::shared_ptr<WaveProblemDGTSVA> prob;
    std::unique_ptr<FWIDiscretizationStack> discStack;
    std::unique_ptr<EleCacheContainer> eleCache;
    std::shared_ptr<PMLHandler> PMLH;
    std::unique_ptr<DGVAcousticPlotter> plotter;
    std::unique_ptr<FWISeismogram> s_m;
    std::shared_ptr<FWIObservationOperatorNEW> Obs;

    std::unique_ptr<Vector> Mat;
    std::unique_ptr<Vectors> U_m;
    std::unique_ptr<FWILinOpRedSecondOrder> LinOpSO;
    std::unique_ptr<FWILinOpAdjRedSecondOrder> LinOpSOAdjoint;
//    const std::list<unsigned int> reconlist{0, 1, 2};//Potentiell nicht zutreffend auf den Fall hier?!
//    will ja nur rho und vp!?
//    wahrscheinlich ändert es aber trotzdem nichts, weil da nur nullen auf anderes trifft
    const std::list<unsigned int> reconlist{0, 1};
    //!!! Achtung, ggf. wird irgendwas über configs ausgelesen.
    // Gibt ggf verschiedene orte, wo configs sachen herkommen. hier "nur" test
    //--- Abstiegsrichtungen checken
public:
    FWITestSecondDerivativeAdjoint() {
        Config::Initialize(ConfigTestSecondDerivativeAdjoint); //Details of Config in .hpp file
        Conf = std::make_unique<WPConf>();
        meshes = getFWIMesh();
        prob = GetInversionProblem(*Conf);
        prob->setParametrization(GetParam().parametrization);
        int deg = 0;
        Config::Get("deg", deg);
        discStack = std::make_unique<FWIDiscretizationStack>(meshes, prob->getParamComps(), prob->GetStateComponents(),
                                                             deg, LevelPair{Conf->level, -1});
        eleCache = std::make_unique<EleCacheContainer>(*discStack->wV, prob->getNumL());
        plotter = std::make_unique<DGVAcousticPlotter>(prob->getNumL(), &eleCache->Get());
        std::shared_ptr<FWIObservationSpecification> spec = FWIObsSpecWPCreator().Witht0(Conf->TConf.t0)
                .Withdt(Conf->TConf.dt).WithT(Conf->TConf.T).CreateShared();
        auto mask = GetMeasureMaskNEW(meshes->dim(), prob->getNumL(), 0);
        Obs = FWIObservationCreator(eleCache->Get(), discStack->GetMesh(), prob->getNumL()).CreateUnique();
        PMLH = std::make_shared<PMLHandler>();
        Mat = std::make_unique<Vector>(*discStack->mV);
        Mat->Clear();
        FillVectorFromProblem(*prob, *Mat);

//      Get Solutions of Forward Problem
        FWITimeSolver TSolveR(*discStack->wV, PMLH, Conf->TConf.rkorder);
        U_m = std::make_unique<Vectors>(Conf->TConf.timeSteps + 1, *discStack->wV);
        s_m = std::make_unique<FWISeismogram>();
        prob->setShot(0);
        TSolveR.SolveAndMeasure(GetAssemble(*Mat), *Obs, *s_m, U_m.get());
    }
    std::shared_ptr<DGAssemble> GetAssemble(const Vector &Material) {
        auto Mp = std::make_unique<Vector>(*discStack->mV);
        Mp->Clear();
        ReparametrizeMaterialVector(Material, *prob, *Mp);
        Mp->Accumulate();
        auto WrappedProblem = std::make_shared<FWIProblemWrapperVA>(*prob, *prob, std::move(Mp),
                                                                    prob->getNumL(), prob->getDim(), prob);
        auto assemble = std::make_shared<DGAssemble>(WrappedProblem, true, -1, Conf->TConf.t0,
                                                     Conf->TConf.T,
                                                     Conf->TConf.dt);
        assemble->SetElementCache(&eleCache->Get());
        return std::move(assemble);
    }

    void TearDown() override {
        Config::Close();
    }
};

TEST_P(FWITestSecondDerivativeAdjoint, FWITestSecondDerivativeAdjoint) {
    /*
     * Preparation Material/Directions and Operators
     */
    /* Mat ist einfach nur rho = 2 und vp = 5 (fast) überall, its ok.. */
    Vector MatDirection1(*Mat);
    Vector MatDirection2(*Mat);
    Vector Rand(*Mat);
    MatDirection1.Clear();
    MatDirection2.Clear();
    Rand.Clear();
    auto srcLoc = prob->getPointSrcLoc(0); //What is this??
    std::experimental::reseed(3);
//    Vectors U_random(*U_m);
//    for (int i = 0; i < U_m->size(); i++) {
        for (cell c = Rand.cells(); c != Rand.cells_end(); ++c) {
            auto r = double(std::experimental::randint(0, 10000) / 10000.0);
//            auto r = 1 + double(std::experimental::randint(0, 10000) / 10000.0);
            if(GetParam().param == 2) {
                Rand(c(), 0) = r;
                r = double(std::experimental::randint(0, 10000) / 10000.0);
                Rand(c(), 1) = r;
            } else {
                Rand(c(), GetParam().param) = r;
            }
        }
        //U_random is Vector Rand in every time step..
        Vectors U_random(U_m->size(), Rand, true);
//    }


//    for (const auto &elemIt: eleCache->Get()) {
//        const auto& elem = *elemIt.second;
//        const auto& c = elem.GetCell();
//        if (PointInCellDGTS(srcLoc, c)) continue;
//        auto r1 = double(std::experimental::randint(0, 10000) / 10000.0);
//        auto r2 = double(std::experimental::randint(0, 10000) / 10000.0);
//        MatDirection1(c(), GetParam().param) = r1;
//        MatDirection2(c(), GetParam().param) = r2;
//    }
//    std::cout << "Testbefüllung Rand1/2 mit eleCache Schleife" << std::endl;

    for (cell c = MatDirection1.cells(); c != MatDirection1.cells_end(); ++c) {
        if (PointInCellDGTS(srcLoc, *c)) continue; //!!! Was macht diese Zeile??
        auto r1 = double(std::experimental::randint(0, 10000) / 10000.0);//!!! Values around 0!!!
//        auto r1 = 1 + double(std::experimental::randint(0, 10000) / 10000.0);
        auto r2 = double(std::experimental::randint(0, 10000) / 10000.0);//!!! Values around 0!!!
//        auto r2 = 1 + double(std::experimental::randint(0, 10000) / 10000.0);
        if(GetParam().param == 2) {
            MatDirection1(c(), 0) = r1;
            MatDirection2(c(), 0) = r2;
            r1 = double(std::experimental::randint(0, 10000) / 10000.0);
            r2 = double(std::experimental::randint(0, 10000) / 10000.0);
            MatDirection1(c(), 1) = r1;
            MatDirection2(c(), 1) = r2;
        } else {
            MatDirection1(c(), GetParam().param) = r1;
            MatDirection2(c(), GetParam().param) = r2;
        }
    }

//    std::cout << "Befüllung Rand1/2 mit Cells Schleife" << std::endl;

    MatDirection1 *= 1.0 / nc::L2MaterialNorm(MatDirection1, eleCache->Get(), reconlist);
    MatDirection2 *= 1.0 / nc::L2MaterialNorm(MatDirection2, eleCache->Get(), reconlist);
    reconstructionContainer recon = reconstructionContainer();
    //Needed in FWIAdjointCalculator
    recon.Rho = true; recon.Vp = true;

    LinOpSO = std::make_unique<FWILinOpRedSecondOrder>(*U_m, *Mat, MatDirection1, prob,
                                                       PMLH, Conf->TConf,
                                                       eleCache->Get(), false);
    LinOpSOAdjoint = std::make_unique<FWILinOpAdjRedSecondOrder>(*U_m, *Mat, MatDirection1, *Obs, prob, PMLH, Conf->TConf, recon, eleCache->Get());

    /*
     * Calculate <\Phi''(Mat)[Dir1,.] MatlDirection2, g >_L²([0,T],X)   =   <MatDirection2, \Phi''(Material)[Dir1,.]^*g>_(L²(D))²
     */
    Vectors U_SecondOrder(*U_m);
    Vectors *pU_SecondOrder = &U_SecondOrder;
    Vectors LambdaFull(*U_m);
    LambdaFull = 0.0;
    Vectors *pLambdaFull = &LambdaFull;
    std::shared_ptr<Vector> AdjointSO = std::make_shared<Vector>((*U_m)[0]);
    AdjointSO->Clear();

    std::shared_ptr<FWISeismogram> s_SO = std::make_shared<FWISeismogram>();
//    *s_SO = 0.0;
//    std::cout << "xxxxxxxxxxx LinearSO to Calculate to go:" << std::endl;
    LinOpSO->SolveAndMeasure(MatDirection2, *Obs, *s_SO, pU_SecondOrder);

    const std::shared_ptr<FWISeismogram> s_u_random = std::make_shared<FWISeismogram>(*s_SO);//Default Constructor!?

    FWISeismogram s_Adj(*s_SO);
    *s_u_random = 0.0;
    for (const auto& rec: (*s_u_random).getSpecification()){
        auto &ts = (*s_u_random).getTimeSeries(rec);
        for (auto& t : ts){
            if (t.first < Eps) continue;
            auto r = double(std::experimental::randint(0, 10000) / 10000.0);
            (*t.second)[0] = r;
        }
    }
//    std::cout << "xxxxxxxxxxx Adjoint to Calculate to go:" << std::endl;
    LinOpSOAdjoint->CalculateAdjoint(*s_u_random, *AdjointSO, pLambdaFull);

    //!!! Muss noch weg, unnötige Referenz
    Vector AdjointWatch = *AdjointSO;//To see it while debugging.

//    MOUT(nc::L2MaterialNorm(MatDirection2, eleCache->Get(), reconlist))// = 1 Normalized
//    MOUT(nc::L2MaterialNorm(AdjointWatch, eleCache->Get(), reconlist));
//    MOUT(MatDirection2.Size())//6136
//    MOUT(AdjointWatch.Size())//24544

    //    SKP RightHandSide                     <MatDirection2   , \Phi''(Material)[MatDir1,.]^*s_u_random>_(L²(D))²
    double rightInnerProd = nc::L2MaterialScalar(MatDirection2, AdjointWatch, eleCache->Get(), reconlist);
    //    SKP LeftHandSide   <\Phi''(Material)[MatDir1,MatDir2], s_u_random >_L²([0,T],X)
    double leftInnerProdSeis = *s_SO * *s_u_random;


//    double leftInnerProd = nc::L2TScalar(*pU_SecondOrder, U_random, eleCache->Get(), Conf->TConf.dt);
    std::cout << leftInnerProdSeis << "<--Linke Seite mit Seismogrammen (Phi''); Rechte Seite Material ----> " << rightInnerProd << std::endl;

    std::vector<double> RelDiff(1, 0.0);
    double absDiff = std::abs(rightInnerProd - leftInnerProdSeis);
    RelDiff[0] = absDiff*(1/std::abs(leftInnerProdSeis));

//    MOUT(absDiff)
    MOUT(RelDiff)
    EXPECT_VECTOR_NEAR(RelDiff, GetParam().values, 0.4);
}

INSTANTIATE_TEST_SUITE_P(FWITestSecondDerivativeAdjoint, FWITestSecondDerivativeAdjoint, Values(
TestParams{"taup",0,false, {0.05} },

TestParams{"taup",1,false, {0.34} },

TestParams{"taup",2,false, {0.23} }//Mit beiden parametern rho, vp gesetzt.
//nicht wirklich gut, aber produziert in TestReducedSO2D Abstiegsrichtungen..

        )
);
/*
 * Weird: Wenn ich rechte Seite für zFull berechnung ändere ist in Test 2 gleich (0.010395359) und Test 1  0.00829783
 */

int main(int argc, char **argv) {
    return MppTest(
            MppTestBuilder(argc, argv).
                    WithoutDefaultConfig().
                    WithScreenLogging().
                    WithConfPath("../").
                    WithGeoPath("../conf/geo/").
                    WithPlotPath("data/vtu/").
                    WithFileLogging().//Diese beiden Befehle speichern Log files automatisch
                    WithLogPath("log/").
                    WithPPM()
    ).RUN_ALL_MPP_TESTS();
}
