#include <iostream>
#include "DGAssemble.hpp"


void DGAssemble::RHS(double t, Vector &b) const {
    rhsAssembler->RHS(t, b);
}

void DGAssemble::FluxMatrixOnly(Matrix &A) const {
    FluxMatrix(A, false, true);
}

void DGAssemble::FluxMatrix(Matrix &A) const {
    FluxMatrix(A, true, true);
}

void DGAssemble::FluxMatrix(Matrix &A, bool damping, bool fwd) const {
    acousticMatrixAssembler.FluxMatrix(A, damping, fwd);
    A.setCollectAfterProd(useCollectTweak);
    A.CreateSparse();

}

void DGAssemble::MassMatrixOnlyKappa(Matrix &M) const {
    acousticMatrixAssembler.MassMatrixOnlyKappa(M);
}

void DGAssemble::MassMatrix(Matrix &M) const {
    acousticMatrixAssembler.MassMatrix(M);
    M.setCollectAfterProd(useCollectTweak);
    M.CreateSparse();
}
void DGAssemble::DampingMatrix(Matrix &D) const {
  acousticMatrixAssembler.DampingMatrix(D);
  D.setCollectAfterProd(useCollectTweak);
  D.CreateSparse();
};

DGAcousticElement *DGAssemble::getElem(const VectorMatrixBase &base, const Cell &C) const {
    if (EleCache == nullptr)
        return new DGAcousticElement(base, C, problem->NumDamping());
    else {
        auto p = EleCache->find(C());
        return p->second;
    }
}

void DGAssemble::destroyElement(DGAcousticElement *elem) const {
    if (EleCache == nullptr) {
        delete elem;
    }
}

void DGAssemble::SystemMatrix(Matrix &S) const {
    FluxMatrix(S);
    S.CreateSparse();
}

void DGAssemble::SetElementCache(const std::unordered_map<Point, DGAcousticElement *> *_EleCache) {
    EleCache = _EleCache;
    plotter.SetElementCache(EleCache);
    acousticMatrixAssembler.SetElementCache(EleCache);
    rhsAssembler->SetElementCache(EleCache);
}

void DGAssemble::PlottingExact(const Vector &u, int step, double t) const {
    Vector PIntComponent(0.0, u);
    for (cell c = PIntComponent.cells(); c != PIntComponent.cells_end(); ++c) {
        const Cell &C = *c;
        DGAcousticElement *elem = getElem(u, C);
        double Area = elem->Area();
        for (int q = 0; q < elem->nQ(); ++q) {
            double w = elem->QWeight(q);
            PIntComponent(c(), 0) += w * problem->ut(t, *c, elem->QPoint(q), 0) * (1.0 / Area);
        }
    }
    char filename[127];
    mpp::plot("PInt") << PIntComponent << mpp::save_plot(NumberName("PInt", filename, step));
    Plotting::Instance().Clear();
}

void DGAssemble::VtkPlotting(const Vector &u, int step) const {
    if (step == 0) InitialPlotting(u);
    plotter.plotWaveVector(u, step);
}

void DGAssemble::Interpolate(const Vector &u, Vector &U) const {
    U = 0;
    int NN = 1 + u.dim();
    for (cell c = u.cells(); c != u.cells_end(); ++c) {
        const Cell &C = *c;
        DGAcousticElement Elem(u, C, problem->NumDamping());
        row r = u.find_row(c());
        for (int i = 0; i < c.Children(); ++i) {
            cell fc = U.find_cell(c.Child(i));
            const Cell &FC = *fc;
            DGAcousticElement Elem_fc(u, FC, problem->NumDamping());
            row fr = U.find_row(fc());
            std::vector<Point> z = U.GetDoF().GetStoragePoints(C);
            for (int iN = 0; iN < NN; ++iN) {
                for (int i = 0; i < Elem.shape_size(); ++i) {
                    Scalar s = Elem.SVValue(C.GlobalToLocal(z[i]), u, iN);
                    U(fr, iN * Elem.shape_size() + i) = s;
                }
            }
        }
    }
    U.Accumulate();
}

void DGAssemble::InitialPlotting(const Vector &u) const {
    plotter.plotMatVec(u, acousticMatrixAssembler.GetMaterial());
  plotter.plotProcs(u);


}

void DGAssemble::FinishStep(Vector &u, double t, double dt, int step, int plot_step) const {
    if (step % plot_step == 0 && plot_step > 0) {
        PrintStep(t, u, step);
        VtkPlotting(u, step);
    }
}

void DGAssemble::InitialStep(double t, Vector &u, bool print) const {
    SetInitialValueOld(u);
    if (print) PrintStep(t, u, 0);
    VtkPlotting(u, 0);
}

void DGAssemble::PrintStep(double t, const Vector &u, int step) const {
    LagrangeDiscretization disc(u.GetDisc().GetMeshes(), 0,
                                u.GetDisc().GetMeshes().Level());
    std::cout << std::fixed;
    std::cout.precision(5);
    Vector PComponent(0.0, disc);
    Vector PIntComponent(0.0, disc);

    mout << "Step(" << step << "), t("
         << t
         << "): E(u) = " << Energy(u);
    Vector u_tmp(u);
    if (problem->HasExactSolution()) {
        get_u_int_assemble(t, u_tmp);
        u_tmp -= u;
        mout << " IntErr " << sqrt(2.0 * Energy(u_tmp));
        mout << " L2Err " << L2Error(t, u);
    }
    Vector u_int(u);
    get_u_int_assemble(t, u_int);
    vector<double> p = {infty, -infty};
    double max_PInt = 0;
    for (cell c = PComponent.cells(); c != PComponent.cells_end(); ++c) {
        const Cell &C = *c;

        DGAcousticElement *elem = getElem(u_tmp, C);
        double P = 0;
        double PInt = 0;
        for (int q = 0; q < elem->nQ(); ++q) {
            double w = elem->QWeight(q);
            P += w * elem->Pressure(q, u);
            PInt += w * elem->Pressure(q, u_int);
        }
        double a = elem->Area();
        P *= (1.0 / a);
        PInt *= (1.0 / a);

        if (p[1] < P)
            p[1] = P;
        if (p[0] > P)
            p[0] = P;

        if (max_PInt < PInt)
            max_PInt = PInt;
        destroyElement(elem);
    }
    p[1] = PPM->Max(p[1]);
    p[0] = PPM->Min(p[0]);
    if (abs(p[0]) < 1e-10) p[0] = 0;

    mout << " max(P) = " << p[1]
         << " min(P) = " << p[0]
         << endl;

    return;

}

double DGAssemble::L2Error(double t, const Vector &u) const {
    double err = 0;
    for (cell c = u.cells(); c != u.cells_end(); ++c) {
        const Cell &C = *c;

        DGAcousticElement *elem = getElem(u, C);
        int p_dim = elem->p_dimension();
        for (int q = 0; q < elem->nQ(); ++q) {
            double w = elem->QWeight(q);
            Point z = elem->QPoint(q);
            VectorField tmp = zero;
            for (int i = 0; i < problem->Dim(); ++i) {
                tmp[i] = problem->ut(t, *c, z, 0);
            }
            VectorField V = elem->Velocity(q, u);
            V -= tmp;
            err += w * V * V;

            for (int n = 0; n < p_dim; ++n) {
                double P_i = elem->Pressure_i(q, u, n)
                             - problem->ut(t, *c, z, problem->Dim() + n);
                err += w * P_i * P_i;
            }
        }
        destroyElement(elem);
    }
    return sqrt(PPM->Sum(err));
}

double DGAssemble::Energy(const Vector &u) const {
    Scalar e = 0;
    for (cell c = u.cells(); c != u.cells_end(); ++c) {
        const Cell &C = *c;

        DGAcousticElement *elem = getElem(u, C);
        e += Energy_local(u, *elem, c);
        destroyElement(elem);
    }
    return 0.5 * PPM->Sum(e);
}

double
DGAssemble::Energy_local(const Vector &u, DGAcousticElement &elem, const cell &c) const {
    Scalar e = 0;
    for (int q = 0; q < elem.nQ(); ++q) {
        double w = elem.QWeight(q);
        Point z = elem.QPoint(q);
        int p_dim = elem.p_dimension();
        double rho = problem->Rho(*c, z);
        VectorField V = elem.Velocity(q, u);
        e += w * rho * V * V;
        for (int n = 0; n < p_dim; ++n) {
            double P_i = elem.Pressure_i(q, u, n);
            double invKappa = problem->invKappa_i(*c, z, n);
            e += w * invKappa * P_i * P_i;
        }
    }
    return e;
}

void DGAssemble::get_u_int_assemble(double t, Vector &u) const {
    u.Clear();
    for (cell c = u.cells(); c != u.cells_end(); ++c) {
        row r = u.find_row(c());
        const Cell &C = *c;
        DGAcousticElement Elem(u, C, problem->NumDamping());
        std::vector<Point> z = u.GetDoF().GetStoragePoints(C);
        for (int i = 0; i < Elem.u_dimension(); ++i)
            for (int j = 0; j < Elem.shape_size(); ++j)
                u(r)[i * Elem.shape_size() + j] = problem->ut(t, *c, z[j], i);
    }
}