//
// Created by cr on 14.07.21.
//

#ifndef FWI_DGVACOUSTICMATRIXASSEMBLER_HPP
#define FWI_DGVACOUSTICMATRIXASSEMBLER_HPP


#include "Cell.hpp"
#include "DGAcousticElement.hpp"
#include "DGViscoAcousticProblem.hpp"

class DGVAcousticMaterial {
protected:
  int NumDamping = 0;
public:
  int GetNumDamping() const { return NumDamping; }

public:
  DGVAcousticMaterial() {}

  virtual ~DGVAcousticMaterial() {};

  virtual double invKappa_i(const Cell &, const Point &, int n) const = 0;

  virtual double Kappa_i(const Cell &, const Point &, int n) const = 0;

  virtual double invTau_i(const Cell &, const Point &, int n) const = 0;

  virtual double Kappa(const Cell &, const Point &) const = 0;

  virtual double Rho(const Cell &, const Point &) const = 0;
  virtual double alpha(const Cell &,
                       const Point &) const = 0;
  virtual double alpha1() const =0;
  virtual double alpha2() const = 0;
};

class DGVAcousticMaterialFromVector : public DGVAcousticMaterial {
protected:
  std::unique_ptr<Vector> MatRepara;
public:
  DGVAcousticMaterialFromVector(const Vector &Mat) : MatRepara(std::make_unique<Vector>(Mat)) {
  }
  virtual double alpha(const Cell &,
                       const Point &) const {
    Exit("not implemented here!");
    return 0.;
  }
  virtual double alpha1() const {
    Exit("not implemented here!");
    return 0.;
  }
  virtual double alpha2() const {
    Exit("not implemented here!");
    return 0.;
  }
};

class DGVAcousticMaterialFromProblem : public DGVAcousticMaterial {
  const ProblemVAcoustic &problem;
public:
  DGVAcousticMaterialFromProblem(const ProblemVAcoustic &problem) : problem(problem) {
    NumDamping = problem.NumDamping();
  }

  double invKappa_i(const Cell &C, const Point &x, int n) const override {
    return problem.invKappa_i(C, x, n);
  }

  double Kappa_i(const Cell &C, const Point &x, int n) const override {
    return problem.Kappa_i(C, x, n);
  }

  double invTau_i(const Cell &C, const Point &x, int n) const override {
    return problem.invTau_i(C, x, n);
  }

  double Kappa(const Cell &C, const Point &x) const override {
    return problem.Kappa(C, x);
  }

  double Rho(const Cell &C, const Point &x) const override {
    return problem.Rho(C, x);
  }
  double alpha(const Cell &C,
                       const Point &x) const override{
    return  problem.alpha(C,x);
  }
  double alpha1() const override{

    return problem.alpha1();
  }
  double alpha2() const override{
    return problem.alpha2();
  }
};

class DGVAcousticMatrixAssembler {
  void assembleDamping(const DGAcousticElement &elem, DGRowEntries &A_c) const;

  void assembleGradP(const DGAcousticElement &, DGRowEntries &, bool) const;

  void assembleDivV(const DGAcousticElement &, DGRowEntries &, bool) const;

  void
  assembleDifferentialPartFlux(const Matrix &A, const DGAcousticElement &elem, DGRowEntries &A_c,
                               bool damping, bool fwd) const;


  virtual void assembleDirichletBoundaryFlux(DGRowEntries &A_c,
                                             const Cell &C, const DGAcousticFaceElement &felem,
                                             int q, bool fwd) const;

  void assembleNeumannBoundaryFlux(DGRowEntries &A_c,
                                   const Cell &C, const DGAcousticFaceElement &felem,
                                   int q, bool fwd) const;

  void assembleRobinBoundaryFlux(DGRowEntries &A_c,
                                 const Cell &C, const DGAcousticFaceElement &felem,
                                 int q, bool fwd) const;

  void assembleBoundaryFlux(DGRowEntries &A_c, const Mesh &M,
                            const Cell &C,
                            const DGAcousticFaceElement &felem, bool fwd) const;

  bool dampingFlux = false;
  bool upwind = true;
  bool fwdBool = true;

  DGAcousticElement *getElem(const VectorMatrixBase &base, const Cell &C) const {
    return (EleCache == nullptr) ? new DGAcousticElement(base, C, Mat->GetNumDamping())
                                 : EleCache->find(
            C())->second;
  }

protected:
  std::unique_ptr<DGVAcousticMaterial> Mat;
  const std::unordered_map<Point, DGAcousticElement *> *EleCache;
public:
  DGVAcousticMatrixAssembler(std::unique_ptr<DGVAcousticMaterial> Mat, bool fwd,
                             const std::unordered_map<Point, DGAcousticElement *> *EleCache = nullptr)
      :
      Mat(std::move(Mat)), fwdBool(fwd), EleCache(EleCache) {
    Config::Get("upwind", upwind);
    Config::Get("dampingFlux", dampingFlux);
  }

  void SetElementCache(const std::unordered_map<Point, DGAcousticElement *> *_EleCache) {
    EleCache = _EleCache;
  }

  const DGVAcousticMaterial &GetMaterial() const { return *Mat; }

  void MassMatrix(Matrix &M) const;

  void MassMatrixOnlyKappa(Matrix &M) const;

  void FluxMatrix(Matrix &A, bool, bool) const;

  void SystemMatrix(Matrix &A) const {
    FluxMatrix(A, true, fwdBool);
  }

  void DampingMatrix(Matrix &D) const;

  void assembleFlux(Matrix &A_loc, const DGAcousticElement &elem,
                    DGRowEntries &A_c, bool fwd) const;


};


#endif //FWI_DGVACOUSTICMATRIXASSEMBLER_HPP
