// $Header: https://svn.math.kit.edu/svn/M++/trunk/DGWave/src/DGWave.C 653 2016-07-28 16:00:30Z ziegler $

#include "m++.hpp"
#include "NonmigratedFiles/DGMatrixGraph.hpp"
#include "Tools.hpp"
#include "NonmigratedFiles/DGAcoustic.hpp"
#include "DGViscoAcoustic.hpp"
#include "NonmigratedFiles/DGViscoAcousticX.hpp"
#include "NonmigratedFiles/DGElasticity.hpp"
#include "NonmigratedFiles/DGViscoElastic.hpp"
#include <fstream>
#include "DGTimeIntegrator.hpp"

class TransferWave_dG : public Operator {
public:
    TransferWave_dG () {}
    virtual void Projection(const Vector& f, Vector& c) const{}
    virtual void Prolongation(const Vector& c, Vector& f) const{}
    virtual void multiply (Vector& u, const Vector& v) const {}
    virtual void multiply_transpose (Vector& u, const Vector& v) const{ }
};

inline constAB<Operator,Vector> 
operator * (const TransferWave_dG& T, const Vector& v) {
    return constAB<Operator,Vector>(T,v);
}
inline constAB<Vector,Operator> 
operator * (const Vector& v, const TransferWave_dG& T) {
    return constAB<Vector,Operator>(v,T);
}


class P2P0TransferWave_2D : public TransferWave_dG {
public:
    P2P0TransferWave_2D () {}
    void Restriction (const Vector& F, Vector& C) const {
        C = 0.0;
        for (cell c=C.cells(); c!=C.cells_end(); ++c) {
            row r = C.find_row(c());
            int nn = 6;
            for (int k=0;k<3;++k){
                for (int i=0;i<6;++i){
                    C(r)[k] += F(r)[k*nn+i];
                }
            }
        }
    }
    void Prolongation (const Vector& C, Vector& F) const {
        for (cell c=C.cells(); c!=C.cells_end(); ++c) {
            row r = C.find_row(c());
            int nn = 6;
            for (int k=0;k<3;++k){
                for (int i=0;i<6;++i){
                    F(r)[k*nn+i] = C(r)[k];
                }
            }
        }
        F.ClearDirichletValues();
        F.MakeAdditive();
        F.Accumulate();
    }
    void multiply (Vector& u, const Vector& v) const { Prolongation(v,u); }
    void multiply_transpose (Vector& u,const Vector& v)const{Restriction(v,u);}
};

class P1P0TransferWave_3D : public TransferWave_dG {
public:
    P1P0TransferWave_3D () {}
    void Restriction (const Vector& F, Vector& C) const {
        C = 0.0;
        for (cell c=C.cells(); c!=C.cells_end(); ++c) {
            row r = C.find_row(c());
            int nn = c.Corners();
            for (int k=0;k<r.n();++k){
                for (int i=0;i<nn;++i){
                    C(r)[k] += F(r)[k*nn+i];
                }
            }
        }
    }
    void Prolongation (const Vector& C, Vector& F) const {
        for (cell c=C.cells(); c!=C.cells_end(); ++c) {
            row r = C.find_row(c());
            int nn = c.Corners();
            for (int k=0;k<r.n();++k){
                for (int i=0;i<nn;++i){
                    F(r)[k*nn+i] = C(r)[k];
                }
            }
        }
        F.ClearDirichletValues();
        F.MakeAdditive();
        F.Accumulate();
    }
    void multiply (Vector& u, const Vector& v) const { Prolongation(v,u); }
    void multiply_transpose (Vector& u,const Vector& v)const{Restriction(v,u);}
};

TransferWave_dG* GetTransferWave_dG (const std::string& name) {
    if (name=="P2P0_2D") return new P2P0TransferWave_2D;
    if (name=="P1P0_3D") return new P1P0TransferWave_3D;
    else Exit("no transfer "+name+" implemented");
}


class dGFVPC : public Preconditioner {
    const TransferWave_dG* TF;
    const DGAssemble& assemble;
    const DGAssemble& assembleFV;
    const Matrix* A;
    Matrix* fvA;
    Preconditioner *SM;
    double theta;
    Preconditioner *BaseP;
    Solver* LS;
    Vector* u_fv;
    int pre;
    int pos;
public:
    dGFVPC (const DGAssemble & Ass, const DGAssemble & AssFV,
            MatrixGraph& GFV) : A(0), theta(1), pre(1), pos(1),
        assemble(Ass), assembleFV(AssFV) {
        std::string pc = "PointBlockJacobi";
        Config::Get("Smoother", pc);
        SM = GetPC(pc);
        u_fv = new Vector(GFV);
        TF = GetTransferWave_dG("P2P0_2D");
        //        TF = GetTransferWave_dG("P1P0_3D");
        std::string solver = "gmres";
        Config::Get("BaseSolver", solver);
        std::string base_pc = "PointBlockJacobi";
        Config::Get("BasePreconditioner", base_pc);
        BaseP = GetPC(base_pc);
        LS = GetSolver(BaseP, solver, "BaseSolver");
        Config::Get("SmootherDamp", theta);
        Config::Get("Presmoothing", pre);
        Config::Get("Postsmoothing", pos);
    }
    ~dGFVPC() {
        Destruct();
        //delete LS;
        // delete SM;
    }
    void Construct (const Matrix& _A) {
        A = &_A;
        SM->Construct(*A);
        fvA = new Matrix(*u_fv);
        assembleFV.Jacobi(*u_fv,*fvA);
        (*LS)(*fvA);
    }
    void Destruct () {
        //if (BaseP) BaseP->Destruct();
        // SM->Destruct();
        // if (u_fv) delete u_fv;
        // u_fv = 0;
        // if (fvA) delete fvA;
        // fvA = 0;
    }

    void Cycle(Vector&u, Vector& r)  const {
        Vector w(u);
        //Presmoothing
        for (int i=0; i<pre; ++i) {
            w = (*SM) * r;
            if (theta != 1) w *= theta;
            r -= (*A) *w;
            u += w;
        }

        Vector d(*u_fv);
        Vector c(d);
        c = 0.0; d = 0.0;
        //	mout << " r " << r.norm() << endl;
        d = r * (*TF); //Projection
        //	mout << " d " << d.norm() << endl;
        c = (*LS) * d;
        //	mout << " c*d " << c * d << endl;
        w = (*TF) * c; //Prolongation
        //	mout << " w * r " << w * r << endl;
        r -= (*A) * w;
        u += w;

        //Postsmoothing
        for (int i=0; i<pos; ++i) {
            w =  r * (*SM);
            if (theta != 1) w *= theta;
            r -= (*A) * w;
            u += w;
        }
    }

    void multiply (Vector& u, const Vector& b) const {
        // u = *SM * b;
        Vector r(b);
        u = 0.0;
        Cycle(u,r);
    }

    // void multiply_transpose (Vector& u, const Vector& v) const {
    //     multiply(u,v);
    // }

    std::string Name () const {
        return "dGFVPC with " + SM->Name(); // + " and " + MG->Name();
    }
};

Preconditioner* GetdGPC (const MatrixGraphs& G_vector, MatrixGraph& G_vectorFV,
                         const DGAssemble & assemble, const DGAssemble & assembleFV) {
    if (assemble.DiscName() == "dGvector_maxwell_FV")
        return GetPC(G_vector,assemble);
    std::string pc = "Multigrid";
    Config::Get("Preconditioner", pc);
    if (pc == "Multigrid") return new dGFVPC(assemble, assembleFV, G_vectorFV);
    else return GetPC(pc);
}
DGVAcousticAssemble* GetAssemble(const DGTSDiscretization& _disc, const ProblemVAcoustic& _prob, Plot& _plot){
    std::string dir = "fwd";
    Config::Get("SolveDirection", dir);
    if (dir == "bwd") return new DGVAcousticAssembleBwd(_disc,_prob,_plot);
    if (dir == "fwd") return new DGVAcousticAssemble(_disc,_prob,_plot);
}

DGVElasticAssemble* GetAssemble(const DGTSDiscretization& _disc, const ProblemVElastic& _prob, Plot& _plot){
    std::string dir = "fwd";
    Config::Get("SolveDirection", dir);
    if (dir == "fwd") return new DGVElasticAssemble(_disc,_prob,_plot);
}
DGVAcousticXAssemble* GetAssembleX(const DGVAcousticXDiscretization& _disc, const ProblemVAcousticX& _prob, Plot& _plot){
    std::string dir = "fwd";
    Config::Get("SolveDirection", dir);
    if (dir == "fwd") return new DGVAcousticXAssemble(_disc,_prob,_plot);
}

#define ref(U,l,s) *(U[s+l*Scales)

void Set (Vector& u, const double& s) {
    double h = infty;
    for (row r=u.rows(); r!=u.rows_end(); ++r)
        if (r()[1] < h)
            h = r()[1];
    h = PPM->Min(h);
     std::unordered_map<Point,int> index;
    typedef  std::unordered_map<Point,int>::const_iterator Iterator;
    int n = 0;
    mout << "h " << h << endl;
    ExchangeBuffer E;
    for (row r=u.rows(); r!=u.rows_end(); ++r) {
        if (abs(r()[1] - h) > 10e-10) continue;
        //	if (u.master(r()))
        E.Send(0) << r();
    }
    E.Communicate();
    if (PPM->proc() == 0) {
        for (short q=0; q<PPM->size(); ++q)
            while (E.Receive(q).size() < E.ReceiveSize(q)) {
                Point z;
                E.Receive(q) >> z;
                index[z] = n++;
                pout << "z " << z << " n " << n-1 << endl;
            }
        PPM->Broadcast(n);
        for (Iterator i=index.begin(); i!=index.end(); ++i) {
            Point z = i->first;
            PPM->Broadcast(z);
            int k = i->second;
            PPM->Broadcast(k);
            pout << "z " << z << " k " << k << endl;
        }
    } else {
        PPM->Broadcast(n);
        for (int j=0; j<n; ++j) {
            Point z;
            PPM->Broadcast(z);
            int k;
            PPM->Broadcast(k);
            index[z] = k;
            pout << "z " << z << " k " << k << endl;
        }
    }

    RVector U(n);

    u = 0;
    for (row r=u.rows(); r!=u.rows_end(); ++r)
        if (abs(r()[1] - h) < 10e-10) {
            u(r,0) = s + r()[0];
            Iterator i = index.find(r());
            U[i->second] = u(r,0);
        }
    for (row r=u.rows(); r!=u.rows_end(); ++r) {
        Iterator i = index.find(Point(r()[0],h));
        u(r,0) = U[i->second];
    }
}

static ostream& operator << (ostream& s, conststd::vector<int>& v) {
    char buf[128];
    for (int i=0; i<v.size(); ++i) {
        sprintf(buf,"%10d",v[i]);
        s << buf;
        if ((i+1)%25==0) s << endl;
    }
    return s << endl;
}

static void PrintValues (const char* name, const RVector& V) {
    //     if (V.norm() == 0.0) return;
    mout << name << V<< endl;
    int N = V.size();
    if (N < 2) return;
    RVector Difference(0.0,N-1);
    for (int i=0; i<N-1; ++i)
        Difference[i] = V[i] - V[i+1];
    mout << "Difference" << endl;
    mout << "                     " << Difference;
    if (N < 3) return;
    RVector Rate(0.0,N-2);
    for (int i=0; i<N-2; ++i){
        if (abs(Difference[i]) > 1e-9){
            Rate[i] = log2(Difference[i] / Difference[i+1]);
        }
        mout << "                     " << Rate<<endl;
    }
}
static void PrintValuesCR (const char* name, const RVector& V) {
    int scales = 1;
    int level = 1;
    int plevel = 1;
    mout << "--------------------------------";
    mout << name;
    mout << "--------------------------------" << endl;
    mout << "###########################" << endl;
    Config::Get("level", level, -1);
    Config::Get("plevel", plevel, -1);
    Config::Get("Scales", scales, -1);
    if (scales == 1) return;
    int L = level - plevel + 1;
    for (int l = 0; l < L; ++l) {
        mout << "SpaceLevel " << l << " Errors:" << endl;
        for (int i = 0; i < scales; ++i) {
            mout << V[l * scales + i] << "     ";
        }
        mout << endl;
        mout << "Error in Time: " << endl;
        for (int i = 0; i < scales - 1; ++i) {
            mout << V[l*scales+i] - V[l*scales+i+1] << "  ";
        }
        mout << endl;
    }
    mout << "###########################" << endl;
}


void MaindG_Elasticity () {
    Meshes M;
    int dim = M.dim();
    int Scales = 1;
    double ScaleFactor = 0.5;

    bool adaptive = false;
    Config::Get("Scales", Scales);
    Config::Get("ScaleFactor", ScaleFactor);
    Config::Get("adaptive", adaptive);
    double finaleTime = 1;
    bool ProblemExact = 0;
    Config::Get("T", finaleTime);
    Config::Get("ProblemExact", ProblemExact);

    DGElasticityDoF *dof = GetDGElasticityDoF();
    DGElasticityDiscretization disc(dof);
    DGElasticityDoF *dof_FV = GetDGElasticityDoF();
    DGElasticityDiscretization disc_FV(dof_FV);
    CellMatrixGraphs G_vector(M, disc);
    Vector u(G_vector.fine());
    std::string problem_name = "Hat2D";
    Config::Get("Problem", problem_name);
    ProblemElasticity *problem = GetProblemElasticity(disc, problem_name);

   std::vector<Vector*> UU(G_vector.size()*Scales);
   std::vector<Vector*> Uexact(G_vector.size()*Scales);
   std::vector<int> S_Sum(G_vector.size()*Scales);
   std::vector<int> M_Sum(G_vector.size()*Scales);
   std::vector<int> K_Sum(G_vector.size()*Scales);
   std::vector<int> K_Pre(G_vector.size()*Scales);
    Plot plot(M.fine());
    
    DGElasticityAssemble assembleFV(disc_FV,*problem,plot);
    DGElasticityAssemble assemble(disc,*problem,plot);
    int L = G_vector.size()-1;
    RVector GF(0.0,(L+1));
    RVector L2E(0.0,(L+1));
    RVector L2I(0.0,(L+1));
   std::vector<int> NDoFsSpace(L+1);
   std::vector<int> NTSteps(L+1);
   std::vector<int> NDoFs(L+1);
    double s = 1;
    for (int level=0; level<=L; ++level) {
        mout << endl << "Level " << level << endl;
        G_vector.SetLevel(level);
        Vector U(G_vector.fine());
        U.ClearDirichletFlags();
        Vector u_1(G_vector.fine());
        u_1.ClearDirichletFlags();
        u_1 = 0;
        // not working
        CellMatrixGraph G_vectorFV(U.GetMesh(),disc);
        Preconditioner* dGPC = GetdGPC(G_vector,G_vectorFV,assemble,assembleFV);
        Solver S(dGPC);
        Preconditioner* dGPC2= GetdGPC(G_vector,G_vectorFV,assemble,assembleFV);
        Solver S2(dGPC2);
        Preconditioner* dGPC3= GetdGPC(G_vector,G_vectorFV,assemble,assembleFV);
        Solver S3(dGPC3);
        Preconditioner* BJ = GetPC("PointBlockJacobi");
        Solver IM(BJ);
        DG_TimeIntegrator dgrk(S,S2,S3,IM);
        assemble.PrintInfo(U);
        if (Scales > 1) s = 1;
        for (int i=0; i<Scales; ++i) {
            TimeSeries TS(s);
            mout << endl << TS;
            NTSteps[level] = int(finaleTime/TS.SecondTStep());
            Date Start;
            dgrk(assemble,TS,U);
            mout<<"Solving done "<< Date() - Start << endl;
            S_Sum[level*Scales+i] = dgrk.S_Sum();
            M_Sum[level*Scales+i] = dgrk.M_Sum();
            K_Sum[level*Scales+i] = dgrk.K_Sum();
            K_Pre[level*Scales+i] = dgrk.K_Pre();
            UU[level*Scales+i] = new Vector(U);
            //             if (ProblemExact){
            //                 assemble.get_u_int_assemble(finaleTime,u_1);
            //                 Uexact[level*Scales+i]=new Vector(u_1);
            //             }
            s *= ScaleFactor;
        }
        GF[level] = assemble.Goal_Functional(finaleTime,U);
        //L2E[level] = assemble.L2Error(U,finaleTime);
        NDoFsSpace[level] = U.pSize();
        NDoFs[level] = 	NDoFsSpace[level] * NTSteps[level];
        //mout<< "GoalFunctional E(U) = " << scientific << setprecision(10) << GF[level] << endl;
    }
    mout << endl;
    mout <<     "space DoFs           " << NDoFsSpace;
    mout <<     "time steps           " << NTSteps;
    mout <<     "space-time DoFs      " << NDoFs;
    //    PrintValues("GoalFunctional E(U)  ",GF);
    PrintValues("L2 Error             ",L2E);
    PrintValues("L2 Interpolation     ",L2I);
}

void MaindG_Acoustic () {
    Meshes M;
    int dim = M.dim();
    int Scales = 1;
    double ScaleFactor = 0.5;

    bool adaptive = false;
    Config::Get("Scales", Scales);
    Config::Get("ScaleFactor", ScaleFactor);
    Config::Get("adaptive", adaptive);
    double finaleTime = 1;
    double ProblemExact = 0;
    Config::Get("T", finaleTime);
    Config::Get("ProblemExact", ProblemExact);
    DGAcousticDoF *dof = GetDGAcousticDoF();
    DGAcousticDiscretization disc(dof);
    DGAcousticDoF *dof_FV = GetDGAcousticDoF();
    DGAcousticDiscretization disc_FV(dof_FV);

    CellMatrixGraphs G_vector(M, disc);
    Vector u(G_vector.fine());
    std::string problem_name = "Hat2D";
    Config::Get("Problem", problem_name);
    ProblemAcoustic *problem = GetProblemAcoustic(disc, problem_name);

   std::vector<Vector*> UU(G_vector.size()*Scales);
   std::vector<Vector*> Uexact(G_vector.size()*Scales);
   std::vector<int> S_Sum(G_vector.size()*Scales);
   std::vector<int> M_Sum(G_vector.size()*Scales);
   std::vector<int> K_Sum(G_vector.size()*Scales);
   std::vector<int> K_Pre(G_vector.size()*Scales);
    Plot plot(M.fine());
    
    DGAcousticAssemble assembleFV(disc_FV,*problem,plot);
    DGAcousticAssemble assemble(disc,*problem,plot);

    int L = G_vector.size()-1;
    RVector GF(0.0,(L+1));
    RVector L2E(0.0,(L+1)*Scales);
    RVector L2I(0.0,(L+1)*Scales);
   std::vector<int> NDoFsSpace(L+1);
   std::vector<int> NTSteps(L+1);
   std::vector<int> NDoFs(L+1);

    DGAcousticDiscretization* discP;
    discP = &disc;
    InterpolateWave_dG* CT;
//    CT = GetNewCellTransfer(2,discP,discP);

    double s = 1;
    for (int level=0; level<=L; ++level) {
        mout << endl << "Level " << level << endl;
        G_vector.SetLevel(level);
        Vector U(G_vector.fine());
        U.ClearDirichletFlags();
        Vector u_1(G_vector.fine());
        u_1.ClearDirichletFlags();
        u_1 = 0;
        // not working
        CellMatrixGraph G_vectorFV(U.GetMesh(),disc);
        Preconditioner* dGPC = GetdGPC(G_vector,G_vectorFV,assemble,assembleFV);
        Solver S(dGPC);
        Preconditioner* dGPC2= GetdGPC(G_vector,G_vectorFV,assemble,assembleFV);
        Solver S2(dGPC2);
        Preconditioner* dGPC3= GetdGPC(G_vector,G_vectorFV,assemble,assembleFV);
        Solver S3(dGPC3);
        Preconditioner* BJ = GetPC("PointBlockJacobi");
        Solver IM(BJ);
        DG_TimeIntegrator dgrk(S,S2,S3,IM);
        assemble.PrintInfo(U);
        if (level > 0) s /= ScaleFactor;
        for (int i=0; i<Scales; ++i) {
            TimeSeries TS(s);
            mout << endl << TS;
            NTSteps[level] = int(finaleTime/TS.SecondTStep());
            Date Start;
            U=0.0;
            dgrk(assemble,TS,U);
            mout<<"Solving done "<< Date() - Start << endl;
	    return;
            S_Sum[level*Scales+i] = dgrk.S_Sum();
            M_Sum[level*Scales+i] = dgrk.M_Sum();
            K_Sum[level*Scales+i] = dgrk.K_Sum();
            K_Pre[level*Scales+i] = dgrk.K_Pre();
            UU[level*Scales+i] = new Vector(U);
            (*UU[level*Scales+i]).ClearDirichletFlags();
            if (problem->Exact()){
                assemble.get_u_int_assemble(finaleTime,u_1);
                Uexact[level*Scales+i]=new Vector(u_1);
                Vector tmp(U);
                tmp = *UU[level*Scales+i];
                tmp -= *Uexact[level*Scales+i];
                L2I[level*Scales+i] = sqrt(2.0*assemble.Energy(tmp));
                L2E[level*Scales+i] = assemble.L2Error(finaleTime,U);
                if (level != 0){
                    tmp = (*CT) * (*UU[(level-1)*Scales+i]);
                    tmp -= *UU[level*Scales+i];
                    mout << "Error of prolongated mesh" << endl;
                    mout << OUT( sqrt(2.0*assemble.Energy(tmp)) ) << endl;
                }
            }
            if (level != 0 && !problem->Exact()){
                Vector tmp(U);
                tmp = (*CT) * (*UU[(level-1)*Scales+i]);
                tmp -= *UU[level*Scales+i];
                mout << "Error of prolongated mesh" << endl;
                L2E[level*Scales+i] =  sqrt(2.0*assemble.Energy(tmp));
                mout << OUT(L2E[level*Scales+i]) << endl;
            }
            s *= ScaleFactor;
        }
        NDoFsSpace[level] = U.pSize();
        NDoFs[level] = 	NDoFsSpace[level] * NTSteps[level];
        continue;
        //        mout<< "GoalFunctional E(U) = "
        //            << scientific << setprecision(10)
        //            << GF[level] << endl;
    }
    mout << endl;
    mout <<     "space DoFs           " << NDoFsSpace;
    mout <<     "time steps           " << NTSteps;
    mout <<     "space-time DoFs      " << NDoFs;
    PrintValues("GoalFunctional E(U)  ",GF);
    std::cout << std::setprecision(10) << endl;
    if (problem->Exact()) {
        PrintValuesCR("L2 Error             ",L2E);
        PrintValuesCR("L2 Interpolation     ",L2I);
    }
    RVector L2I_fine(0.0,L+1);
    RVector L2E_fine(0.0,L+1);
    mout << "S_Sum"<< endl;
    for (int l=0; l<L+1;++l ){
        mout << "Error Level " << l << endl;
        for (int s = 0;s<Scales; ++s){
            mout << S_Sum[l*Scales + s] << " ";
        }
        mout << endl;
    }
    mout << endl;
    mout << "M_Sum"<< endl;
    for (int l=0; l<L+1;++l ){
        mout << "Error Level " << l << endl;
        for (int s = 0;s<Scales; ++s){
            mout << M_Sum[l*Scales + s] << " ";
        }
        mout << endl;
    }
    mout << endl;
    mout << "K_Sum"<< endl;
    for (int l=0; l<L+1;++l ){
        mout << "Error Level " << l << endl;
        for (int s = 0;s<Scales; ++s){
            mout << M_Sum[l*Scales + s] << " ";
        }
        mout << endl;
    }
    mout << endl;
    mout << "K_Pre"<< endl;
    for (int l=0; l<L+1;++l ){
        mout << "Error Level " << l << endl;
        for (int s = 0;s<Scales; ++s){
            mout << K_Pre[l*Scales + s] << " ";
        }
        mout << endl;
    }
    mout << endl;
    if (problem->Exact()){
        mout << "L2I" << endl;
        for (int l=0; l<L+1;++l ){
            mout << "Error Level " << l << endl;
            for (int s = 0;s<Scales; ++s){
                mout << L2I[l*Scales + s] << " ";
            }
            mout << endl;
        }
        mout << endl;
    }
    mout << "L2E"<< endl;
    for (int l=0; l<L+1;++l ){
        mout << "Error Level " << l << endl;
        for (int s = 0;s<Scales; ++s){
            mout << L2E[l*Scales + s] << " ";
        }
        mout << endl;
    }
    mout << endl;
    for (int l=0; l<L+1;++l){
        L2I_fine[l] = L2I[Scales-1+l*Scales];
        L2E_fine[l] = L2E[Scales-1+l*Scales];
    }
    if (problem->Exact()){
        PrintValues("L2 Interpolation     ",L2I_fine);
    }
    PrintValues("L2 Error             ",L2E_fine);
}


void MaindG_ViscoAcousticX () {
    Meshes M;
    int Scales = 1;
    double ScaleFactor = 0.5;
    bool adaptive = false;
    Config::Get("Scales", Scales);
    Config::Get("ScaleFactor", ScaleFactor);
    Config::Get("adaptive", adaptive);
    double finaleTime = 1;
    Config::Get("T", finaleTime);

    DGVAcousticXDoF *dof = GetDGVAcousticXDoF();
    DGVAcousticXDiscretization disc(dof);
    DGVAcousticXDoF *dof_FV = GetDGVAcousticXDoF();
    DGVAcousticXDiscretization disc_FV(dof_FV);

    CellMatrixGraphs G_vector(M, disc);
    Vector u(G_vector.fine());
    std::string problem_name = "Hat2D";
    Config::Get("Problem", problem_name);
    ProblemVAcousticX* problem = GetProblemVAcousticX(disc,problem_name);

   std::vector<Vector*> UU(G_vector.size()*Scales);
   std::vector<Vector*> Uexact(G_vector.size()*Scales);
   std::vector<int> S_Sum(G_vector.size()*Scales);
   std::vector<int> M_Sum(G_vector.size()*Scales);
   std::vector<int> K_Sum(G_vector.size()*Scales);
   std::vector<int> K_Pre(G_vector.size()*Scales);
    Plot plot(M.fine());
    DGVAcousticXAssemble* assembleFV = GetAssembleX(disc_FV,*problem,plot);
    DGVAcousticXAssemble* assemble   = GetAssembleX(disc,*problem,plot);
//    Transfer* T = new MatrixTransfer(*assemble);
    int L = G_vector.size()-1;
    RVector GF(0.0,(L+1));
    RVector L2E(0.0,(L+1)*Scales);
    RVector L2I(0.0,(L+1)*Scales);
   std::vector<int> NDoFsSpace(L+1);
   std::vector<int> NTSteps(L+1);
   std::vector<int> NDoFs(L+1);
    int SpeedUp =0;
    Config::Get("SpeedUp", SpeedUp);
    std::string pre;
    Config::Get("Preconditioner", pre);
    double s = 1;
    for (int level=0; level<=L; ++level) {
        mout << endl << "Level " << level << endl;
        G_vector.SetLevel(level);
        Vector U(G_vector.fine());
        if (SpeedUp == 1){
            assemble->initializeEleCache(G_vector.fine());
            if(pre == "Multigrid"){
                assembleFV->initializeEleCache(G_vector.fine());
            }
        }
        U.ClearDirichletFlags();
        Vector u_1(G_vector.fine());
        u_1.ClearDirichletFlags();
        u_1 = 0;
        // not working
        CellMatrixGraph G_vectorFV(U.GetMesh(),disc);
        Preconditioner* dGPC = GetdGPC(G_vector,G_vectorFV,*assemble,*assembleFV);
        Solver S(dGPC);
        Preconditioner* dGPC2= GetdGPC(G_vector,G_vectorFV,*assemble,*assembleFV);
        Solver S2(dGPC2);
        Preconditioner* dGPC3= GetdGPC(G_vector,G_vectorFV,*assemble,*assembleFV);
        Solver S3(dGPC3);
        Preconditioner* BJ = GetPC("PointBlockJacobi");
        Solver IM(BJ);
        DG_TimeIntegrator dgrk(S,S2,S3,IM);
        assemble->PrintInfo(U);
        if (level > 0) s /= ScaleFactor;
        //        if (Scales > 1) s = 1;
        for (int i=0; i<Scales; ++i) {
            TimeSeries TS(s);
            mout << endl << TS;
            NTSteps[level] = int(finaleTime/TS.SecondTStep());
            Date Start;
            dgrk(*assemble,TS,U);
            mout<<"Solving done "<< Date() - Start << endl;
            S_Sum[level*Scales+i] = dgrk.S_Sum();
            M_Sum[level*Scales+i] = dgrk.M_Sum();
            K_Sum[level*Scales+i] = dgrk.K_Sum();
            K_Pre[level*Scales+i] = dgrk.K_Pre();
            UU[level*Scales+i] = new Vector(U);

            s *= ScaleFactor;
        }
        GF[level] = assemble->Goal_Functional(finaleTime,U);
        mout<< "GoalFunctional E(U) = "
            << scientific << setprecision(10)
            << GF[level] << endl;
    }
    mout << endl;
    mout <<     "space DoFs           " << NDoFsSpace;
    mout <<     "time steps           " << NTSteps;
    mout <<     "space-time DoFs      " << NDoFs;
    PrintValues("GoalFunctional E(U)  ",GF);
}
template<typename Problem,typename DGAssembleType>
void MaindG_template () {
//    mout << "HALLO" << endl;
    int cell_std_deg = 1;
    Config::Get("deg", cell_std_deg);

    std::string problem_name = "Hat2D";
    Config::Get("Problem", problem_name);

    int Scales = 1;
    double ScaleFactor = 0.5;
    Config::Get("Scales", Scales);
    Config::Get("ScaleFactor", ScaleFactor);

    int SpeedUp = 0;
    Config::Get("SpeedUp", SpeedUp);

    std::string pre;
    Config::Get("Preconditioner", pre);

    Problem *problem = nullptr;
    problem = GetProblem(problem_name, Problem());

    double finaleTime = 1;
    Config::Get("T", finaleTime);

    DoF* dof = GetDoF(problem,cell_std_deg);
    DGTSDiscretization disc(dof,problem->getNumL());
    DoF* dof_FV = GetDoF(problem,cell_std_deg);
    DGTSDiscretization disc_FV(dof_FV,problem->getNumL());

    Meshes M;
//    mout << OUT(M.fine().dim()) << endl;
//    M.PrintMeshWidths();
    CellMatrixGraphs G_vector(M,disc);
    Vector u(G_vector.fine());
   std::vector<Vector*> UU(G_vector.size()*Scales);
   std::vector<Vector*> Uexact(G_vector.size()*Scales);
   std::vector<int> S_Sum(G_vector.size()*Scales);
   std::vector<int> M_Sum(G_vector.size()*Scales);
   std::vector<int> K_Sum(G_vector.size()*Scales);
   std::vector<int> K_Pre(G_vector.size()*Scales);
    Plot plot(M.fine());

    DGAssembleType* assemble = nullptr;
    DGAssembleType* assembleFV =nullptr;
    assemble = GetAssemble(disc,*problem,plot);
    assembleFV =GetAssemble(disc_FV,*problem,plot);
    int L = G_vector.size()-1;
    RVector GF(0.0,(L+1));
    RVector L2E(0.0,(L+1)*Scales);
    RVector L2T(0.0,Scales);
    RVector L2I(0.0,(L+1)*Scales);
   std::vector<int> NDoFsSpace(L+1);
   std::vector<int> NTSteps(L+1);
   std::vector<int> NDoFs(L+1);

    double s = 1;
    for (int level=0; level<=L; ++level) {
        mout << endl << "Level " << level << endl;
        G_vector.SetLevel(level);
        Vector U(G_vector.fine());
        if (SpeedUp == 1){
            assemble->initializeEleCache(G_vector.fine());
            if(pre == "Multigrid"){
                assembleFV->initializeEleCache(G_vector.fine());
            }
        }
        U.ClearDirichletFlags();
        Vector u_1(G_vector.fine());
        u_1.ClearDirichletFlags();
        u_1 = 0;
        // not working
        CellMatrixGraph G_vectorFV(U.GetMesh(),disc);
        Preconditioner* dGPC = GetdGPC(G_vector,G_vectorFV,*assemble,*assembleFV);
        Solver S(dGPC);
        Preconditioner* dGPC2= GetdGPC(G_vector,G_vectorFV,*assemble,*assembleFV);
        Solver S2(dGPC2);
        Preconditioner* dGPC3= GetdGPC(G_vector,G_vectorFV,*assemble,*assembleFV);
        Solver S3(dGPC3);
        Preconditioner* BJ = GetPC("PointBlockJacobi");
//        Preconditioner* BJ = GetPC("CellJacobi");
        Solver IM(BJ);
        DG_TimeIntegrator dgrk(S,S2,S3,IM);
        assemble->PrintInfo(U);
        if (level > 0) s /= ScaleFactor;
        //        if (Scales > 1) s = 1;
        for (int i=0; i<Scales; ++i) {
            TimeSeries TS(s);
            mout << endl << TS;
            NTSteps[level] = int(finaleTime/TS.SecondTStep());
            Date Start;
            dgrk(*assemble,TS,U);
            mout<<"Solving done "<< Date() - Start << endl;
            S_Sum[level*Scales+i] = dgrk.S_Sum();
            M_Sum[level*Scales+i] = dgrk.M_Sum();
            K_Sum[level*Scales+i] = dgrk.K_Sum();
            K_Pre[level*Scales+i] = dgrk.K_Pre();
            UU[level*Scales+i] = new Vector(U);
            if (problem->Exact()){
                Vector u_tmp(u);
                u_tmp=0.0;
                assemble->get_u_int_assemble(finaleTime,u_1);
                Vector u_tmp2(u_1);
                u_tmp2-=u_tmp;
                Uexact[level*Scales+i]=new Vector(u_1);
                Vector tmp(U);
                tmp = *UU[level*Scales+i];
                tmp -= *Uexact[level*Scales+i];
                L2I[level*Scales+i] = sqrt(2.0*assemble->Energy(tmp));
                L2E[level*Scales+i] = assemble->L2Error(finaleTime,U);

            }
            if (i>0){
                Vector tmp2(U);
                tmp2 = *UU[level*Scales +i];
                tmp2 -= *UU[level*Scales +i-1];
                L2T[i] = assemble->Norm(tmp2);
            }
            s *= ScaleFactor;

        }
        for ( int i=1; i<Scales-1; i++){
            if( i==1) mout << "EOCS:" <<endl;
            if (PPM->proc() == 0){
                printf("%1.10f \n",L2T[i]/L2T[i+1]);
            }
        }
        GF[level] = assemble->Goal_Functional(finaleTime,U);
        mout<< "GoalFunctional E(U) = "
            << scientific << setprecision(10)
            << GF[level] << endl;
    }
    mout << endl;
    for ( int i=0; i<Scales; i++){
        for ( int level=0; level<L; level++){
            mout << "EOCS:" <<endl;
            if (PPM->proc() == 0){
                printf("%1.10f \n",L2I[(level+1)*Scales+i]/L2I[level*Scales+i]);
            }
        }
    }
//    mout <<     "space DoFs           " << NDoFsSpace;
//    mout <<     "time steps           " << NTSteps;
//    mout <<     "space-time DoFs      " << NDoFs;
//    PrintValues("GoalFunctional E(U)  ",GF);
//    PrintValues("GoalFunctional E(U)  ",GF);
    std::cout << std::setprecision(10) << endl;
    if (problem->Exact()) {
        PrintValuesCR("L2 Error             ",L2E);
        PrintValuesCR("L2 Interpolation     ",L2I);
    }
    return;

}
void MaindG_ViscoAcoustic() {
    MaindG_template<ProblemVAcoustic,DGVAcousticAssemble>();
}
void MaindG_ViscoElastic() {
    MaindG_template<ProblemVElastic,DGVElasticAssemble>();
}
