import numpy as np

from numpy import pi, abs, sqrt, exp, real, conj, nan_to_num, heaviside
from numpy.fft import fft, ifft, fftfreq, fftshift, ifftshift, irfft

import scipy as sp
import scipy.signal
import scipy.signal
import scipy.special
from scipy.special import hankel1, hankel2

import matplotlib.pyplot as plt
import sys
import copy

from seismogram_lib import FWISeismogram, plot_phase_spectrum, plot_magnitude_spectrum


# --------------------------------
# time domain

def H(x):
    return heaviside(x, 0.0)

def G(t,r, shift=0, c=1):
    t = t-shift
    v = 1.0/sqrt(t**2 - r**2/c**2) * H(t - abs(r)/c)
    return 1.0/(2*pi)*nan_to_num(v)

def dt_gauss(t, sigma=1):
    return -2*t*exp(-(t/sigma)**2/2) / (sigma*sqrt(2*pi))

def ricker(t, freq): 
    a = (pi**2)*(freq**2)
    return (1.-2.*a*t**2)*exp(-a*t**2)

def dt_ricker(t, freq):
    a = (pi**2)*(freq**2)
    return (4*a**2*t**3 - 6*a*t) * exp(-a*t**2)


def get_pressure_trace(conf, receiver_pos):
    dt = conf["dt"]
    r = np.linalg.norm(conf["source_pos"]-receiver_pos)
    t_solution = np.arange(*conf["time_interval"], dt)
    t_source = np.arange(*conf["source_time_interval"], dt/100)
    t_padded = np.arange(conf["time_interval"][0]+conf["source_time_interval"][0], 
                         conf["time_interval"][1]+conf["source_time_interval"][1], dt/100)

    shift = conf["source_t_offset"]
    green = G(t_padded,r, shift=shift, c=conf["c"])
    source = conf["rho"]*conf["source_func"](t_source)
    
    #print("AAA", fft(green)*(dt/100))
    
    solution = sp.signal.fftconvolve(green, source, mode="valid")*(dt/100)
    solution = solution[::100]

    #plt.plot(t_padded,green); plt.show(); sys.exit()
    #plt.plot(t_source,source); plt.show(); sys.exit()
    #plt.plot(solution); plt.show(); sys.exit()
    
    return t_solution, solution
    

# --------------------------------
# Frequency domain

def ricker_freq(omega, freq):
    omega_p = 2*pi*freq
    omega2 = omega**2
    return omega2/(2. * pi**2.5 * freq**3) * exp(-omega2/omega_p**2)

def dt_ricker_freq(omega,freq):
    r = ricker_freq(omega,freq)
    return 1j*omega * r

def delta_freq(omega, freq):
    omega_p = 2*pi*freq
    return exp(-omega_p)

def dt_delta_freq(omega,freq):
    r = delta_freq(omega,freq)
    return 1j*omega * r


def G_freq(omega, r, c=1.):
    # Note: the magic constant 0.2781899 is obtained experimetally by transforming the time-domain Greens function
    # Calzone says that setting the 0-th order term to 0 does not change much since the corresponding part in the source is usually small

    #return np.hstack(([0.2781899], 1j * pi * hankel1(0, arg[1:]) / (4*pi)))
        
    arg = -(r/c)*omega
    N = (len(omega)-1)//2
    arg_pos = arg[1:N+1]
    G_pos = 1j * pi * hankel1(0, arg_pos) / (4*pi)
    
    G = np.hstack(([0.2781899], G_pos, G_pos.conj()[::-1]))
    return G
    
    #return -1j * pi * nan_to_num(hankel2(0, arg)) / (4*pi)


def get_pressure_trace_freq(conf,dim, receiver_pos):
    dt = conf["dt"]
    r = np.linalg.norm(conf["source_pos"]-receiver_pos)
    
    t_min, t_max = conf["time_interval"]
    t_solution = np.linspace(t_min,t_max, int(round((t_max-t_min)/dt))+1)
    
    N, fac = len(t_solution), 7
    fft_omegas = (2.*pi) * fftfreq(N*fac, dt/fac)
    #N_pos = (len(fft_omegas)-1)//2
    #fft_omegas = fft_omegas[0:N_pos+1]
    
    fft_green = G_freq(fft_omegas, r, c=conf["c"])
    
    fft_source = exp(1j*t_min * fft_omegas) * conf["rho"] \
    #                * dt_ricker_freq(fft_omegas,conf["f"])
    
    # diract delta rhs
    #fft_source = exp(1j*t_min * fft_omegas) * conf["rho"] * 1
    
    
    #fft_source = exp(1j*t_min * fft_omegas) * conf["rho"] \
                    #* dt_delta_freq(fft_omegas,conf["f"])
                
    #fft_source *= exp(-1.j*conf["source_t_offset"] * fft_omegas) # time offset of source
    
    fft_solution = fft_green * fft_source 
    solution_from_fft = (ifft(fft_solution)) / (dt/fac)
    #solution_from_fft = (irfft(fft_solution, n=N_pos*2+1)) / (dt/fac)
    
    solution_from_fft = solution_from_fft[::fac]

    #plt.plot(t_padded,green); plt.show(); sys.exit()
    #plt.plot(t_source,source); plt.show(); sys.exit()
    #plt.plot(solution); plt.show(); sys.exit()
    
    return t_solution, solution_from_fft

# --------------------------------


def create_pressure_seismogram(seismogram, conf,dim):
    conf = conf.copy()
    spec = seismogram.spec
    conf["time_interval"] = [spec.t_start, spec.t_end]
    conf["dt"] = spec.t_delta
    
    new_s = copy.deepcopy(seismogram)
    for ts in new_s.timeseries:
        receiver_pos = np.array([ts.receiver.x, ts.receiver.y])
        print(receiver_pos)
        #t, p = get_pressure_trace(conf, receiver_pos=receiver_pos)
        t, p = get_pressure_trace_freq(conf,dim, receiver_pos=receiver_pos)
        print(min(p), max(p), min(ts.states), max(ts.states))
        ts.states[0:len(p),0] = p.copy()

    return new_s

dim = 1
conf = {}
if dim is 1:
    comp = 1
    conf["c"] = 5.
    conf["rho"] = 1.
    conf["f"] = 25.0
    conf["source_pos"] = np.array([600, 1000]),
    conf["source_t_offset"] = 1.5/conf["f"]
    conf["source_func"] = lambda t: dt_ricker(t, freq=conf["f"])
    conf["source_time_interval"] = [-2./conf["f"],2./conf["f"]]
if dim is 2:
    comp = 2
    conf["c"] = 3500.
    conf["rho"] = 2000.
    conf["f"] = 25.0
    conf["source_pos"] = np.array([600, 1000]),
    conf["source_t_offset"] = 1.5/conf["f"]
    conf["source_func"] = lambda t: dt_ricker(t, freq=conf["f"])
    conf["source_time_interval"] = [-2./conf["f"],2./conf["f"]]


s = FWISeismogram.readFromFile(sys.argv[1]).extractComponent(comp)

s_sol = create_pressure_seismogram(s, conf,dim)

d_min, d_max = np.inf, -np.inf
for seis in [s, s_sol]:
    d_min = min(d_min, seis.min())
    d_max = max(d_max, seis.max())

plot_spectra = 0
if plot_spectra:
    print("WARNING: hardcoded stuff....")
    
    plot_magnitude_spectrum([s_sol], 0, figsize=(6,4), xlim=(0,270), ylim=(0,2000), filename="exact_amplitude.pdf")
    
    plot_phase_spectrum([s_sol], 0, figsize=(6,4), xlim=(0,250), ylim=(-350,0), filename="exact_phase.pdf")
    
else:
    figsize=(3.5,10)
    figure = plt.figure(figsize=figsize)
    #data_factor = "normalized"
    data_factor = 2
    s.plot(data_factor=data_factor, data_range=(d_min,d_max),
        figure=figure, show=False, fill_below=False, r_step=3)
    s_sol.plot(data_factor=data_factor, data_range=(d_min,d_max),
            figure=figure, show=False, fill_below=False,
            linestyle="--", r_step=3)

    diff = s_sol - s
    print("Rel. L2-Difference: ", diff.norm2()/s_sol.norm2())

    plt.show()
