from __future__ import print_function

import glob
import os
import os.path
import sys
import time


def make_vector(s):
    return tuple(map(float,s.strip().split()))


def write_vtk(fn, points, cells):
    with open(fn, "w") as f:
        f.write("# vtk DataFile Version 2.0\nUnstructured Grid by M++\nASCII\nDATASET UNSTRUCTURED_GRID\n")
        
        print("Writing points...")
        
        f.write("POINTS {} float\n".format(len(points)))
        for p in points:
            f.write("{:15.10f}\t{:15.10f}\t{:15.10f}\n".format(*p))
            
        point_count = sum(len(c)+1 for c in cells)
        f.write("CELLS {} {}\n".format(len(cells), point_count))
        
        print("Writing cells...")
        
        for c in cells:
            f.write("{} {}\n".format(len(c), " ".join(map(str, c))))
            
        print("Writing cell types...")
            
        cell_types = {
            8: 12
            }
        
        f.write("CELL_TYPES {}\n".format(len(cells)))
        for c in cells:
            f.write("{}\n".format(cell_types[len(c)]))
            
        print("Writing cell data...")
            
        f.write("CELL_DATA {}\nSCALARS scalar_value float 1\nLOOKUP_TABLE default".format(len(cells)))
        for c in cells:
            d = cells[c]
            f.write("{:15.10f}\n".format(d[0]))
        
        
def read_file(input_fn):
    points, cells = set(), {}
    
    with open(input_fn) as f:
        for l in f:
            l = l.strip()
            center, rest = l.split(":", 1)
            corners, data = rest.split("\t-\t", 1)
            corners = tuple(map(make_vector, corners.strip().split("\t")))
            data = make_vector(data)
                  
            points |= set(corners)
            cells[corners] = data
            
    return points, cells


def read_data(input_dir, limit, skip):
    points, cells = {}, {}
    cell_list = []

    def key(s):
        return int(s.split("_")[-1][:-4])

    files = sorted(glob.glob(os.path.join(input_dir, "P*.txt")), key=key)
    files_to_read = min(len(files), limit)
    print("Reading {} files of {} present ...".format(files_to_read, len(files)))

    read_files = 0
    for fn in files:
        read_files += 1
        if read_files <= skip:
            print("Skipping file nr {}: {}".format(read_files, fn))
            continue
        else:
            print("Reading file nr {}: {}".format(read_files, fn))
                            
        f_points, f_cells = read_file(fn)
        
        for p in f_points:
            if p not in points:
                points[p] = len(points)
                
        for cell,data in f_cells.items():
            cells[tuple(points[c] for c in cell)] = data
            
        if read_files - skip >= files_to_read:
            break
                
    return points, cells


if __name__ == "__main__":
    import argparse
        
    parser = argparse.ArgumentParser(description='Merges parallel plot files into vtk-File')
    parser.add_argument("input_dir", type=str, nargs="?", default="data/parallel_plotting/", help="path to input files")
    parser.add_argument("output_fn", type=str, nargs="?", default="data/vtk/merged.vtk", help="path to the output file")

    parser.add_argument("-s", metavar="s", type=int, default=0, help="skip first <s> files")
    parser.add_argument("-l", metavar="l", type=int, default=100000, help="only use the first <l> files")
    args = parser.parse_args()


    print("Merging files in ", args.input_dir, " to ", args.output_fn)
    points, cells = read_data(args.input_dir, args.l, args.s)

    print("Writing output to ", args.output_fn)
    write_vtk(args.output_fn, points, cells)
