from __future__ import print_function

import glob
import os
import os.path
import sys
import concurrent.futures



def make_vector(s):
    return tuple(map(float, s.strip().split()))


def write_vtk(fn, points, cells):
    with open(fn, "w") as f:
        f.write("# vtk DataFile Version 2.0\nUnstructured Grid by M++\nASCII\nDATASET UNSTRUCTURED_GRID\n")
        
        print("Writing points...")
        
        f.write("POINTS {} float\n".format(len(points)))
        for p in points:
            f.write("{:15.10f}\t{:15.10f}\t{:15.10f}\n".format(*p))
            
        point_count = sum(len(c)+1 for c in cells)
        f.write("CELLS {} {}\n".format(len(cells), point_count))
        
        print("Writing cells...")
        
        for c in cells:
            f.write("{} {}\n".format(len(c), " ".join(str(p) for p in c)))
            
        print("Writing cell types...")
            
        cell_types = {
            8: 12
            }
        
        f.write("CELL_TYPES {}\n".format(len(cells)))
        for c in cells:
            f.write("{}\n".format(cell_types[len(c)]))
            
        print("Writing cell data...")
            
        f.write("CELL_DATA {}\nSCALARS scalar_value float 1\nLOOKUP_TABLE default".format(len(cells)))
        for c in cells:
            d = cells[c]
            f.write("{:15.10f}\n".format(d[0]))
        
        
def read_file(input_fn):
    points, cells = set(), set()
    
    print("Reading ", input_fn)
    
    with open(input_fn) as f:
        for l in f:
            l = l.strip()
            center, rest = l.split(":")
            corners, data = rest.split("\t-\t")
            corners = tuple(map(make_vector, corners.strip().split("\t")))
            data = make_vector(data)
            
            points |= set(corners)
            
            cells.add((corners, data))
            
    return points, cells


def read_data(input_dir, limit, skip):
    points, cells = {}, {}

    def key(s):
        return int(s.split("_")[-1][:-4])

    files = sorted(glob.glob(os.path.join(input_dir, "P*.txt")), key=key)
    files_to_read = min(len(files), limit)
    print("Reading {} files of {} present ...".format(files_to_read, len(files)))
    
    files = files[skip:skip+limit]
      
    with concurrent.futures.ProcessPoolExecutor() as executor:
        future_to_fn = {executor.submit(read_file, fn): (i,fn) for i,fn in enumerate(files, start=skip)}
        
        for future in concurrent.futures.as_completed(future_to_fn):
            i, fn = future_to_fn[future]
            try:
                f_points, f_cells = future.result()
            except Exception as exc:
                print("File nr {}: {} generated an exception: {}".format(i, fn, exc))
            else:
                print("Read file nr {}: {}".format(i, fn))
                
                for p in f_points:
                    if p not in points:
                        points[p] = len(points)
                 
                for cell,data in f_cells:
                    cells[tuple(points[c] for c in cell)] = data
    
    #print("Numbering {} points ... ".format(len(points)))
    ##for i,p in enumerate(points):
    ##    points[p] = i
    #points = {p:i for i,p in enumerate(points)}
    #print(points)
    #print("done")
    
    #numbered_cells = {}
    #print("Collecting cells ... ")
    ##for cell,data in cells.items():
        ##numbered_cells[tuple(points[c] for c in cell)] = data
    #numbered_cells = {tuple(points[c] for c in cell): data for cell,data in cells}
    #print("done")
    
    return points, cells


if __name__ == "__main__":
    import argparse
        
    parser = argparse.ArgumentParser(description='Merges parallel plot files into vtk-File')
    parser.add_argument("input_dir", type=str, nargs="?", default="data/parallel_plotting/", help="path to input files")
    parser.add_argument("output_fn", type=str, nargs="?", default="data/vtk/merged.vtk", help="path to the output file")

    parser.add_argument("-s", metavar="s", type=int, default=0, help="skip first <s> files")
    parser.add_argument("-l", metavar="l", type=int, default=100000, help="only use the first <l> files")
    args = parser.parse_args()


    print("Merging files in ", args.input_dir, " to ", args.output_fn)
    points, cells = read_data(args.input_dir, args.l, args.s)

    print("Writing output to ", args.output_fn)
    write_vtk(args.output_fn, points, cells)
