import matplotlib.pyplot as plt
import numpy as np
from mpl_toolkits.axes_grid1 import make_axes_locatable
from matplotlib.colors import Normalize

def QGSLS(omegas,tauls,taup):
    Qomega = np.zeros(omegas.shape)
    for i,omega in enumerate(omegas):
        sum2 = 0.0
        sum1 = 0.0
        for taul in tauls:
            prod = omega * taul/(1.0+omega*taul*omega*taul)
            sum1 += omega * taul * prod
            sum2 += prod
        Qomega[i] = (1.0/taup + sum1)/sum2
    return Qomega
class MidpointNormalize(Normalize):
    def __init__(self, vmin=None, vmax=None, midpoint=None, clip=False):
        self.midpoint = midpoint
        Normalize.__init__(self, vmin, vmax, clip)

    def __call__(self, value, clip=None):
        x, y = [self.vmin, self.midpoint, self.vmax], [0, 0.5, 1]
        return np.ma.masked_array(np.interp(value, x, y))


def savePlot(Img, filename, vmin, vmax, title="", colorbar=True, show_axis=False, ticksmiddle=None, text=None,
             norm=None, cmap="viridis"):

    plt.clf()
    plt.cla()
    fig, ax = plt.subplots()

    [i.set_linewidth(0.2) for i in plt.gca().spines.values()]

    ax.get_xaxis().set_visible(show_axis)
    ax.get_yaxis().set_visible(show_axis)
    if norm is not None:
        imgplot = plt.imshow(Img, interpolation='nearest', norm=norm)
    else:
        imgplot = plt.imshow(Img, interpolation='nearest', vmin=vmin, vmax=vmax, norm=norm)
    imgplot.set_cmap(cmap)
    if colorbar:
        # cb = plt.colorbar(imgplot, shrink=.45, pad=.02, aspect=10)
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("top", size="5%", pad=0.1)
        if ticksmiddle is None:
            tickz = [vmin, vmax]
        else:
            tickz = [vmin]
            tickz.extend(ticksmiddle)
            tickz.append(vmax)
        cb = plt.colorbar(imgplot, cax=cax, orientation='horizontal', ticks=tickz)
        cax.xaxis.set_ticks_position("top")
        cb.outline.set_linewidth(0.2)

    if title:
        # plt.title(title)
        ax.text(0.05, 0.05, title,
                verticalalignment='bottom', horizontalalignment='left',
                transform=ax.transAxes, color='white', fontsize=14)

    if text:
        # plt.title(title)
        ax.text(0.05, 0.03, text,
                verticalalignment='bottom', horizontalalignment='left',
                transform=ax.transAxes, color='white', fontsize=14)

    # fig.set_size_inches(6,4)
    plt.savefig(filename, dpi=300, bbox_inches="tight", pad_inches=0.02)
    plt.close()


def createImageMapping(centers):
    def take_first_of_first(elem):
        return elem[0][0]

    def take_second_of_first(elem):
        return elem[0][1]

    return sorted(sorted([(centers.points[i, 0:2], i) for i in range(0, len(centers.points))],
                         key=take_first_of_first), key=take_second_of_first)


def get_coordinate_values(x_list,y_list, bounding_box_):
  #  x_set = sorted(set([points[j][0] for j in range(len(points))]))
  #  y_set = sorted(set([points[j][1] for j in range(len(points))]))
    x_values_full = len(x_list)
    if bounding_box_ is not None:
        x_values_new = [j for j in range(x_values_full) if bounding_box_[0][0] <= x_list[j] <= bounding_box_[0][1]]
        y_values_new = [j for j in range(len(y_list)) if bounding_box_[1][0] <= y_list[j] <= bounding_box_[1][1]]
        return x_values_full, x_values_new, y_values_new
    else:
        return x_values_full, [j for j in range(x_values_full)], [j for j in range(len(y_list))]

def preprocess(points):
    x = sorted([p[0] for p in points])
    y = sorted([p[1] for p in points])
    x_ = [x[0]]
    y_ = [y[0]]
    xcur = x[0]
    ycur = y[0]
    for i in range(len(x)):
        if (abs(xcur - x[i])) > 1e-4:
            xcur = x[i]
            x_.append(xcur)
    for i in range(len(y)):
        if (abs(ycur - y[i])) > 1e-4:
            ycur = y[i]
            y_.append(ycur)
    return x_, y_
def vtuMapping(vtu, bounding_box):
    centers = vtu.cell_centers()
    points = centers.points

    imageMapping = createImageMapping(centers)
    x,y = preprocess(points)
    x_val_f, x_val, y_val = get_coordinate_values(x, y, bounding_box)
    return x_val_f, x_val, y_val, imageMapping


def fillImage(cellData, imageMapping, xFullLength, yValues, xValues):
    image = np.zeros((len(yValues), len(xValues)), dtype=float)
    for i in range(0, len(yValues)):
        for j in range(0, len(xValues)):
            image[i][j] = cellData[imageMapping[yValues[i] * xFullLength + xValues[j]][1]]
    return image

def fillImageAndMapping(vtu, comp, bounding_box):
    x_full_length, x_values, y_values, image_mapping = vtuMapping(vtu, bounding_box)
    image = np.zeros((len(y_values), len(x_values)), dtype=float)
    for i in range(0, len(y_values)):
        for j in range(0, len(x_values)):
            image[i][j] = vtu.cell_data[comp][image_mapping[y_values[i] * x_full_length + x_values[j]][1]]
    return image

def fillCelldata(cellData, image, imageMapping, xFullLength, yValues, xValues):
    test = np.zeros(cellData.shape)
    for i in range(0, len(yValues)):
        for j in range(0, len(xValues)):
            test[imageMapping[yValues[i] * xFullLength + xValues[j]][1]] = image[j][i]
  #          cellData[imageMapping[yValues[i] * xFullLength + xValues[j]][1]] = image[i][j]
    return test

def calculateErrors(image):
    error_l2 = np.linalg.norm(image, 'fro')
    error_l_inf = max(abs(image.max()), abs(image.min()))
    return error_l2, error_l_inf
