import numpy as np
import multiprocessing as mp
from numpy import pi, abs, sqrt, exp, real, conj, nan_to_num, heaviside
from numpy.fft import fft, ifft, fftfreq, fftshift, ifftshift

import scipy as sp
import scipy.signal
import scipy.signal
import scipy.special
from scipy.special import hankel1, hankel2

import matplotlib.pyplot as plt
import sys
import copy

from pprint import pprint

from seismogram_lib import FWISeismogram, plot_phase_spectrum, plot_magnitude_spectrum

from functools import wraps
import time


def timeit(func):
    @wraps(func)
    def timeit_wrapper(*args, **kwargs):
        start_time = time.perf_counter()
        result = func(*args, **kwargs)
        end_time = time.perf_counter()
        total_time = end_time - start_time
        print(f'Function {func.__name__} Took {total_time:.4f} seconds')
        return result

    return timeit_wrapper


# --------------------------------
# Frequency domain

def ricker_freq(omega, freq):
    omega_p = 2 * pi * freq
    omega2 = omega ** 2
    return omega2 / (2. * pi ** 2.5 * freq ** 3) * exp(-omega2 / omega_p ** 2)


def dt_ricker_freq(omega, freq):
    r = ricker_freq(omega, freq)
    return 1j * omega * r

def dt_n_ricker_freq(omega, freq,n):
    r = ricker_freq(omega, freq)
    return 1.j**n * omega ** n * r

def delta_freq(omega, freq):
    return exp(-2j * pi * freq)


def dt_delta_freq(omega, freq):
    r = delta_freq(omega, freq)
    return 1j * omega * r


def GVA_freq_old(omega, r, rho, f0, f, c0, tau_p):
    def Kr_fct(rho, c0, omega0, tau_l, tau_p):
        alpha = 0.
        for l in range(len(tau_l)):
            alpha += omega0 ** 2 * tau_l[l] ** 2 / (1. + omega0 ** 2 * tau_l[l] ** 2)

        return c0 ** 2 * rho / (1. + alpha * tau_p)

    def c_fct(omega, rho, omega0, c0, tau_l, tau_p):
        sum1 = 0.
        for l in range(len(tau_l)):
            sum1 += omega ** 2 * tau_l[l] ** 2 / (1. + omega ** 2 * tau_l[l] ** 2)

        Kr = Kr_fct(rho, c0, omega0, tau_l, tau_p)
        return sqrt(Kr / rho * (1. + sum1 * tau_p))

    def Q_fct(omega, tau_l, tau_p):
        sum1, sum2 = 0., 0.
        for l in range(len(tau_l)):
            sum1 += omega ** 2 * tau_l[l] ** 2 / (1. + omega ** 2 * tau_l[l] ** 2)
            sum2 += omega * tau_l[l] / (1. + omega ** 2 * tau_l[l] ** 2)

        return (1. + sum1 * tau_p) / (sum2 * tau_p)

    omega0 = 2 * pi * f0
    tau_l = 1. / (2. * pi * f)

    c_omega = c_fct(omega, rho, omega0, c0, tau_l, tau_p)
    k = omega / c_omega
    arg = -k * r

    acoustic_G = -1j * pi * nan_to_num(hankel2(0, arg)) / (4 * pi)  # see Carcione

    norm_omega = abs(omega)
    Q = Q_fct(omega, tau_l, tau_p)
    va_correction = exp(-r * omega / (2 * c_omega * Q))

    return acoustic_G * va_correction


def GVA_freq(omega, r, rho, f0, f, c0, tau_p):
    # Note: the magic constant 0.2781899 is obtained experimetally by transforming the time-domain Greens function
    # Calzone says that setting the 0-th order term to 0 does not change much since the corresponding part in the source is usually small

    # return np.hstack(([0.2781899], 1j * pi * hankel1(0, arg[1:]) / (4*pi)))

    def c_fct(omega, rho, omega0, c0, tau_l, tau_p):
        sum1, sum2 = 0., 0.
        for l in range(len(tau_l)):
            sum1 += 1j * omega * tau_l[l] / (1. + 1j * omega * tau_l[l])
            sum2 += omega0 ** 2 * tau_l[l] ** 2 / (1. + omega0 ** 2 * tau_l[l] ** 2)

        Kr = c0 ** 2 * rho / (1. + sum2 * tau_p)  # Bohlen
        Kc = Kr * (1. + sum1 * tau_p)
        return sqrt(Kc / rho)

    omega0 = 2 * pi * f0
    tau_l = 1. / (2. * pi * f)

    c_omega = c_fct(omega, rho, omega0, c0, tau_l, tau_p)
    arg = omega * r / c_omega

    acoustic_G = -1j * pi * nan_to_num(hankel2(0, arg)) / (4 * pi)  # see Carcione

    # if len(tau_l) == 1: # L = 1
    #     zero_coeff = 0.2781899
    # elif len(tau_l) == 5: # L = 5
    # else:
    #     zero_coeff = 0.0
    #
    # if zero_coeff != 0.0:
    #     print("WARNING: setting Fourier coefficient of 0th freq to {}".format(zero_coeff))
    zero_coeff = 0
    zero_coeff = 0.2781899

    acoustic_G[0] = zero_coeff

    return acoustic_G


def GVA_freq(omega, r, rho, f0, f, c0, tau_p):
    # Note: the magic constant 0.2781899 is obtained experimetally by transforming the time-domain Greens function
    # Calzone says that setting the 0-th order term to 0 does not change much since the corresponding part in the source is usually small

    # return np.hstack(([0.2781899], 1j * pi * hankel1(0, arg[1:]) / (4*pi)))

    def c_fct(omega, rho, omega0, c0, tau_l, tau_p):
        sum1, sum2 = 0., 0.
        for l in range(len(tau_l)):
            sum1 += 1j * omega * tau_l[l] / (1. + 1j * omega * tau_l[l])
            sum2 += omega0 ** 2 * tau_l[l] ** 2 / (1. + omega0 ** 2 * tau_l[l] ** 2)

        Kr = c0 ** 2 * rho / (1. + sum2 * tau_p)  # Bohlen
        Kc = Kr * (1. + sum1 * tau_p)
        return sqrt(Kc / rho)

    omega0 = 2 * pi * f0
    tau_l = 1. / (2. * pi * f)

    c_omega = c_fct(omega, rho, omega0, c0, tau_l, tau_p)
    arg = omega * r / c_omega

    acoustic_G = -1j * pi * nan_to_num(hankel2(0, arg)) / (4 * pi)  # see Carcione

    # if len(tau_l) == 1: # L = 1
    #     zero_coeff = 0.2781899
    # elif len(tau_l) == 5: # L = 5
    # else:
    #     zero_coeff = 0.0
    #
    # if zero_coeff != 0.0:
    #     print("WARNING: setting Fourier coefficient of 0th freq to {}".format(zero_coeff))
    zero_coeff = 0
    # zero_coeff = 0.2781899

    acoustic_G[0] = zero_coeff

    return acoustic_G


def get_pressure_trace_freq_VA(receiver_pos, dt, time_interval, source_pos, source_t_offset, rho, f0, freqs, c0, tau_p,
                               minimal_freq_count=None, reduceAfterFFT=True):
    r = np.linalg.norm(source_pos - receiver_pos)
    t_min, t_max = time_interval
    t_solution = np.linspace(t_min, t_max, int(round((t_max - t_min) / dt)) + 1)
    N, fac = len(t_solution), 1
    if minimal_freq_count is not None:
        while N * fac < minimal_freq_count:
            fac += 1

    fft_omegas = (2. * pi) * fftfreq(N * fac, dt / fac)

    fft_green = GVA_freq(fft_omegas, r,
                         rho=rho, f0=f0,
                         f=freqs,
                         c0=c0, tau_p=tau_p)

    # fft_source = exp(1j*t_min * fft_omegas) * rho \
    #                * dt_ricker_freq(fft_omegas, f0)

    # fft_source = exp(1j*t_min * fft_omegas)  * rho

    # fft_source = exp(1j * t_min * fft_omegas) * rho \
    #              * ricker_freq(fft_omegas, f0)

    # shift source according to time offset
    # fft_source *= exp(-1.j * source_t_offset * fft_omegas)

    fft_solution = dt_delta_freq(fft_omegas, fftfreq(N * fac, dt / fac)) * fft_green  # * fft_source
    #    fft_solution = fft_green * fft_source

    solution_from_fft = real(ifft(fft_solution)) / (dt / fac)
    if reduceAfterFFT is True:
        solution_from_fft = solution_from_fft[::fac]
    else:
        t_solution = np.linspace(t_min, t_max, len(solution_from_fft))

    return t_solution, solution_from_fft


# --------------------------------
def getAdjoint(seismogram, source_pos, source_t_offset, rho, f0, freqs, c0, tau_p, L,
               minimal_freq_count=None, time_interval=None):
    spec = seismogram.spec
    new_s = copy.deepcopy(seismogram)
    receiver_pos = np.array([new_s.timeseries[0].receiver.x, new_s.timeseries[0].receiver.y])
    sample_pos = (receiver_pos + source_pos) * 0.5
    rSrc = np.linalg.norm(sample_pos - source_pos)
    rReceive = np.linalg.norm(sample_pos - receiver_pos)

    if time_interval is None:
        time_interval = [spec.t_start, spec.t_end]
    dt = spec.t_delta

    t_min, t_max = time_interval
    t_solution = np.linspace(t_min, t_max, int(round((t_max - t_min) / dt)) + 1)
    N, fac = len(t_solution), 1
    if minimal_freq_count is not None:
        while N * fac < minimal_freq_count:
            fac += 1

    fft_omegas = (2. * pi) * fftfreq(N * fac, dt / fac)
    fft_green = GVA_freq(fft_omegas, rSrc, rho=rho, f0=f0, f=freqs, c0=c0, tau_p=tau_p)
    fft_source = exp(1j * t_min * fft_omegas) * rho * dt_n_ricker_freq(fft_omegas, f0, 1)
    fft_source *= exp(-1.j * source_t_offset * fft_omegas)
    fft_solution = fft_green * fft_source
    solution_from_fft = real(ifft(fft_solution)) / (dt / fac)
    t_solution = np.linspace(t_min, t_max, len(solution_from_fft))
    solution_from_fft = solution_from_fft[::-1]
    fft_green2 = GVA_freq(fft_omegas,  rReceive, rho=rho, f0=f0, f=freqs, c0=c0, tau_p=tau_p)
    nrproc = 8
    adj = np.zeros(len(t_solution[::fac]))
    t_split = np.array_split(t_solution[::fac], nrproc)
    i_split = np.array_split(range(len(t_solution[::fac])), nrproc)
    with mp.Pool(nrproc) as pool:
        result = pool.starmap(f, [(fft_green2, solution_from_fft,fft_omegas,
                                   fac, Is, ts) for Is, ts
                                  in zip(i_split, t_split)])
    for i, val in result:
        adj[i] = val
    plt.plot(t_solution[::fac], adj)
    plt.show()
    return new_s


@timeit
def f(fft_green2, solution, fft_omegas, fac, Is, ts):
    ret = np.zeros(ts.shape)
    for i, tau in enumerate(ts):
        #fft_backprop = exp(-1.j * tau * fft_omegas) * (-1) * fft_omegas * fft_omegas * fft_green2
        fft_backprop = exp(-1.j * tau * fft_omegas) * 1.j * fft_omegas * fft_green2
        #fft_backprop = exp(-1.j * tau * fft_omegas) * fft_green2
        td_backprop = real(ifft(fft_backprop)) / (1.0 / fac)
        #td_backprop = td_backprop[::-1]
        affe = np.multiply(td_backprop, solution)
        ret[i] = np.trapz(affe)
    return Is, ret


def create_pressure_seismogram(seismogram, source_pos, source_t_offset, rho, f0, freqs, c0, tau_p, L,
                               minimal_freq_count=None, time_interval=None):
    spec = seismogram.spec

    if time_interval is None:
        time_interval = [spec.t_start, spec.t_end]
    dt = spec.t_delta

    new_s = copy.deepcopy(seismogram)
    for ts in new_s.timeseries:
        receiver_pos = np.array([ts.receiver.x, ts.receiver.y])
        t, p = get_pressure_trace_freq_VA(receiver_pos=receiver_pos,
                                          dt=dt, time_interval=time_interval,
                                          source_pos=source_pos, source_t_offset=source_t_offset,
                                          rho=rho, f0=f0, freqs=freqs, c0=c0, tau_p=tau_p,
                                          minimal_freq_count=minimal_freq_count)
        ts.times = t.copy()
        # ts.states = p.reshape((len(p), 1))
        # ts.states = np.zeros((len(p),3+L))
        ts.states = np.zeros((len(p), 1))
        ts.states[:, 0] = p

    return new_s


def read_pressure_seismogram(fn, comp=2, state_components_override=False):
    s = FWISeismogram.readFromFile(fn)
    s_p = s.extractComponent(comp)

    # p = p0 + p1 + ... + pL
    if state_components_override is False:
        for l in range(1, s.state_components - 2):
            s_pl = s.extractComponent(2 + l)
            s_p = s_p + s_pl

    s_p.label = fn
    return s_p


def get_va_parameters(L):
    if L == 1:
        Q0 = 100.00
        tau_p = 2. / Q0
        freqs = np.array([25.0])
    elif L == 5:
        tau_p = 0.0767
        freqs = np.array([0.4971, 2.1266, 10.3723, 46.693, 120.0193])

        # Bohlen
        # tau_p = 0.056
        # freqs = np.array([0.28,  1.15, 5.72, 26.2, 76.5])
    else:
        raise ValueError(f"Only L=1 and L=5 are supported (requested L={L})")

    return tau_p, freqs


if __name__ == "__main__":

    c0 = 1.
    rho = 1.
    f0 = 25.0

    minimal_freq_count = 8192 * 2
    #    minimal_freq_count = None

    # L = 1
    L = 5
    # L =0
    print("Using L =", L)

    tau_p, freqs = get_va_parameters(L)

    print("Using tau_p={}".format(tau_p))

    source_pos = np.array([1.51, 2.0000]),
    # source_pos = np.array([600.000, 1000.0000]),
    # source_pos = np.array([600.000, 1000.0000]),

    # source_pos = np.array([600.000, 370.0000]),
    source_t_offset = 1.5 / f0
    # source_t_offset = 1.5e-5 * 2e5
    source_func = lambda t: dt_ricker(t, freq=f0)
    source_time_interval = [-2. / f0, 2. / f0]

    S = []

    # for fn in sys.argv[1:]:
    #   s_p = read_pressure_seismogram(fn)
    #   S += [s_p]
    # fn_list = ["/home/cr/FWI/tools/far/s_point_point_level_3_dt_8e-05_deg_4_rkorder_-2_shot_0_far"]
    # for fn in fn_list:
    s_p = read_pressure_seismogram("/home/cr/Documents/DissNEW/PythonScripts/PlotPdfs/Obs2Damping__",
                                   comp=0,
                                   state_components_override=True)
    S += [s_p]
    getAdjoint(S[-1], source_pos=source_pos, L=L,
               source_t_offset=source_t_offset,
               rho=rho, f0=f0, freqs=freqs, c0=c0, tau_p=tau_p,
               minimal_freq_count=minimal_freq_count)
    exit("ciao")
    s_sol = create_pressure_seismogram(S[-1], source_pos=source_pos, L=L,
                                       source_t_offset=source_t_offset,
                                       rho=rho, f0=f0, freqs=freqs, c0=c0, tau_p=tau_p,
                                       minimal_freq_count=minimal_freq_count)

    s_sol.writeToFile("/tmp/s_sol")
    s_sol_p = read_pressure_seismogram("/tmp/s_sol",
                                       comp=0,
                                       state_components_override=True)
    plot_spectra = 0
    if plot_spectra:
        print("WARNING: hardcoded stuff....")

        plot_magnitude_spectrum([s_sol], 0, figsize=(6, 4), xlim=(0, 270), ylim=(-1000, 1000),
                                filename="exact_amplitude.pdf")

        plot_phase_spectrum([s_sol], 0, figsize=(6, 4), xlim=(0, 250), ylim=(-350, 350), filename="exact_phase.pdf")

    else:
        d_min, d_max = np.inf, -np.inf
        for seis in S + [s_sol_p]:
            d_min = min(d_min, seis.min())
            d_max = max(d_max, seis.max())

        # data_factor = "normalized"
        data_factor = 1

        t0 = 0.2
        #    t0 = -np.inf
        norm_s_sol = s_sol_p.norm2(t0=t0)
        print(norm_s_sol)
        s_p *= 1.0 / 10000.0
        for s_p in S:
            diff = s_sol_p - s_p
            # print(diff)
            # print(s_p.norm2(t0=0.22))
            # print(diff.norm2(t0=0.22))
            print("Rel. L2-Difference for {}: ".format(s_p.label), diff.norm2(t0=t0) / norm_s_sol)

        figsize = (10, 10)
        figure = plt.figure(figsize=figsize)
        for s_p in S:
            s_p.plot(data_factor=data_factor, data_range=(d_min, d_max),
                     figure=figure, show=False, fill_below=False, r_step=3, time_range=[0, 1.5])

        s_sol_p.label = "ref"
        s_sol_p.plot(data_factor=data_factor, data_range=(d_min, d_max),
                     figure=figure, show=False, fill_below=False,
                     linestyle="--", r_step=3, time_range=[0, 1.5])

        plt.legend(bbox_to_anchor=(0., 1.02, 1., .102), loc=3,
                   ncol=2, mode="expand", borderaxespad=0.)
        plt.show()
